/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"


static int
check_logfile(const int debuglvl, const char *logloc)
{
	int	fd;

	/* safetly */
	if(!logloc)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if((fd = open(logloc, O_RDONLY)) == -1)
	{
		/* logfile does not yet exist, try to create it. */
		if((fd = open(logloc, O_WRONLY|O_CREAT|O_EXCL, 0600)) == -1)
		{
			(void)vrprint.error(-1, "Error", "creating of logfile '%s' failed: %s.", logloc, strerror(errno));
			return(-1);
		}
		close(fd);
	}
	else
	{
		close(fd);
	}

	return(0);
}


int
config_check_logdir(const int debuglvl, const char *logdir)
{
	DIR	*dir_p = NULL;

	/* safetly */
	if(!logdir)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* try to open */
	if(!(dir_p = opendir(logdir)))
	{
		(void)vrprint.error(-1, "Error", "opening '%s' failed: %s. Please select an existing directory.", logdir, strerror(errno));
		return(-1);
	}

	closedir(dir_p);
	return(0);
}


/* difference with config_check_logdir:
   this functions uses vuurmuur_opendir which repairs permissions if needed. */
int
config_check_vuurmuurdir(const int debuglvl, const char *logdir)
{
	DIR	*dir_p = NULL;

	/* safetly */
	if(!logdir)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* this isn't the right approach _at all_ but i'm don't really know how to do it better */
	if(
		strcmp(logdir, "/bin") == 0 ||
		strcmp(logdir, "/boot") == 0 ||
		strcmp(logdir, "/dev") == 0 ||
		strcmp(logdir, "/etc") == 0 ||
		strcmp(logdir, "/home") == 0 ||
		strcmp(logdir, "/lib") == 0 ||
		strcmp(logdir, "/mnt") == 0 ||
		strcmp(logdir, "/opt") == 0 ||
		strcmp(logdir, "/proc") == 0 ||
		strcmp(logdir, "/root") == 0 ||
		strcmp(logdir, "/sbin") == 0 ||
		strcmp(logdir, "/sys") == 0 ||
		strcmp(logdir, "/tmp") == 0 ||

		strcmp(logdir, "/usr") == 0 ||
		strcmp(logdir, "/usr/local") == 0 ||
		strcmp(logdir, "/usr/share") == 0 ||
		strcmp(logdir, "/usr/lib") == 0 ||
		strcmp(logdir, "/usr/bin") == 0 ||
		strcmp(logdir, "/usr/sbin") == 0 ||

		strcmp(logdir, "/var") == 0 ||
		strcmp(logdir, "/var/log") == 0 ||

		strcmp(logdir, "/") == 0
	)
	{
		(void)vrprint.error(-1, "Error", "directory '%s' is on my blacklist. Please select another.", logdir);
		return(-1);
	}

	/* try to open, error reporting is done by others */
	if(!(dir_p = vuurmuur_opendir(logdir)))
	{
		return(-1);
	}

	closedir(dir_p);
	return(0);
}


/*
	TODO:
		check if scripts dir exists
*/
int
check_iptables_command(const int debuglvl, char *iptables_location, char quiet)
{
	char	cmd[sizeof(conf.iptables_location) + 32] = "";

	/* safety */
	if(!iptables_location)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* first check if there even is a value */
	if(strcmp(iptables_location, "") == 0)
	{
		if(!quiet)
			(void)vrprint.error(0, "Error", "The path to the 'iptables'-command was not set.", iptables_location);

		return(0);
	}
	else
	{
		/* now check the command */
		snprintf(cmd, sizeof(cmd), "%s --version &> /dev/null", iptables_location);

		if(pipe_command(debuglvl, cmd, PIPE_QUIET) < 0)
		{
			if(!quiet)
				(void)vrprint.error(0, "Error", "The path '%s' to the 'iptables'-command seems to be wrong.", iptables_location);

			return(0);
		}
	}

	return(1);
}


/*
	TODO:
		check if scripts dir exists
*/
int
check_iptablesrestore_command(const int debuglvl, char *iptablesrestore_location, char quiet)
{
	char	cmd[sizeof(conf.iptablesrestore_location) + 32] = "";

	/* safety */
	if(!iptablesrestore_location)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* first check if there even is a value */
	if(strcmp(iptablesrestore_location, "") == 0)
	{
		if(!quiet)
			(void)vrprint.error(0, "Error", "The path to the 'iptables-restore'-command was not set.", iptablesrestore_location);

		return(0);
	}
	else
	{
		/* now check the command */
		snprintf(cmd, sizeof(cmd), "%s < /dev/null &> /dev/null", iptablesrestore_location);

		if(pipe_command(debuglvl, cmd, PIPE_QUIET) < 0)
		{
			if(!quiet)
				(void)vrprint.error(0, "Error", "The path '%s' to the 'iptables-restore'-command seems to be wrong.", iptablesrestore_location);

			return(0);
		}
	}

	return(1);
}


/* updates the logdirlocations in the cnf struct based on cnf->vuurmuur_log_dir */
int
config_set_log_names(const int debuglvl, struct vuurmuur_config *cnf)
{
	int	retval = 0;

	/* safety */
	if(!cnf)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(snprintf(cnf->vuurmuurlog_location, sizeof(cnf->vuurmuurlog_location), "%s/vuurmuur.log", cnf->vuurmuur_logdir_location) >= sizeof(cnf->vuurmuurlog_location))
	{
		(void)vrprint.error(-1, "Error", "vuurmuur.log location was truncated (in: %s:%d).", __FUNCTION__, __LINE__);
		retval = -1;
	}

	if(snprintf(cnf->trafficlog_location,  sizeof(cnf->trafficlog_location),  "%s/traffic.log",  cnf->vuurmuur_logdir_location) >= sizeof(cnf->trafficlog_location))
	{
		(void)vrprint.error(-1, "Error", "traffic.log location was truncated (in: %s:%d).", __FUNCTION__, __LINE__);
		retval = -1;
	}

	if(snprintf(cnf->debuglog_location,    sizeof(cnf->debuglog_location),    "%s/debug.log",    cnf->vuurmuur_logdir_location) >= sizeof(cnf->debuglog_location))
	{
		(void)vrprint.error(-1, "Error", "debug.log location was truncated (in: %s:%d).", __FUNCTION__, __LINE__);
		retval = -1;
	}

	if(snprintf(cnf->errorlog_location,    sizeof(cnf->errorlog_location),    "%s/error.log",    cnf->vuurmuur_logdir_location) >= sizeof(cnf->errorlog_location))
	{
		(void)vrprint.error(-1, "Error", "error.log location was truncated (in: %s:%d).", __FUNCTION__, __LINE__);
		retval = -1;
	}

	return(retval);
}


/*
	Note: we cannot use vrprint.debug and vrprint.info in this, because in most cases
	we want those function to print to the log, however the log locations are only known
	after this function! (unless libvuurmuur_verbose == 1)
*/
int
init_config(const int debuglvl, char *configfile_location, struct vuurmuur_config *cnf)
{
	int	retval = VR_CNF_OK,
		result = 0;
	char	answer[32] = "";
	FILE	*fp = NULL;


	/* safety first */
	if(!configfile_location || !cnf)
		return(VR_CNF_E_PARAMETER);

	/* check the file */
	if(!(fp = fopen(configfile_location, "r")))
	{
		(void)vrprint.error(-1, "Error", "could not open configfile '%s': %s (in: %s).", configfile_location, strerror(errno), __FUNCTION__);
		if(errno == ENOENT)
			return(VR_CNF_E_FILE_MISSING);
		else if(errno == EACCES)
			return(VR_CNF_E_FILE_PERMISSION);
		else
			return(VR_CNF_E_UNKNOWN_ERR);
	}
	fclose(fp);

	/* check if we like the configfile */
	if(!(stat_ok(debuglvl, configfile_location)))
		return(VR_CNF_E_FILE_PERMISSION);


	result = ask_configfile(debuglvl, "SERVICES_BACKEND", cnf->serv_backend_name, configfile_location, sizeof(cnf->serv_backend_name));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "Variable SERVICES_BACKEND not found in '%s'.", configfile_location);
		retval = VR_CNF_E_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "ZONES_BACKEND", cnf->zone_backend_name, configfile_location, sizeof(cnf->zone_backend_name));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "Variable ZONES_BACKEND not found in '%s'.", configfile_location);
		retval = VR_CNF_E_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "INTERFACES_BACKEND", cnf->ifac_backend_name, configfile_location, sizeof(cnf->ifac_backend_name));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "Variable INTERFACES_BACKEND not found in '%s'.", configfile_location);
		retval = VR_CNF_E_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "PLUGINDIR", cnf->plugindir, configfile_location, sizeof(cnf->plugindir));
	if(result == 1)
	{
		if(libvuurmuur_verbose > 0)
			(void)vrprint.info("Info", "Using '%s' as plugindir.", cnf->plugindir);
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "Variable PLUGINDIR not found in '%s'.", configfile_location);
		retval = VR_CNF_E_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "RULESFILE", cnf->rules_location, configfile_location, sizeof(cnf->rules_location));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.error(-1, "Error", "Variable RULESFILE not found in '%s'.", configfile_location);
		retval = VR_CNF_E_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "BLOCKLISTFILE", cnf->blocklist_location, configfile_location, sizeof(cnf->blocklist_location));
	if(result == 1)
	{
		/* ok, found */
		if(strcmp(cnf->blocklist_location, "") == 0 || strcasecmp(cnf->blocklist_location, "none") == 0)
		{
			(void)vrprint.info("Info", "not using blocklist.");
			cnf->use_blocklist = 0;
		}
		else
		{
			cnf->use_blocklist = 1;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable BLOCKLISTFILE not found in '%s'. Disabling blocklist.", configfile_location);
		cnf->use_blocklist = 0;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* OLD_CREATE_METHOD */
	result = ask_configfile(debuglvl, "OLD_CREATE_METHOD", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->old_rulecreation_method = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->old_rulecreation_method = 0;
		}
		else
		{
			(void)vrprint.warning("Warning", "'%s' is not a valid value for option OLD_CREATE_METHOD.", answer);
			cnf->old_rulecreation_method = DEFAULT_OLD_CREATE_METHOD;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		cnf->old_rulecreation_method = DEFAULT_OLD_CREATE_METHOD;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* DYN_INT_CHECK */
	result = ask_configfile(debuglvl, "DYN_INT_CHECK", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->dynamic_changes_check = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->dynamic_changes_check = 0;
		}
		else
		{
			(void)vrprint.warning("Warning", "'%s' is not a valid value for option DYN_INT_CHECK.", answer);
			cnf->dynamic_changes_check = DEFAULT_DYN_INT_CHECK;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		cnf->dynamic_changes_check = DEFAULT_DYN_INT_CHECK;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* LOG_POLICY_LIMIT */
	result = ask_configfile(debuglvl, "DYN_INT_INTERVAL", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative DYN_INT_INTERVAL-limit (%d) can not be used, using default (%d).", result, DEFAULT_DYN_INT_INTERVAL);
			cnf->dynamic_changes_interval = DEFAULT_DYN_INT_INTERVAL;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->dynamic_changes_interval = result;
		}
	}
	else if(result == 0)
	{
		cnf->dynamic_changes_interval = DEFAULT_DYN_INT_INTERVAL;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* SYN_LIMIT */
	result = ask_configfile(debuglvl, "SYN_LIMIT", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative SYN-limit (%d) can not be used, using default (%d).", result, DEFAULT_SYN_LIMIT);
			cnf->syn_limit = DEFAULT_SYN_LIMIT;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else if(result == 0)
		{
			(void)vrprint.warning("Warning", "A SYN-limit of 0 can not be used, using default (%d).", DEFAULT_SYN_LIMIT);
			cnf->syn_limit = DEFAULT_SYN_LIMIT;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->syn_limit = result;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable SYN_LIMIT not found in '%s'. Using default.", configfile_location);
		cnf->syn_limit = DEFAULT_SYN_LIMIT;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* SYN_LIMIT_BURST */
	result = ask_configfile(debuglvl, "SYN_LIMIT_BURST", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative SYN-limit-burst (%d) can not be used, using default (%d).", result, DEFAULT_SYN_LIMIT_BURST);
			cnf->syn_limit_burst = DEFAULT_SYN_LIMIT_BURST;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else if(result == 0)
		{
			(void)vrprint.warning("Warning", "A SYN-limit-burst of 0 can not be used, using default (%d).", DEFAULT_SYN_LIMIT_BURST);
			cnf->syn_limit_burst = DEFAULT_SYN_LIMIT_BURST;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->syn_limit_burst = result;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable SYN_LIMIT_BURST not found in '%s'. Using default.", configfile_location);
		cnf->syn_limit_burst = DEFAULT_SYN_LIMIT_BURST;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* UDP_LIMIT */
	result = ask_configfile(debuglvl, "UDP_LIMIT", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative UDP-limit (%d) can not be used, using default (%d).", result, DEFAULT_UDP_LIMIT);
			cnf->udp_limit = DEFAULT_UDP_LIMIT;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else if(result == 0)
		{
			(void)vrprint.warning("Warning", "A UDP-limit of 0 can not be used, using default (%d).", DEFAULT_UDP_LIMIT);
			cnf->udp_limit = DEFAULT_UDP_LIMIT;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->udp_limit = result;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable UDP_LIMIT not found in '%s'. Using default.", configfile_location);
		cnf->udp_limit = DEFAULT_UDP_LIMIT;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* UDP_LIMIT_BURST */
	result = ask_configfile(debuglvl, "UDP_LIMIT_BURST", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative UDP-limit-burst (%d) can not be used, using default (%d).", result, DEFAULT_UDP_LIMIT_BURST);
			cnf->udp_limit_burst = DEFAULT_UDP_LIMIT_BURST;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else if(result == 0)
		{
			(void)vrprint.warning("Warning", "A UDP-limit-burst of 0 can not be used, using default (%d).", DEFAULT_UDP_LIMIT_BURST);
			cnf->udp_limit_burst = DEFAULT_UDP_LIMIT_BURST;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->udp_limit_burst = result;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable UDP_LIMIT_BURST not found in '%s'. Using default.", configfile_location);
		cnf->udp_limit_burst = DEFAULT_UDP_LIMIT_BURST;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* LOG_POLICY */
	result = ask_configfile(debuglvl, "LOG_POLICY", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->log_policy = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->log_policy = 0;
		}
		else
		{
			(void)vrprint.warning("Warning", "'%s' is not a valid value for option LOG_POLICY.", answer);
			cnf->log_policy = DEFAULT_LOG_POLICY;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable LOG_POLICY not found in '%s'. Using default.", configfile_location);
		cnf->log_policy = DEFAULT_LOG_POLICY;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* LOG_POLICY_LIMIT */
	result = ask_configfile(debuglvl, "LOG_POLICY_LIMIT", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		result = atoi(answer);
		if(result < 0)
		{
			(void)vrprint.warning("Warning", "A negative LOG-limit (%d) can not be used, using default (%d).", result, DEFAULT_LOG_POLICY_LIMIT);
			cnf->log_policy_limit = DEFAULT_LOG_POLICY_LIMIT;
			cnf->log_policy_burst = cnf->log_policy_limit * 2;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
		else
		{
			cnf->log_policy_limit = result;
			cnf->log_policy_burst = cnf->log_policy_limit * 2;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable LOG_POLICY_LIMIT not found in '%s'. Using default.", configfile_location);

		cnf->log_policy_limit = DEFAULT_LOG_POLICY_LIMIT;
		cnf->log_policy_burst = cnf->log_policy_limit * 2;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* PROTECT_SYNCOOKIES */
	result = ask_configfile(debuglvl, "PROTECT_SYNCOOKIE", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->protect_syncookie = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->protect_syncookie = 0;
		}
		else
		{
			(void)vrprint.warning("Warning", "'%s' is not a valid value for option PROTECT_SYNCOOKIE.", answer);
			cnf->protect_syncookie = DEFAULT_PROTECT_SYNCOOKIE;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable PROTECT_SYNCOOKIE not found in '%s'. Using default.", configfile_location);
		cnf->protect_syncookie = DEFAULT_PROTECT_SYNCOOKIE;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* PROTECT_ECHOBROADCAST */
	result = ask_configfile(debuglvl, "PROTECT_ECHOBROADCAST", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->protect_echobroadcast = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->protect_echobroadcast = 0;
		}
		else
		{
			(void)vrprint.warning("Warning", "'%s' is not a valid value for option PROTECT_ECHOBROADCAST.", answer);
			cnf->protect_echobroadcast = DEFAULT_PROTECT_ECHOBROADCAST;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable PROTECT_ECHOBROADCAST not found in '%s'. Using default.", configfile_location);
		cnf->protect_echobroadcast = DEFAULT_PROTECT_ECHOBROADCAST;

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "IPTABLES", cnf->iptables_location, configfile_location, sizeof(cnf->iptables_location));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable IPTABLES not found in '%s', using default value.", configfile_location);
		strlcpy(cnf->iptables_location, DEFAULT_IPTABLES_LOCATION, sizeof(cnf->iptables_location));

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);

	result = ask_configfile(debuglvl, "IPTABLES_RESTORE", cnf->iptablesrestore_location, configfile_location, sizeof(cnf->iptablesrestore_location));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable IPTABLES_RESTORE not found in '%s', using default value.", configfile_location);
		strlcpy(cnf->iptablesrestore_location, DEFAULT_IPTABLES_REST_LOCATION, sizeof(cnf->iptablesrestore_location));

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	result = ask_configfile(debuglvl, "MODPROBE", cnf->modprobe_location, configfile_location, sizeof(cnf->modprobe_location));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		strlcpy(cnf->modprobe_location, DEFAULT_MODPROBE_LOCATION, sizeof(cnf->modprobe_location));
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* PROTECT_SYNCOOKIES */
	result = ask_configfile(debuglvl, "LOAD_MODULES", answer, configfile_location, sizeof(answer));
	if(result == 1)
	{
		/* ok, found */
		if(strcasecmp(answer, "yes") == 0)
		{
			cnf->load_modules = 1;
		}
		else if(strcasecmp(answer, "no") == 0)
		{
			cnf->load_modules = 0;
		}
		else
		{
			cnf->load_modules = DEFAULT_LOAD_MODULES;

			retval = VR_CNF_W_ILLEGAL_VAR;
		}
	}
	else if(result == 0)
	{
		cnf->load_modules = DEFAULT_LOAD_MODULES;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* directory containing scripts */
	result = ask_configfile(debuglvl, "SCRIPTS_DIR", cnf->scripts_location, configfile_location, sizeof(cnf->scripts_location));
	if(result == 1)
	{
		/* ok */
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable SCRIPTS_DIR not found in '%s', using default value: %s", configfile_location, DEFAULT_SCRIPTS_DIR);
		strlcpy(cnf->scripts_location, DEFAULT_SCRIPTS_DIR, sizeof(cnf->scripts_location));

		retval = VR_CNF_W_MISSING_VAR;
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* check if the configfile value is overridden by the commandline */
	if(cnf->loglevel_cmdline == 0)
	{
		result = ask_configfile(debuglvl, "LOGLEVEL", cnf->loglevel, configfile_location, sizeof(cnf->loglevel));
		if(result == 1)
		{
			// ok
			if(libvuurmuur_verbose > 0)
				(void)vrprint.info("Info", "Loglevel is '%s'.", cnf->loglevel);
		}
		else if(result == 0)
		{
			(void)vrprint.warning("Warning", "Variable LOGLEVEL not found in the configfile, using default value.");
			strcpy(cnf->loglevel, "");
			retval = VR_CNF_W_MISSING_VAR;
		}
		else
			return(VR_CNF_E_UNKNOWN_ERR);
	}


	/* systemlog */
	result = ask_configfile(debuglvl, "SYSTEMLOG", cnf->systemlog_location, configfile_location, sizeof(cnf->systemlog_location));
	if (result == 1 )
	{
		/* ok */
		if(libvuurmuur_verbose > 0)
			(void)vrprint.info("Info", "Using '%s' as systemlogfile.", cnf->systemlog_location);
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable SYSTEMLOG not found in '%s', using default value (%s).", configfile_location, DEFAULT_SYSTEMLOG_LOCATION);
		strlcpy(cnf->systemlog_location, DEFAULT_SYSTEMLOG_LOCATION, sizeof(cnf->systemlog_location));
		retval = VR_CNF_W_MISSING_VAR;
//TODO: check if location really exists
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* get the logfile dir */
	result = ask_configfile(debuglvl, "LOGDIR", cnf->vuurmuur_logdir_location, configfile_location, sizeof(cnf->vuurmuur_logdir_location));
	if(result == 1)
	{
		if(libvuurmuur_verbose > 0)
			(void)vrprint.info("Info", "Using '%s' as normal logdir.", cnf->vuurmuur_logdir_location);
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "Variable LOGDIR not found in '%s', using default value.", configfile_location);
		strlcpy(cnf->vuurmuur_logdir_location, DEFAULT_LOGDIR_LOCATION, sizeof(cnf->vuurmuur_logdir_location));
		
		/* we return here because we don't want the logfile checks if we know
		   that this is wrong. */
		return(VR_CNF_W_MISSING_VAR);
	}
	else
		return(VR_CNF_E_UNKNOWN_ERR);


	/* check plugin dir */
	if(config_check_vuurmuurdir(debuglvl, cnf->plugindir) < 0)
		return(VR_CNF_W_ILLEGAL_VAR);
	
	/* check scripts dir */
	if(config_check_vuurmuurdir(debuglvl, cnf->scripts_location) < 0)
		return(VR_CNF_W_ILLEGAL_VAR);

	/* check if we can open the logdir */
	if(config_check_logdir(debuglvl, cnf->vuurmuur_logdir_location) < 0)
		return(VR_CNF_W_ILLEGAL_VAR);


	/* set/update the lognames */
	if(config_set_log_names(debuglvl, cnf) < 0)
		return(VR_CNF_E_UNKNOWN_ERR);


	/* vuurmuur.log */
	if(libvuurmuur_verbose > 0)
		(void)vrprint.info("Info", "Using '%s' as vuurmuur.log.", cnf->vuurmuurlog_location);

	if(check_logfile(debuglvl, cnf->vuurmuurlog_location) < 0)
	{
		retval = VR_CNF_E_ILLEGAL_VAR;
	}


	/* error.log */
	if(libvuurmuur_verbose > 0)
		(void)vrprint.info("Info", "Using '%s' as error.log.", cnf->errorlog_location);

	if(check_logfile(debuglvl, cnf->errorlog_location) < 0)
	{
		retval = VR_CNF_E_ILLEGAL_VAR;
	}


	/* debug.log */
	if(libvuurmuur_verbose > 0)
		(void)vrprint.info("Info", "Using '%s' as debug.log.", cnf->debuglog_location);

	if(check_logfile(debuglvl, cnf->debuglog_location) < 0)
	{
		retval = VR_CNF_E_ILLEGAL_VAR;
	}


	/* traffic.log */
	if(libvuurmuur_verbose > 0)
		(void)vrprint.info("Info", "Using '%s' as traffic.log.", cnf->trafficlog_location);

	if(check_logfile(debuglvl, cnf->trafficlog_location) < 0)
	{
		retval = VR_CNF_E_ILLEGAL_VAR;
	}

	/* check the iptables command */
	if(!check_iptables_command(debuglvl, cnf->iptables_location, 0))
	{
		retval = VR_CNF_E_ILLEGAL_VAR;
	}
	/* if we are going to use the iptables-restore command, check it */
	if(cnf->old_rulecreation_method == 0)
	{
		if(!check_iptablesrestore_command(debuglvl, cnf->iptablesrestore_location, 0))
		{
			retval = VR_CNF_E_ILLEGAL_VAR;
		}
	}

	return(retval);
}


int
reload_config(const int debuglvl, struct vuurmuur_config *old_cnf)
{
	struct vuurmuur_config	new_cnf;
	int			retval = VR_CNF_OK;

	/* safety */
	if(!old_cnf)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(VR_CNF_E_PARAMETER);
	}

	/* loglevel can be overrided by commandline */
	new_cnf.loglevel_cmdline = old_cnf->loglevel_cmdline;

	/* reload the configfile */
	if((retval = init_config(debuglvl, old_cnf->configfile_location, &new_cnf)) >= VR_CNF_OK)
	{
		/* first copy the config file location to the new config since it is not loaded by init_config */
		strlcpy(new_cnf.configfile_location, old_cnf->configfile_location, sizeof(new_cnf.configfile_location));

		/* rule creation method is not allowed to change */
		new_cnf.old_rulecreation_method = old_cnf->old_rulecreation_method;

		/* in old_create_method mode, loglevel is not allowed to change at runtime */
		if(new_cnf.old_rulecreation_method == 1)
		{
			strlcpy(new_cnf.loglevel, old_cnf->loglevel, sizeof(new_cnf.loglevel));
		}

		/* copy the data to the old struct */
		*old_cnf = new_cnf;
	}

	return(retval);
}


/* ask_configfile

    This function ask questions from the configfile.

    Returncodes:
     1: ok
     0: ok, but question not found.
    -1: error
*/
int
ask_configfile(const int debuglvl, char *question, char *answer_ptr, char *file_location, size_t size)
{
	int	retval=0,
		i=0,
		k=0,
		l=0,
		j=0;
	FILE	*fp = NULL;
	char	line[512],
		variable[128],
		value[256];

	if(!question || !file_location || size <= 0)
		return(-1);

	if(!(fp = vuurmuur_fopen(file_location,"r")))
	{
		(void)vrprint.error(-1, "Error", "unable to open configfile '%s': %s (in: ask_configfile).", file_location, strerror(errno));
		return(-1);
	}
	else
	{
		while(fgets(line, 512, fp) != NULL)
		{
			i=strlen(line);
			k=0; l=0; j=0;

			if ((line[0] == '#' ) || (strlen(line) < 1) || (line[0] == '\n'))
			{
				//do nothing, its a comment.
			}
			else
			{
				// variable
				while(line[k] != '=' && k < size)
				{
					variable[j]=line[k];
					k++; j++;
				}
				variable[j]='\0';

				if (debuglvl >= HIGH && libvuurmuur_verbose)
					(void)vrprint.debug("Debug", "ask_configfile: question '%s' variable '%s' (%d)", question, variable, size);


				// value
				j=0;
				l=k+1;
				while(line[l] != '\0' && line[l] != '\n' && l < (size+k+1))
				{
					// if the first character is a '"' we strip it.
					if(j == 0 && line[l] == '\"')
					{
						l++;
					}
					else
					{
						value[j]=line[l];
						l++;
						j++;
					}
				}
				// if the last character is a'"' we strip it.
				if(j > 0)
				{
					if(value[j-1] == '\"')
					{
						value[j-1]='\0';
					}
					else
						value[j]='\0';
				}
				else
					value[j]='\0';

				if (debuglvl >= HIGH && libvuurmuur_verbose)
					(void)vrprint.debug("Debug", "ask_configfile: question '%s' value '%s' (%d)", question, value, size);

				if(strcmp(question, variable) == 0)
				{
					if (debuglvl >= HIGH && libvuurmuur_verbose)
						(void)vrprint.debug("Debug", "ask_configfile: question '%s' matched, value: '%s'", question, value);

					strlcpy(answer_ptr, value, size);

					retval=1;
					break;
				}
			}
		}

		if(fclose(fp) != 0)
		{
			(void)vrprint.error(-1, "Error", "closing file '%s' failed: %s.", file_location, strerror(errno));
			retval=-1;
		}
	}
	if (debuglvl >= HIGH && libvuurmuur_verbose)
		(void)vrprint.debug("Debug", "ask_configfile: ** end **, return=%d", retval);

	return(retval);
}

/*	write_configfile

	Writes the config to disk.
*/
int
write_configfile(const int debuglvl, char *file_location)
{
	FILE *fp = NULL;

	/* safety */
	if(!file_location)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* open for over-writing */
	if(!(fp = fopen(file_location, "w+")))
	{
		(void)vrprint.error(-1, "Error", "unable to open configfile for writing: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* start writing the config to the file */
	fprintf(fp, "# vuurmuur config file\n\n");

	fprintf(fp, "# The directory containting the plugins (full path).\n");
	fprintf(fp, "PLUGINDIR=\"%s\"\n\n", conf.plugindir);

	fprintf(fp, "# Which plugin to use for which type of data.\n");
	fprintf(fp, "SERVICES_BACKEND=\"%s\"\n\n", conf.serv_backend_name);
	fprintf(fp, "ZONES_BACKEND=\"%s\"\n\n", conf.zone_backend_name);
	fprintf(fp, "INTERFACES_BACKEND=\"%s\"\n\n", conf.ifac_backend_name);

	fprintf(fp, "# Location of the rulesfile (full path).\n");
	fprintf(fp, "RULESFILE=\"%s\"\n\n", conf.rules_location);
	fprintf(fp, "# Location of the blocklistfile (full path).\n");
	fprintf(fp, "BLOCKLISTFILE=\"%s\"\n\n", conf.blocklist_location);

	fprintf(fp, "# Location of the iptables-command (full path).\n");
	fprintf(fp, "IPTABLES=\"%s\"\n\n", conf.iptables_location);
	fprintf(fp, "# Location of the iptables-restore-command (full path).\n");
	fprintf(fp, "IPTABLES_RESTORE=\"%s\"\n\n", conf.iptablesrestore_location);

	fprintf(fp, "# Location of the modprobe-command (full path).\n");
	fprintf(fp, "MODPROBE=\"%s\"\n\n", conf.modprobe_location);

	fprintf(fp, "# Load modules if needed? (yes/no)\n");
	fprintf(fp, "LOAD_MODULES=\"%s\"\n\n", conf.load_modules ? "Yes" : "No");

	fprintf(fp, "# If set to yes, each rule will be loaded into the system individually using\n");
	fprintf(fp, "# iptables. Otherwise iptables-restore will be used (yes/no).\n");
	fprintf(fp, "OLD_CREATE_METHOD=\"%s\"\n\n", conf.old_rulecreation_method ? "Yes" : "No");

	fprintf(fp, "# The directory containting the scripts (full path).\n");
	fprintf(fp, "SCRIPTS_DIR=\"%s\"\n\n", conf.scripts_location);

	fprintf(fp, "# The directory where the logs will be written to (full path).\n");
	fprintf(fp, "LOGDIR=\"%s\"\n\n", conf.vuurmuur_logdir_location);
	fprintf(fp, "# The logfile where the kernel writes the logs to e.g. /var/log/messages (full path).\n");
	fprintf(fp, "SYSTEMLOG=\"%s\"\n\n", conf.systemlog_location);
	fprintf(fp, "# The loglevel to use when logging traffic. For use with syslog.\n");
	fprintf(fp, "LOGLEVEL=\"%s\"\n\n", conf.loglevel);

	fprintf(fp, "# Check the dynamic interfaces for changes?\n");
	fprintf(fp, "DYN_INT_CHECK=\"%s\"\n\n", conf.dynamic_changes_check ? "Yes" : "No");
	fprintf(fp, "# Check every x seconds.\n");
	fprintf(fp, "DYN_INT_INTERVAL=\"%d\"\n\n", conf.dynamic_changes_interval);
	
	fprintf(fp, "# LOG_POLICY controls the logging of the default policy.\n");
	fprintf(fp, "LOG_POLICY=\"%s\"\n\n", conf.log_policy ? "Yes" : "No");

	fprintf(fp, "# LOG_POLICY_LIMIT sets the maximum number of logs per second.\n");
	fprintf(fp, "LOG_POLICY_LIMIT=\"%d\"\n\n", conf.log_policy_limit);

	fprintf(fp, "# SYN_LIMIT sets the maximum number of SYN-packets per second.\n");
	fprintf(fp, "SYN_LIMIT=\"%d\"\n", conf.syn_limit);
	fprintf(fp, "SYN_LIMIT_BURST=\"%d\"\n\n", conf.syn_limit_burst);

	fprintf(fp, "# UDP_LIMIT sets the maximum number of udp 'connections' per second.\n");
	fprintf(fp, "UDP_LIMIT=\"%d\"\n", conf.udp_limit);
	fprintf(fp, "UDP_LIMIT_BURST=\"%d\"\n\n", conf.udp_limit_burst);

	/* protect */
	fprintf(fp, "# Protect against syn-flooding? (yes/no)\n");
	fprintf(fp, "PROTECT_SYNCOOKIE=\"%s\"\n", conf.protect_syncookie ? "Yes" : "No");

	fprintf(fp, "# Ignore echo-broadcasts? (yes/no)\n");
	fprintf(fp, "PROTECT_ECHOBROADCAST=\"%s\"\n\n", conf.protect_echobroadcast ? "Yes" : "No");
	
	fprintf(fp, "# end of file\n");

	/* flush buffer */
	fflush(fp);
	/* close file */
	fclose(fp);

	(void)vrprint.info("Info", "Rewritten config file.");
	return(0);
}
