/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"


/*	rule_malloc

	Allocates memory for a rule, and inits all variables.

	Returns the address of the memory on succes and NULL on failure.
*/
void *
rule_malloc(void)
{
	struct RuleData_ *rule_ptr = NULL;

	rule_ptr = malloc(sizeof(struct RuleData_));
	if(rule_ptr == NULL)
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(NULL);
	}

	memset(rule_ptr, 0, sizeof(struct RuleData_));

	rule_ptr->type = TYPE_RULE;

	rule_ptr->rulecache.from = NULL;
	rule_ptr->rulecache.to = NULL;
	rule_ptr->rulecache.who = NULL;
	rule_ptr->rulecache.service = NULL;
	rule_ptr->rulecache.description = NULL;

	rule_ptr->rulecache.option.rule_log=0;

	strcpy(rule_ptr->rulecache.option.logprefix, "");
	rule_ptr->rulecache.option.rule_logprefix=0;

	rule_ptr->rulecache.option.remoteport=0;
	rule_ptr->rulecache.option.RemoteportList.len = 0;

	rule_ptr->rulecache.option.listenport=0;
	rule_ptr->rulecache.option.ListenportList.len = 0;

	rule_ptr->rulecache.option.redirectport=0;

	rule_ptr->rulecache.option.reject_option=0;
	strcpy(rule_ptr->rulecache.option.reject_type, "");

	rule_ptr->rulecache.option.rule_comment = 0;
	strcpy(rule_ptr->rulecache.option.comment, "");

	rule_ptr->rulecache.from_firewall = 0;
	rule_ptr->rulecache.to_firewall = 0;
	rule_ptr->rulecache.ruletype = 0;

	rule_ptr->rulecache.iptcount.input = 0;
	rule_ptr->rulecache.iptcount.output = 0;
	rule_ptr->rulecache.iptcount.forward = 0;
	rule_ptr->rulecache.iptcount.preroute = 0;
	rule_ptr->rulecache.iptcount.postroute = 0;

	rule_ptr->rulecache.iptcount.start_input = 0;
	rule_ptr->rulecache.iptcount.start_output = 0;
	rule_ptr->rulecache.iptcount.start_forward = 0;
	rule_ptr->rulecache.iptcount.start_preroute = 0;
	rule_ptr->rulecache.iptcount.start_postroute = 0;

	rule_ptr->number = 0;
	rule_ptr->status = 0;
	rule_ptr->active = 0;

	strcpy(rule_ptr->action, "");
	strcpy(rule_ptr->service, "");
	strcpy(rule_ptr->from, "");
	strcpy(rule_ptr->to, "");
	strcpy(rule_ptr->danger, "");
	strcpy(rule_ptr->who, "");
	strcpy(rule_ptr->source, "");

	rule_ptr->opt = NULL;

	return(rule_ptr);
}


void *
ruleoption_malloc(const int debuglvl)
{
	struct options	*opt_ptr = NULL;

	/* alloc the memory */
	if(!(opt_ptr = malloc(sizeof(struct options))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(NULL);
	}
	/* initialize the mem */
	memset(opt_ptr, 0, sizeof(struct options));

	/* setup the lists */
	(void)d_list_setup(debuglvl, &opt_ptr->RemoteportList, NULL);
	(void)d_list_setup(debuglvl, &opt_ptr->ListenportList, NULL);

	return(opt_ptr);
}


void *
zone_malloc(const int debuglvl)
{
	struct ZoneData_ *zone_ptr = NULL;

	if(!(zone_ptr = malloc(sizeof(struct ZoneData_))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(NULL);
	}

	strcpy(zone_ptr->name, "");
	strcpy(zone_ptr->host_name, "");
	strcpy(zone_ptr->network_name, "");
	strcpy(zone_ptr->zone_name, "");

	zone_ptr->zone_parent = NULL;
	zone_ptr->network_parent = NULL;

	strcpy(zone_ptr->mac, "");
	zone_ptr->has_mac = 0;

	strcpy(zone_ptr->ipv4.ipaddress, "");
	strcpy(zone_ptr->ipv4.netmask, "");
	strcpy(zone_ptr->ipv4.network, "");
	strcpy(zone_ptr->ipv4.broadcast, "");

	zone_ptr->group_member_count = 0;

	zone_ptr->GroupList.len = 0;
	if(d_list_setup(debuglvl, &zone_ptr->GroupList, NULL) < 0)
		return(NULL);

	zone_ptr->interface_count = 0;
	zone_ptr->active_interfaces = 0;

	zone_ptr->InterfaceList.len = 0;
	if(d_list_setup(debuglvl, &zone_ptr->InterfaceList, NULL) < 0)
		return(NULL);
	
	zone_ptr->ProtectList.len = 0;
	if(d_list_setup(debuglvl, &zone_ptr->ProtectList, free) < 0)
		return(NULL);

	zone_ptr->active = 0;
	zone_ptr->status = 0;
	zone_ptr->type = TYPE_UNSET;

	return(zone_ptr);
}


void
zone_free(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	if(!zone_ptr)
		return;

	if(zone_ptr->type == TYPE_GROUP)
		(void)d_list_cleanup(debuglvl, &zone_ptr->GroupList);

	if(zone_ptr->type == TYPE_NETWORK)
	{
		(void)d_list_cleanup(debuglvl, &zone_ptr->InterfaceList);
		(void)d_list_cleanup(debuglvl, &zone_ptr->ProtectList);
	}

	free(zone_ptr);
}


void *
service_malloc(void)
{
	struct ServicesData_ *ser_ptr = NULL;

	/* alloc some mem */
	ser_ptr = malloc(sizeof(struct ServicesData_));
	if(ser_ptr == NULL)
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(NULL);
	}

	/* init */
	memset(ser_ptr, 0, sizeof(struct ServicesData_));

	ser_ptr->type = TYPE_SERVICE;

	return(ser_ptr);
}


/* returns a initialized interface memory area */
void *
interface_malloc(const int debuglvl)
{
	struct InterfaceData_ *iface_ptr = NULL;

	iface_ptr = malloc(sizeof(struct InterfaceData_));
	if(iface_ptr == NULL)
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(NULL);
	}

	memset(iface_ptr, 0, sizeof(struct InterfaceData_));

	iface_ptr->type = TYPE_INTERFACE;

	iface_ptr->active = 1;

	iface_ptr->ProtectList.len = 0;
	if(d_list_setup(debuglvl, &iface_ptr->ProtectList, free) < 0)
		return(NULL);

	iface_ptr->cnt = NULL;

	return(iface_ptr);
}


/*

*/
int
LockSHM(int lock, int sem_id, char quiet)
{
	int			z;
	static struct sembuf	sops = { 0, -1, 0 };

	sops.sem_num = 0;
	sops.sem_op  = (short int)(lock ? -1 : 1);

	do
	{
		z = semop(sem_id, &sops, 1);
	}
	while (z == -1 && errno == EINTR);

	if(z == -1)
	{
		if(!quiet)
		{
			(void)vrprint.error(-1, "Error", "%s: %s. Connection lost.", __FUNCTION__, strerror(errno));
		}

		return(0);
	}

	return(1);
}

void
tell_libvuurmuur_version(int *major, int *minor, int *sub, int *min_vuur_major, int *min_vuur_minor, int *min_vuur_sub)
{
	*major = LIBVUURMUUR_VERSION_MAJOR;
	*minor = LIBVUURMUUR_VERSION_MINOR;
	*sub = LIBVUURMUUR_VERSION_SUB;

	*min_vuur_major = MIN_VUURMUUR_VERSION_MAJOR;
	*min_vuur_minor = MIN_VUURMUUR_VERSION_MINOR;
	*min_vuur_sub = MIN_VUURMUUR_VERSION_SUB;

	return;
}

/*	range_strcpy

	src must be NULL-terminated.

	Returncodes:
		0: ok
		-1: error

*/
int
range_strcpy(char *dest, const char *src, const int start, const int end, size_t size)
{
	int	s,
		d,
		src_len;

	if(!dest || !src)
		return(-1);

	src_len = (int)strlen(src);

	if(start >= src_len || end > src_len)
		return(-1);

	for(s = start, d = 0; s < end && d < size && s < src_len; s++, d++)
	{
		dest[d] = src[s];
	}
	dest[d]='\0';

	return(0);
}



/* textfield regex

	General regular expression for filtering out problematic
	chars like '"' and %...
*/
//#define TEXTFIELD_REGEX		"^([a-zA-Z0-9_-][.] )$"


/*
	Actions:	 1: setup
			 0: setdown

	Returncodes:
		 	 0: ok
			-1: error

*/
int
setup_rgx(int action, struct rgx_ *reg)
{
	if(!reg)
		return(-1);

	if(action < 0 || action > 1)
		return(-1);

	if(action == 1)
	{
		/* regex setup */
		if(!(reg->zonename = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->zonename, ZONE_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}

		/* regex setup */
		if(!(reg->servicename = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->servicename, SERV_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}

		/* regex setup */
		if(!(reg->interfacename = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->interfacename, IFAC_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}

		/* regex setup */
		if(!(reg->macaddr = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->macaddr, MAC_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}

		/* regex setup */
		if(!(reg->configline = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->configline, CONFIG_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}

		/* regex setup */
/*		if(!(reg->comment = malloc(sizeof(regex_t))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(-1);
		}

		if(regcomp(reg->comment, TEXTFIELD_REGEX, REG_EXTENDED) != 0)
		{
			(void)vrprint.error(-1, "Internal Error", "setting up the regular expression with regcomp failed.");
			return(-1);
		}
*/
	}
	else
	{
		/* zone */
		regfree(reg->zonename);
		free(reg->zonename);

		/* service */
		regfree(reg->servicename);
		free(reg->servicename);

		/* interface */
		regfree(reg->interfacename);
		free(reg->interfacename);

		/* mac */
		regfree(reg->macaddr);
		free(reg->macaddr);

		/* config */
		regfree(reg->configline);
		free(reg->configline);

		/* comment */
//		regfree(reg->comment);
//		free(reg->comment);
	}

	return(0);
}
