/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "main.h"

/* iptables tables */
enum TB
{
	TB_ERROR,
	TB_FILTER,
	TB_MANGLE,
	TB_NAT,
	TB_TOOBIG,
};

/* iptables chains */
enum CH
{
	CH_ERROR,
	CH_PREROUTING,
	CH_INPUT,
	CH_FORWARD,
	CH_OUTPUT,
	CH_POSTROUTING,
	CH_BLOCKLIST,
	CH_BLOCKTARGET,
	CH_ANTISPOOF,
	CH_BADTCP,
	CH_SYNLIMITTARGET,
	CH_UDPLIMITTARGET,
	CH_TOOBIG,
};


static int
pipe_iptables_command(const int debuglvl, char *table, char *chain, char *cmd)
{
	char	str[MAX_PIPE_COMMAND] = "";

	/* safety */
	if(!cmd || !table || !chain)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/*
		assemble the command string
	*/
	if(snprintf(str, sizeof(str), "%s %s %s %s", conf.iptables_location, table, chain, cmd) >= sizeof(str))
	{
		(void)vrprint.error(-1, "Error", "iptables command creation overflow (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/*
		finally try to create the rule
	*/
	if(pipe_command(debuglvl, str, PIPE_VERBOSE) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


static int
process_rule(const int debuglvl, /*@null@*/RuleSet *ruleset, char table, char chain, char *cmd, unsigned long long packets, unsigned long long bytes)
{
	char	*chn = NULL,
		*tbl = NULL;

	/* safety */
	if(!cmd || table <= TB_ERROR || table >= TB_TOOBIG || chain <= CH_ERROR || chain >= CH_TOOBIG)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* the table */
	if(table == TB_FILTER)			tbl = "-t filter";
	else if(table == TB_MANGLE)		tbl = "-t mangle";
	else if(table == TB_NAT)		tbl = "-t nat";
	/* the chain */
	if(chain == CH_PREROUTING)		chn = "-A PREROUTING";
	else if(chain == CH_INPUT)		chn = "-A INPUT";
	else if(chain == CH_OUTPUT)		chn = "-A OUTPUT";
	else if(chain == CH_FORWARD)		chn = "-A FORWARD";
	else if(chain == CH_POSTROUTING)	chn = "-A POSTROUTING";
	else if(chain == CH_BLOCKLIST)		chn = "-A BLOCKLIST";
	else if(chain == CH_BLOCKTARGET)	chn = "-A BLOCK";
	else if(chain == CH_ANTISPOOF)		chn = "-A ANTISPOOF";
	else if(chain == CH_BADTCP)		chn = "-A BADTCP";
	else if(chain == CH_SYNLIMITTARGET)	chn = "-A SYNLIMIT";
	else if(chain == CH_UDPLIMITTARGET)	chn = "-A UDPLIMIT";

	if(!tbl || !chn)
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(!ruleset)
	{
		return(pipe_iptables_command(debuglvl, tbl, chn, cmd));
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "packets: %llu, bytes: %llu.", packets, bytes);

	if(table == TB_FILTER)
	{
		if(chain == CH_INPUT)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_input, chn, cmd, packets, bytes));
		else if(chain == CH_FORWARD)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_forward, chn, cmd, packets, bytes));
		else if(chain == CH_OUTPUT)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_output, chn, cmd, packets, bytes));

		else if(chain == CH_BLOCKTARGET)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_blocktarget, chn, cmd, packets, bytes));
		else if(chain == CH_BLOCKLIST)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_blocklist, chn, cmd, packets, bytes));
		else if(chain == CH_BADTCP)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_badtcp, chn, cmd, packets, bytes));
		else if(chain == CH_ANTISPOOF)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_antispoof, chn, cmd, packets, bytes));
		else if(chain == CH_SYNLIMITTARGET)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_synlimittarget, chn, cmd, packets, bytes));
		else if(chain == CH_UDPLIMITTARGET)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->filter_udplimittarget, chn, cmd, packets, bytes));
	}
	else if(table == TB_MANGLE)
	{
		if(chain == CH_PREROUTING)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->mangle_preroute, chn, cmd, packets, bytes));
		else if(chain == CH_INPUT)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->mangle_input, chn, cmd, packets, bytes));
		else if(chain == CH_FORWARD)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->mangle_forward, chn, cmd, packets, bytes));
		else if(chain == CH_OUTPUT)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->mangle_output, chn, cmd, packets, bytes));
		else if(chain == CH_POSTROUTING)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->mangle_postroute, chn, cmd, packets, bytes));
	}
	else if(table == TB_NAT)
	{
		if(chain == CH_PREROUTING)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->nat_preroute, chn, cmd, packets, bytes));
		else if(chain == CH_OUTPUT)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->nat_output, chn, cmd, packets, bytes));
		else if(chain == CH_POSTROUTING)
			return(ruleset_add_rule_to_set(debuglvl, &ruleset->nat_postroute, chn, cmd, packets, bytes));
	}

	/* default case, should never happen */
	return(-1);
}


/*	create_rule_input

	Creates a rule in the input chain.

	Returncodes:
		-1: error
		 0: ok
*/
int
create_rule_input(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	int	retval=0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "";
	char	reverse_input_device[sizeof(rule->from_int) + 3] = "";


	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* check caps */
	if(!bash_out)
	{
		if(!iptcap->target_queue && strcmp(rule->action, "QUEUE") == 0)
		{
			(void)vrprint.warning("Warning", "input rule not created: QUEUE not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_log && strncmp(rule->action, "LOG", 3) == 0)
		{
			(void)vrprint.warning("Warning", "input rule not created: LOG not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_reject && strncmp(rule->action, "REJECT", 6) == 0)
		{
			(void)vrprint.warning("Warning", "input rule not created: REJECT not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);

	/* we use rule->to_int here because it describes the interface of the firewall */
	snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s %s %s -m state --state NEW -j %s",
						input_device,
						rule->proto,
						rule->from_ip,
						rule->from_netmask,
						rule->temp_src_port,
						rule->to_ip,
						rule->to_netmask,
						rule->temp_dst_port,
						rule->from_mac,
						rule->limit,
						rule->action);

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		return(-1);

	create->iptcount.input++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* check cap */
		if(!bash_out)
		{
			if(iptcap->target_mark == 0)
			{
				(void)vrprint.warning("Warning", "mark rules not created: MARK not supported by this system.", __FUNCTION__, __LINE__);
				return(0); /* this is not an error */
			}
		}

		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_input_device, input_device, sizeof(reverse_input_device));
		reverse_input_device[1] = 'o';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
								input_device,
								rule->proto,
								rule->from_ip,
								rule->from_netmask,
								rule->temp_src_port,
								rule->to_ip,
								rule->to_netmask,
								rule->temp_dst_port,
								rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);

			/* related */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
								reverse_input_device,
								rule->proto,
								rule->to_ip,
								rule->to_netmask,
								temp_dst_port,
								rule->from_ip,
								rule->from_netmask,
								temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);
	
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* establised */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->temp_src_port,
								rule->to_ip,
								rule->to_netmask,
								rule->temp_dst_port,
								rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);

			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								temp_dst_port,
								rule->from_ip,
								rule->from_netmask,
								temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);

			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				/* RELATED */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);

				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);
									
				/* ESTABLISHED */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);

				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);
			}
		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* new, related, established */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							input_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);

			/* related, established */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);

			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);

				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
			
				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);
			}
		}
	}

	return(retval);
}


/*	create_rule_output

	Creates a rule in the output chain.

	Returncodes:
		-1: error
		 0: ok
*/
int
create_rule_output(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	output_device[sizeof(rule->to_int) + 3] = "";
	char	reverse_output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check caps */
	if(!bash_out)
	{
		if(!iptcap->target_queue && strcmp(rule->action, "QUEUE") == 0)
		{
			(void)vrprint.warning("Warning", "output rule not created: QUEUE not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_log && strncmp(rule->action, "LOG", 3) == 0)
		{
			(void)vrprint.warning("Warning", "output rule not created: LOG not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_reject && strncmp(rule->action, "REJECT", 6) == 0)
		{
			(void)vrprint.warning("Warning", "output rule not created: REJECT not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);

	snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW -j %s",
					output_device,
					rule->proto,
					rule->from_ip,
					rule->from_netmask,
					rule->temp_src_port,
					rule->to_ip,
					rule->to_netmask,
					rule->temp_dst_port,
					rule->limit,		/* log limit */
					rule->action);

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		return(-1);

	/* update rule counter */
	create->iptcount.output++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* check cap */
		if(!bash_out)
		{
			if(iptcap->target_mark == 0)
			{
				(void)vrprint.warning("Warning", "mark rules not created: MARK not supported by this system.", __FUNCTION__, __LINE__);
				return(0); /* this is not an error */
			}
		}

		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_output_device, output_device, sizeof(reverse_output_device));
		reverse_output_device[1] = 'i';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
							output_device,
							rule->proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! related */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
							reverse_output_device,
							rule->proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);
		
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
		
			/* established */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							reverse_output_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);

			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				/* RELATED */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);

				/* ESTABLISHED */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);

			}
		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
		
			/* new, related, established */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! related, established */
			snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							reverse_output_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
				return(-1);


			/* helperrrr */
			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				/* RELATED,ESTABLISHED */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_OUTPUT, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_INPUT, cmd, 0, 0) < 0)
					return(-1);
			}
		}
	}

	return(retval);
}


int
create_rule_forward(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "",
		output_device[sizeof(rule->to_int) + 3] = "";
	char	reverse_input_device[sizeof(rule->from_int) + 3] = "",
		reverse_output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check caps */
	if(!bash_out)
	{
		if(!iptcap->target_queue && strcmp(rule->action, "QUEUE") == 0)
		{
			(void)vrprint.warning("Warning", "forward rule not created: QUEUE not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_log && strncmp(rule->action, "LOG", 3) == 0)
		{
			(void)vrprint.warning("Warning", "forward rule not created: LOG not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
		else if(!iptcap->target_reject && strncmp(rule->action, "REJECT", 6) == 0)
		{
			(void)vrprint.warning("Warning", "forward rule not created: REJECT not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);


	/* create the rule */
	snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s %s %s -m state --state NEW -j %s",
					input_device,
					output_device,
					rule->proto,
					rule->from_ip,
					rule->from_netmask,
					rule->temp_src_port,
					rule->to_ip,
					rule->to_netmask,
					rule->temp_dst_port,
					rule->from_mac,
					rule->limit,
					rule->action);

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		return(-1);

	create->iptcount.forward++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* check cap */
		if(!bash_out)
		{
			if(iptcap->target_mark == 0)
			{
				(void)vrprint.warning("Warning", "mark rules not created: MARK not supported by this system.", __FUNCTION__, __LINE__);
				return(0);/* this is not an error */
			}
		}

		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_input_device, input_device, sizeof(reverse_input_device));
		reverse_input_device[1] = 'o';
		strlcpy(reverse_output_device, output_device, sizeof(reverse_output_device));
		reverse_output_device[1] = 'i';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related. from -> dst */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
							input_device,
							output_device,
							rule->proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! related. dst -> from  */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
							reverse_output_device,
							reverse_input_device,
							rule->proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

	
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* established. both ways */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							input_device,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							reverse_output_device,
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

		
			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				/* RELATED */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);


				/* ESTABLISHED */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);
			}
		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* new,related,established */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							input_device,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

			/* REVERSE! related,established */
			snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							reverse_output_device,
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);

			if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
				return(-1);

		
			if(strcmp(rule->helper, "") != 0)
			{
				/* check cap */
				if(!bash_out)
				{
					if(iptcap->match_helper == 0)
					{
						(void)vrprint.warning("Warning", "mark rules not created: helper-match not supported by this system.", __FUNCTION__, __LINE__);
						return(0); /* this is not an error */
					}
				}

				/* RELATED & ESTABLISHED */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);

				/* REVERSE! */
				snprintf(cmd, sizeof(cmd), "%s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);

				if(process_rule(debuglvl, ruleset, TB_MANGLE, CH_FORWARD, cmd, 0, 0) < 0)
					return(-1);
			}
		}
	}

	return(retval);
}


int
create_rule_masq(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check cap */
	if(!bash_out)
	{
		if(iptcap->target_masquerade == 0)
		{
			(void)vrprint.warning("Warning", "masquerade rules not created: MASQUERADE-target not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);

	/* assemble the string */
	snprintf(cmd, sizeof(cmd), "%s -s %s/%s -d %s/%s -j %s",
					output_device,
					rule->from_ip,
					rule->from_netmask,
					rule->to_ip,
					rule->to_netmask,
					rule->action);

	if(process_rule(debuglvl, ruleset, TB_NAT, CH_POSTROUTING, cmd, 0, 0) < 0)
		return(-1);

	/* update the chain counter */
	create->iptcount.postroute++;

	return(0);
}


/*
	TODO: maybe we want an option to use only one interface.
*/
int
create_rule_snat(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	char	cmd[MAX_PIPE_COMMAND];
	char	output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check cap */
	if(!bash_out)
	{
		if(iptcap->target_snat == 0)
		{
			(void)vrprint.warning("Warning", "snat rules not created: SNAT-target not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);
	
	/* assemble SNAT string - we do this here because LOG can't handle --to-source */
	if(strcmp(rule->action, "SNAT") == 0)
	{
		snprintf(rule->action, sizeof(rule->action), "SNAT --to-source %s", rule->serverip);
	}

	/* assemble the string */
	snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/%s %s -j %s",
					output_device,
					rule->proto,
					rule->from_ip,
					rule->from_netmask,
					rule->temp_src_port,
					rule->to_ip,
					rule->to_netmask,
					rule->temp_dst_port,
					rule->action);

	if(process_rule(debuglvl, ruleset, TB_NAT, CH_POSTROUTING, cmd, 0, 0) < 0)
		return(-1);

	/* update the counter */
	create->iptcount.postroute++;

	return(0);
}


/*	create_rule_portfw

	Here we create the iptablesrules for portfw. They concist of a PREROUTING rule and a FORWARD rule.
	Both are created for PORTFW.

	For PORTFW we handle both listenport and remoteport options.
*/
int
create_rule_portfw(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check cap */
	if(!bash_out)
	{
		if(iptcap->target_dnat == 0)
		{
			(void)vrprint.warning("Warning", "portfw rules not created: DNAT-target not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* see pp 278 linux firewall 2nd edition for portforwarding && 274-275 for redirect */
	
	/* assembeling rule->action, we start with rule->remoteip */

	/* for remote port use, thats easy we want to use the given remoteport as --to-destination ports */
	if(create->option.remoteport == 1 && rule->remoteport_ptr != NULL)
	{
		if(rule->remoteport_ptr->dst_high == -1)
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d", create->to->ipv4.ipaddress, rule->remoteport_ptr->dst_low);
		else
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d-%d", create->to->ipv4.ipaddress, rule->remoteport_ptr->dst_low, rule->remoteport_ptr->dst_high);
	}
	/* if we use listenport, we want --to-destination to be the original port(s) of the service. */
	else if(create->option.listenport == 1 && rule->portrange_ptr != NULL)
	{
		if(rule->portrange_ptr->dst_high <= 0)
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d", create->to->ipv4.ipaddress, rule->portrange_ptr->dst_low);
		else
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d-%d", create->to->ipv4.ipaddress, rule->portrange_ptr->dst_low, rule->portrange_ptr->dst_high);
	}
	/* if no listenport or remoteport --to-destination is just the ip. */
	else
	{
		snprintf(rule->remoteip, sizeof(rule->remoteip), "%s", create->to->ipv4.ipaddress);
	}

	/* we set this here because we need remoteip */
	if(strncmp(rule->action, "DNAT", 4) == 0)
	{
		snprintf(rule->action, sizeof(rule->action), "DNAT --to-destination %s", rule->remoteip);
	}

	/* set --dport here, but only if we need to change it. */
	if(create->option.listenport == 1 && rule->listenport_ptr != NULL)
	{
		if(rule->listenport_ptr->dst_high == -1)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->listenport_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->listenport_ptr->dst_low, rule->listenport_ptr->dst_high);
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);
	
	/* here we pipe the rule, but only if its not a log rule, because we only log the forward rule for portfw */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
	{
		snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/255.255.255.255 %s %s -m state --state NEW -j %s",
						input_device,
						rule->proto,
						rule->from_ip,
						rule->from_netmask,
						rule->temp_src_port,
						rule->serverip,
						rule->temp_dst_port,
						rule->from_mac,
						rule->action);

		if(process_rule(debuglvl, ruleset, TB_NAT, CH_PREROUTING, cmd, 0, 0) < 0)
			return(-1);

		create->iptcount.preroute++;
	}

	/* the forward rule */

	/* if we use remoteport, it will be our destination */
	if (create->option.remoteport == 1 && rule->remoteport_ptr != NULL)
	{
		if(rule->remoteport_ptr->dst_high <= 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->remoteport_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->remoteport_ptr->dst_low, rule->remoteport_ptr->dst_high);
	}

	/*	if we have a listenport option temp_dst_port for the DNAT rule is different from the FORWARD rule, so we fix that here
		we only do this if remoteport == 0, otherwise we use the remoteport
	*/
	if(create->option.listenport == 1 && create->option.remoteport == 0 && rule->portrange_ptr != NULL)
	{
		if(rule->portrange_ptr->dst_high <= 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->portrange_ptr->dst_low, rule->portrange_ptr->dst_high);
	}

	/* set rule->chain */
	strlcpy(create->chain, "FORWARD", sizeof(create->chain));

	/* set the action */
	if(strncmp(rule->action, "DNAT", 4) == 0)
	{
		if(!create->option.queue)
			snprintf(rule->action, sizeof(rule->action), "ACCEPT");
		else
			snprintf(rule->action, sizeof(rule->action), "QUEUE");
	}

	if(create_rule_forward(debuglvl, ruleset, rule, create, iptcap) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating forward rule for portfw failed (in: %s).", __FUNCTION__);
		retval = -1;
	}

	/* reset rule->chain */
	strlcpy(create->chain, "PREROUTING", sizeof(create->chain));
	
	return(retval);
}


/*	create_rule_redirect

	see pp 278 linux firewall 2nd edition 274-275 for redirect
*/
int
create_rule_redirect(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	tmp_chain[sizeof(create->chain)] = "",
		tmp_port[sizeof(rule->temp_dst_port)] = "",
		tmp_action[sizeof(rule->action)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "";

	/* safety */
	if(!rule || !create || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check cap */
	if(!bash_out)
	{
		if(iptcap->target_redirect == 0)
		{
			(void)vrprint.warning("Warning", "redirect rules not created: REDIRECT-target not supported by this system.", __FUNCTION__, __LINE__);
			return(0); /* this is not an error */
		}
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);

	/* here we pipe the rule, but only if its not a log rule, because we only log the forward rule for portfw */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
	{
		snprintf(cmd, sizeof(cmd), "%s %s -s %s/%s %s -d %s/255.255.255.255 %s %s -m state --state NEW -j %s",
						input_device,
						rule->proto,
						rule->from_ip,
						rule->from_netmask,
						rule->temp_src_port,
						rule->to_ip,
						rule->temp_dst_port,
						rule->from_mac,
						rule->action);

		if(process_rule(debuglvl, ruleset, TB_NAT, CH_PREROUTING, cmd, 0, 0) < 0)
			return(-1);

		create->iptcount.preroute++;
	}
	
	/*
		create the corresponding input rule
	*/
	
	/* temp store the realports */
	strlcpy(tmp_port, rule->temp_dst_port, sizeof(tmp_port));
	/* temp store the chain */
	strlcpy(tmp_chain, create->chain, sizeof(tmp_chain));
	/* temp store the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
		strlcpy(tmp_action, rule->action, sizeof(tmp_action));

	/* set the redirectport to rule->temp_dst_port */
	snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", create->option.redirectport);
	/* set the chain */
	snprintf(create->chain, sizeof(create->chain), "INPUT");
	/* set the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
	{
		if(!create->option.queue)
			snprintf(rule->action, sizeof(rule->action), "ACCEPT");
		else
			snprintf(rule->action, sizeof(rule->action), "QUEUE");
	}
		
	/* now create the input rule */
	if(create_rule_input(debuglvl, ruleset, rule, create, iptcap) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating input rule for redirect failed (in: %s).", __FUNCTION__);
		return(-1);
	}
		
	/* restore the realports */
	strlcpy(rule->temp_dst_port, tmp_port, sizeof(rule->temp_dst_port));
	/* restore the chain */
	strlcpy(create->chain, tmp_chain, sizeof(create->chain));
	/* restore the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
		strlcpy(rule->action, tmp_action, sizeof(rule->action));

	return(0);
}


/* pre_rules

	Cleanup
		chains
	Sets up
		default policies
		connection tracking
		portscan detection
*/
int
pre_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, Interfaces *interfaces, IptCap *iptcap)
{
	int			retval = 0,
				result = 0;
	char			cmd[MAX_PIPE_COMMAND] = "";
	d_list_node		*d_node = NULL;
	struct InterfaceData_ 	*iface_ptr = NULL;
	char			limit[] = "-m limit --limit 1/s --limit-burst 2";


	/* safety */
	if(!interfaces || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* check cap */
	if(!bash_out)
	{
		if(iptcap->target_log == 0)
		{
			(void)vrprint.warning("Warning", "not creating logrules. LOG-target not supported by system.");
		}
		else
		{
			if(!iptcap->match_limit)
			{
				(void)vrprint.warning("Warning", "not setting limits on logrules. Limit-match not supported by system.");
				memset(limit, 0, sizeof(limit));
			}
		}
	}

	/*
		first flush the chains
	*/
	if(!ruleset)
	{
		if(bash_out == 1) fprintf(stdout, "\n# Flushing chains...\n");
		if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Flushing chains...");
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval = -1;
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat --flush", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval = -1;
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle --flush", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval = -1;
	}


	/*
		allow local loopback
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Allowing local loopback...\n");
	if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Allowing local loopback...");

	snprintf(cmd, sizeof(cmd), "-i lo -j ACCEPT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval = -1;
	ipt_input_count++;

	snprintf(cmd, sizeof(cmd), "-o lo -j ACCEPT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;


	/*
		create an empty rule in INPUT, OUTPUT and FORWARD, which count
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Creating interface counters...\n");
	if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Creating interface counters...");

	for(d_node = interfaces->list.top; d_node; d_node = d_node->next)
	{
		iface_ptr = d_node->data;

		/* if the interface active? */
		if(iface_ptr->active)
		{
			/* does the interface have an ipaddress? */
			if(strcmp(iface_ptr->ipv4.ipaddress, "") != 0)
			{
				/* Check for empty device string and virtual device. */
				if(strcmp(iface_ptr->device, "") != 0 && !iface_ptr->device_virtual)
				{
					/*
						first in the input chain
					*/
					snprintf(cmd, sizeof(cmd), "-i %s", iface_ptr->device);
					if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, iface_ptr->cnt ? iface_ptr->cnt->input_packets : 0, iface_ptr->cnt ? iface_ptr->cnt->input_bytes : 0) < 0)
						retval=-1;
					else
						ipt_input_count++;

					/*
						then in the output chain
					*/
					snprintf(cmd, sizeof(cmd), "-o %s", iface_ptr->device);
					if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, iface_ptr->cnt ? iface_ptr->cnt->output_packets : 0, iface_ptr->cnt ? iface_ptr->cnt->output_bytes : 0) < 0)
						retval=-1;
					else
						ipt_output_count++;

					/*
						then in the forward chain, in
					*/
					snprintf(cmd, sizeof(cmd), "-i %s", iface_ptr->device);
					if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, iface_ptr->cnt ? iface_ptr->cnt->forwardin_packets : 0, iface_ptr->cnt ? iface_ptr->cnt->forwardin_bytes : 0) < 0)
						retval=-1;
					else
						ipt_forward_count++;

					/*
						then in the forward chain, out
					*/
					snprintf(cmd, sizeof(cmd), "-o %s", iface_ptr->device);
					if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, iface_ptr->cnt ? iface_ptr->cnt->forwardout_packets : 0, iface_ptr->cnt ? iface_ptr->cnt->forwardout_bytes : 0) < 0)
						retval=-1;
					else
						ipt_forward_count++;
				}
			}
		}
	}


	if(!ruleset)
	{
		/*
			set default policies to DROP
		*/
		if(bash_out == 1) fprintf(stdout, "\n# Setting default policies...\n");
		if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Setting default policies...");
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy INPUT DROP", conf.iptables_location);
		result=pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval=-1;
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy OUTPUT DROP", conf.iptables_location);
		result=pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval=-1;
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy FORWARD DROP", conf.iptables_location);
		result=pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0) retval=-1;
	}
	else
	{
		ruleset->filter_input_policy = 1;	/* drop */
		ruleset->filter_output_policy = 1;	/* drop */
		ruleset->filter_forward_policy = 1;	/* drop */
	}


	/*
		stealthscan protection
	*/
	if(bash_out==1) fprintf(stdout, "\n# Setting up stealth scan protection...\n");
	if((debuglvl >= LOW) || (bash_out == 1)) (void)vrprint.debug(__FUNCTION__, "Setting up stealth scan protection...");

	/* ALL NONE */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ALL NONE %s -j LOG --log-prefix \"vrmr: DROP probe ALL \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ALL NONE -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* SYN - FIN */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags SYN,FIN SYN,FIN %s -j LOG --log-prefix \"vrmr: DROP probe SYN-FIN \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags SYN,FIN SYN,FIN -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* SYN - RST */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags SYN,RST SYN,RST %s -j LOG --log-prefix \"vrmr: DROP probe SYN-RST \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags SYN,RST SYN,RST -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* FIN - RST */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags FIN,RST FIN,RST %s -j LOG --log-prefix \"vrmr: DROP probe FIN-RST \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags FIN,RST FIN,RST -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* ACK - FIN */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,FIN FIN %s -j LOG --log-prefix \"vrmr: DROP probe FIN \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,FIN FIN -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* ACK - PSH */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,PSH PSH %s -j LOG --log-prefix \"vrmr: DROP probe PSH \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,PSH PSH -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* ACK - URG */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,URG URG %s -j LOG --log-prefix \"vrmr: DROP probe URG \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp --tcp-flags ACK,URG URG -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/* New tcp but no SYN */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-p tcp ! --syn -m state --state NEW %s -j LOG --log-prefix \"vrmr: DROP no SYN \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-p tcp ! --syn -m state --state NEW -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/*
		Fragmented packets
	*/
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-f %s -j LOG --log-prefix \"vrmr: DROP FRAG \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-f -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/*
		set up connectiontracking for QUEUE (mark 3)
	*/
	if((iptcap->target_queue && iptcap->match_mark) || bash_out)
	{
		if(bash_out == 1) fprintf(stdout, "\n# Setting up connection-tracking for QUEUE targets...\n");
		if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Setting up connection-tracking for QUEUE targets...");

		snprintf(cmd, sizeof(cmd), "-m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;

		snprintf(cmd, sizeof(cmd), "-m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_output_count++;

		snprintf(cmd, sizeof(cmd), "-m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
			retval=-1;
		ipt_forward_count++;
	}
	else
	{
		(void)vrprint.info("Info", "connection tracking for QUEUE not setup. QUEUE-target and/or mark-match not supported by system.");
	}


	/*
		set up connectiontracking for QUEUE with IptState for Snort_inline.
	*/
	if((iptcap->target_queue && iptcap->match_mark) || bash_out)
	{
		if(bash_out == 1) fprintf(stdout, "\n# Setting up connection-tracking for QUEUE with IptState for Snort_inline.\n");
		if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Setting up connection-tracking for QUEUE with IptState for Snort_inline..");

		/* input */
		snprintf(cmd, sizeof(cmd), "-m mark --mark 1 -m state --state RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;

		snprintf(cmd, sizeof(cmd), "-m mark --mark 2 -m state --state ESTABLISHED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;

		/* output */
		snprintf(cmd, sizeof(cmd), "-m mark --mark 1 -m state --state RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_output_count++;

		snprintf(cmd, sizeof(cmd), "-m mark --mark 2 -m state --state ESTABLISHED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_output_count++;

		/* forward */
		snprintf(cmd, sizeof(cmd), "-m mark --mark 1 -m state --state RELATED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
			retval=-1;
		ipt_forward_count++;

		snprintf(cmd, sizeof(cmd), "-m mark --mark 2 -m state --state ESTABLISHED -j QUEUE");
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
			retval=-1;
		ipt_forward_count++;
	}
	else
	{
		(void)vrprint.warning("Info", "connection tracking for QUEUE (Snort-Inline) not setup. QUEUE-target and/or mark-match not supported by system.");
	}


	/*
		set up connectiontracking
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Setting up connection-tracking...\n");
	if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Setting up connection-tracking...");

	snprintf(cmd, sizeof(cmd), "-m state --state ESTABLISHED,RELATED -j ACCEPT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	snprintf(cmd, sizeof(cmd), "-m state --state ESTABLISHED,RELATED -j ACCEPT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;

	snprintf(cmd, sizeof(cmd), "-m state --state ESTABLISHED,RELATED -j ACCEPT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;


	/*
		invalid input
	*/
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-m state --state INVALID %s -j LOG --log-prefix \"vrmr: DROP in INVALID \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_input_count++;
	}

	snprintf(cmd, sizeof(cmd), "-m state --state INVALID -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	/*
		invalid output
	*/
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-m state --state INVALID %s -j LOG --log-prefix \"vrmr: DROP out INVALID \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
			retval=-1;
		ipt_output_count++;
	}

	snprintf(cmd, sizeof(cmd), "-m state --state INVALID -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;

	/*
		invalid forward
	*/
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-m state --state INVALID %s -j LOG --log-prefix \"vrmr: DROP fw INVALID \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
			retval=-1;
		ipt_forward_count++;
	}

	snprintf(cmd, sizeof(cmd), "-m state --state INVALID -j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;


	/*
		Setup Block lists
	*/
	if(bash_out == 1)
		fprintf(stdout, "\n# Setting up blocklist...\n");
	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "Setting up blocklist...");

	if(!ruleset)
	{
		/* create the chain and insert it into input, output and forward.
	
			NOTE: we ignore the returncode and want no output (although we get some
			in the errorlog) because if we start vuurmuur when a ruleset is already in
			place, the chain will exist and iptables will complain.
		*/
		snprintf(cmd, sizeof(cmd), "%s -N BLOCKLIST 2>/dev/null", conf.iptables_location);
		(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	}

	snprintf(cmd, sizeof(cmd), "-j BLOCKLIST");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;

	if(!ruleset)
	{
		/* create the BLOCK action
	
	   	NOTE: see BLOCKLIST creation. */
		snprintf(cmd, sizeof(cmd), "%s -N BLOCK 2>/dev/null", conf.iptables_location);
		(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	}

	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "%s -j LOG --log-prefix \"vrmr: DROP BLOCKED \" %s",
						limit,
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_BLOCKTARGET, cmd, 0, 0) < 0)
			retval=-1;
	}

	snprintf(cmd, sizeof(cmd), "-j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_BLOCKTARGET, cmd, 0, 0) < 0)
		retval=-1;


	/*
		syn-flooding protection
	*/
	if(!ruleset)
	{
		snprintf(cmd, sizeof(cmd), "%s -N SYNLIMIT 2>/dev/null", conf.iptables_location);
		(void)pipe_command(debuglvl, cmd, 1);
	}

	/* create the rules */
	if(update_synlimit_rules(debuglvl, ruleset, iptcap) < 0)
		retval = -1;

	/* create the chain and insert it into input, output and forward.
	
	   NOTE: we ignore the returncode and want no output (although we get some
	   in the errorlog) because if we start vuurmuur when a ruleset is already in
	   place, the chain will exist and iptables will complain. */
	snprintf(cmd, sizeof(cmd), "-p tcp --syn -j SYNLIMIT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;


	/*
		udp-flooding protection
	*/
	if(!ruleset)
	{
		snprintf(cmd, sizeof(cmd), "%s -N UDPLIMIT 2>/dev/null", conf.iptables_location);
		(void)pipe_command(debuglvl, cmd, 1);
	}

	/* create the rules */
	if(update_udplimit_rules(debuglvl, ruleset, iptcap) < 0)
		retval = -1;

	/* create the chain and insert it into input, output and forward.
	
	   NOTE: we ignore the returncode and want no output (although we get some
	   in the errorlog) because if we start vuurmuur when a ruleset is already in
	   place, the chain will exist and iptables will complain. */
	snprintf(cmd, sizeof(cmd), "-p udp -m state --state NEW -j UDPLIMIT");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_output_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;


	if(!ruleset)
	{
		/* anti spoof rules */
		if(bash_out == 1)
			fprintf(stdout, "\n# Setting up anti-spoofing rules...\n");

		/* create the chain and insert it into input, output and forward.
	
			NOTE: we ignore the returncode and want no output (although we get some
			in the errorlog) because if we start vuurmuur when a ruleset is already in
			place, the chain will exist and iptables will complain. */
		snprintf(cmd, sizeof(cmd), "%s -N ANTISPOOF 2>/dev/null", conf.iptables_location);
		(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	}

	snprintf(cmd, sizeof(cmd), "-j ANTISPOOF");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
		retval=-1;
	ipt_input_count++;

	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
		retval=-1;
	ipt_forward_count++;

	return(retval);
}


/*	update_synlimit_rules

	Creates/updates the the rules in the SYNLIMIT chain.
	
	Note: if the limit-match is not supported, bail out of here.
*/
int
update_synlimit_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, IptCap *iptcap)
{
	int			retval = 0,
				result = 0;
	char			cmd[MAX_PIPE_COMMAND] = "";

	/* caps */
	if(!bash_out)
	{
		if(iptcap->match_limit == 0)
		{
			(void)vrprint.warning("Warning", "synlimit rules not setup. Limit-match not supported by system.");
			return(0); /* no error */
		}
	}

	if(conf.syn_limit == 0 || conf.syn_limit_burst == 0)
	{
		(void)vrprint.error(-1, "Error", "limit of 0 cannot be used (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* flush the chain if we are not in ruleset mode */
	if(!ruleset)
	{
		/* first flush the chain */
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush SYNLIMIT", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0)
			retval = -1;
	}

	/* create the return rule */
	snprintf(cmd, sizeof(cmd), "-m limit --limit %u/s --limit-burst %u -j RETURN", conf.syn_limit, conf.syn_limit_burst);
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_SYNLIMITTARGET, cmd, 0, 0) < 0)
		retval=-1;

	/* the log rule */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix \"vrmr: DROP SYNLIMIT reach.\" %s", 
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_SYNLIMITTARGET, cmd, 0, 0) < 0)
			retval=-1;
	}

	/* and finally the drop rule */
	snprintf(cmd, sizeof(cmd), "-j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_SYNLIMITTARGET, cmd, 0, 0) < 0)
		retval=-1;

	return(retval);
}


/*	update_udplimit_rules

	Creates/updates the the rules in the UDPLIMIT chain.
	
	Note: if the limit-match is not supported, bail out of here.
*/
int
update_udplimit_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, IptCap *iptcap)
{
	int			retval = 0,
				result = 0;
	char			cmd[MAX_PIPE_COMMAND] = "";

	/* caps */
	if(!bash_out)
	{
		if(iptcap->match_limit == 0)
		{
			(void)vrprint.warning("Warning", "udplimit rules not setup. Limit-match not supported by system.");
			return(0); /* no error */
		}
	}

	if(conf.udp_limit == 0 || conf.udp_limit_burst == 0)
	{
		(void)vrprint.error(-1, "Error", "limit of 0 cannot be used (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* flush the chain if we are not in ruleset mode */
	if(!ruleset)
	{
		/* first flush the chain */
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush UDPLIMIT", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
		if(result < 0)
			retval = -1;
	}

	/* create the return rule */
	snprintf(cmd, sizeof(cmd), "-m limit --limit %u/s --limit-burst %u -j RETURN", conf.udp_limit, conf.udp_limit_burst);
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_UDPLIMITTARGET, cmd, 0, 0) < 0)
		retval=-1;

	/* the log rule */
	if(iptcap->target_log || bash_out)
	{
		snprintf(cmd, sizeof(cmd), "-m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix \"vrmr: DROP UDPLIMIT reach.\" %s",
						loglevel);

		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_UDPLIMITTARGET, cmd, 0, 0) < 0)
			retval=-1;
	}

	/* and finally the drop rule */
	snprintf(cmd, sizeof(cmd), "-j DROP");
	if(process_rule(debuglvl, ruleset, TB_FILTER, CH_UDPLIMITTARGET, cmd, 0, 0) < 0)
		retval=-1;

	return(retval);
}


/*	post_rules

	Enables logging in the INPUT, OUTPUT and FORWARD chains.
	
	If forward_rules == 1, then /proc/sys/net/ipv4/ip_forward will be enabled. Otherwise it will be disabled.

	Returncode:
		 0: ok
		-1: error
*/
int
post_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, IptCap *iptcap, int forward_rules)
{
	int	retval=0,
		result=0;
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	limit[42] = "";


	/* safety */
	if(!iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* do we want to log the default policy? */
	if(conf.log_policy)
	{
		/* cap */
		if(!bash_out)
		{
			if(iptcap->target_log == 0)
			{
				(void)vrprint.warning("Warning", "not creating logrules. LOG-target not supported by system.");
				return(0); /* no error */
			}
		}

		/* see if we want to limit the logging of the default policy */
		if(conf.log_policy_limit > 0)
		{
			/* cap */
			if(iptcap->match_limit || bash_out)
			{
				snprintf(limit, sizeof(limit), "-m limit --limit %d/s --limit-burst %d",
										conf.log_policy_limit,
										conf.log_policy_burst);
			}
			else
			{
				(void)vrprint.warning("Warning", "not setting limits on logrules. Limit-match not supported by system.");
				return(0); /* no error */
			}
		}

		/* enable logging for all packets which don't have a rule */
		if(bash_out == 1) fprintf(stdout, "\n# Enabling logging...\n");
		if(debuglvl >= LOW) (void)vrprint.debug(__FUNCTION__, "Enabling logging...");

		/* input */
		snprintf(cmd, sizeof(cmd), "%s -j LOG --log-prefix \"vrmr: DROP in policy \" %s",
						limit,
						loglevel);
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_INPUT, cmd, 0, 0) < 0)
			retval=-1;

		/* output */
		snprintf(cmd, sizeof(cmd), "%s -j LOG --log-prefix \"vrmr: DROP out policy \" %s",
						limit,
						loglevel);
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_OUTPUT, cmd, 0, 0) < 0)
			retval=-1;

		/* forward */
		snprintf(cmd, sizeof(cmd), "%s -j LOG --log-prefix \"vrmr: DROP fw policy \" %s",
						limit,
						loglevel);
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_FORWARD, cmd, 0, 0) < 0)
			retval=-1;
	}

	/* enable or disable ip-forwarding */
	if(forward_rules)
	{
		(void)vrprint.info("Info", "Created forward-rules: enabling ip-forwarding.");

		result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/ip_forward", 1, NULL);
		if(result != 0)
		{
			/* if it fails, we dont really care, its not fatal */
			(void)vrprint.error(-1, "Error", "enabling ip-forwarding failed.");
		}
	}
	else
	{
		(void)vrprint.info("Info", "Created no forward-rules: disabling ip-forwarding.");

		result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/ip_forward", 0, NULL);
		if(result != 0)
		{
			/* if it fails, we dont really care, its not fatal */
			(void)vrprint.error(-1, "Error", "enabling ip-forwarding failed.");
		}
	}

	return(retval);
}


/*
	we don't care if the network is active or not
*/
int
create_network_protect_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, Zones *zones, IptCap *iptcap)
{
	struct RuleCache_	*create = NULL;
	d_list_node		*d_node = NULL,
				*net_d_node = NULL,
				*from_if_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	struct RuleData_	*rule_ptr = NULL;
	int			retval = 0;
	struct InterfaceData_	*from_if_ptr = NULL;
	char			cmd[MAX_PIPE_COMMAND] = "";
	char			input_device[16 + 3] = "";	/* 16 + '-i ' */
	char			limit[] = "-m limit --limit 1/s --limit-burst 5";


	/* safety */
	if(!zones || !iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* check capability */
	if(!bash_out)
	{
		if(!iptcap->match_limit)
		{
			(void)vrprint.warning("Warning", "not setting limits on logrules. Limit-match not supported by system.");
			memset(limit, 0, sizeof(limit));
		}
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading anti-spoofing rules...\n");

	/* loop through the zones to look for networks */
	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		if(zone_ptr->type == TYPE_NETWORK)
		{
			/* now loop through the ruleslist */
			for(net_d_node = zone_ptr->ProtectList.top; net_d_node; net_d_node = net_d_node->next)
			{
				if(!(rule_ptr = net_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
				if(!(create = &rule_ptr->rulecache))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
				if(!create->who)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				/* here we print the description if we are in bashmode */
				if(bash_out == 1 && create->description != NULL)
				{
					fprintf(stdout, "\n# %s\n", create->description);
		
					free(create->description);
					create->description = NULL;
				}

				/* iptables protect */
				if(create->danger.solution == PROT_IPTABLES)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug(__FUNCTION__, "protect iptables.");

					/* check if all is filled in right */
					if(	strcmp(create->danger.source_ip.ipaddress, "") != 0 &&
						strcmp(create->danger.source_ip.netmask, "") != 0 &&
						strcmp(create->danger.type, "") != 0 &&
						strcmp(create->danger.source, "") != 0)
					{
						/* if the network has no interfaces, don't bother */
						if(create->who->InterfaceList.len > 0)
						{
							for(from_if_node = create->who->InterfaceList.top; from_if_node; from_if_node = from_if_node->next)
							{
								if(!(from_if_ptr = from_if_node->data))
								{
									(void)vrprint.error(-1, "Internal Error", "from_if_ptr == NULL! (in: create_rule, PROT_IPTABLES).");
									return(-1);
								}

								/*	see if the interface is active */
								if(from_if_ptr->active == 0)
								{
									/* here we print the description if we are in bashmode */
									if(bash_out == 1)
									{
										fprintf(stdout, "# anti-spoof rule for interface '%s' not created. The interface is inactive.\n", from_if_ptr->name);
									}
								}
								else
								{
									/*	get the input_device

										if the device is virtual (oldstyle), we don't add it to the iptables string.
									*/
									if(!from_if_ptr->device_virtual_oldstyle)
										snprintf(input_device, sizeof(input_device), "-i %s", from_if_ptr->device);
									else
										memset(input_device, 0, sizeof(input_device));


									/*	only create the rule if the interface is active

										NOTE: this is not the same as an interface that is down!
									*/
									if(from_if_ptr->active)
									{
										/* create the log rule */
										if(iptcap->target_log || bash_out)
										{
											snprintf(cmd, sizeof(cmd), "%s -s %s/%s -d %s/255.255.255.255 %s -j LOG --log-prefix \"vrmr: DROP %s %s \" %s",
															input_device,
															create->danger.source_ip.ipaddress,
															create->danger.source_ip.netmask,
															from_if_ptr->ipv4.ipaddress,
															limit,
															create->danger.type,
															create->danger.source,
															loglevel);

											if(process_rule(debuglvl, ruleset, TB_FILTER, CH_ANTISPOOF, cmd, 0, 0) < 0)
												retval=-1;
										}

										/* and the DROP rule */
										snprintf(cmd, sizeof(cmd), "%s -s %s/%s -d %s/255.255.255.255 -j DROP",
														input_device,
														create->danger.source_ip.ipaddress,
														create->danger.source_ip.netmask,
														from_if_ptr->ipv4.ipaddress);

										if(process_rule(debuglvl, ruleset, TB_FILTER, CH_ANTISPOOF, cmd, 0, 0) < 0)
											retval=-1;
									}
								}
							}
						}
					}
				}
				/* Whoops, this is serious. */
				else
				{
					(void)vrprint.error(-1, "Internal Error", "unknown protect danger type %d (in: %s).", create->danger.solution, __FUNCTION__);
					return(-1);
				}
			}
		}
	}

	return(0);
}


int
create_block_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, d_list *blocklist)
{
	char		cmd[MAX_PIPE_COMMAND] = "",
			*ipaddress = NULL;
	d_list_node	*d_node = NULL;
	int		retval = 0;


	/* safety */
	if(!blocklist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading Blocklist...\n");

	if(blocklist->len == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "no items in blocklist.");

		return(0);
	}

	/* create two rules for each ipaddress */
	for(d_node = blocklist->top; d_node; d_node = d_node->next)
	{
		if(!(ipaddress = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "ipaddress to add: '%s'.", ipaddress);

		/* ip is source */
		snprintf(cmd, sizeof(cmd), "-s %s/255.255.255.255 -j BLOCK", ipaddress);
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_BLOCKLIST, cmd, 0, 0) < 0)
			retval=-1;

		/* ip is dst */
		snprintf(cmd, sizeof(cmd), "-d %s/255.255.255.255 -j BLOCK", ipaddress);
		if(process_rule(debuglvl, ruleset, TB_FILTER, CH_BLOCKLIST, cmd, 0, 0) < 0)
			retval=-1;
	}

	return(retval);
}
