/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __MAIN_H__
#define __MAIN_H__

/*****************************************************************************************************************\
 ********************************* INCLUDES **********************************************************************
\*****************************************************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>     /* for catching signals */
#include <time.h>       /* included for logging */
#include <errno.h>      /* error handling */
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/wait.h>

/* our own vuurmuurlib */
#include <vuurmuur.h>

/*****************************************************************************************************************\
 ********************************* DEFINES ***********************************************************************
\*****************************************************************************************************************/

#define VUURMUUR_VERSION_MAJOR		0
#define VUURMUUR_VERSION_MINOR		5
#define VUURMUUR_VERSION_SUB		58

#define MIN_LIBVUURMUUR_VERSION_MAJOR	0
#define MIN_LIBVUURMUUR_VERSION_MINOR	5
#define MIN_LIBVUURMUUR_VERSION_SUB	57

#define LOGPREFIX_PREFIX		"vrmr:"		/* the prefix in the logprefix */
#define LOGPREFIX_LOG_MAXLEN		29
#define LOGPREFIX_ULOG_MAXLEN		32		/* not yet used */

#define LOOP_INT			1

#define YES				1
#define NO				0

#define PIDFILE				"/var/run/vuurmuur.pid"

#define LOCK(x)				LockSHM(1, x, 0)
#define UNLOCK(x)			LockSHM(0, x, 0)

/*************************************************************************************************************************\
 ******************************************************* DATATYPES *******************************************************
\*************************************************************************************************************************/

char	loglevel[32];	/* --log-level warning */

struct RuleCreateData_
{
	char		action[sizeof(RuleCache.action)];
	char		chain[48]; /* why 48? */

	/* ipaddress */
	char		from_ip[16];
	char		to_ip[16];

	/* netmasks */
	char		from_netmask[16];
	char		to_netmask[16];

	/* mac - only from */
	char		from_mac[48]; /* --m mac --from-mac mac */

	struct ipdata	ipv4_from;
	struct ipdata	ipv4_to;

	/* interfaces */
	char		from_int[16];
	char		to_int[16];

	/* proto */
	char		proto[16+6]; // why 16+6? <- the 6 is for ' --syn' for tcp
	char		helper[32];

	/* ports */
	char		temp_dst_port[32];
	char		temp_src_port[32];

	struct portdata *portrange_ptr;
	struct portdata *listenport_ptr;
	struct portdata *remoteport_ptr;

	char		limit[42]; /*  -m limit --limit 999/s --limit-burst 9999 */

	/* loop stuff - needs a cleanup and needs it BAD */
	char		logged;
	char		normal;
	char		broadcasted;
	char		broadlog;

	/* portfw stuff - needs to go-> later we can put it in the function for creating portfw rules! */
	char		serverip[MAX_NET_ZONE];
	char		remoteip[32]; // max 15 (for ip) + 1 (for :) + 5 (for port) = 21? + 1 = 22
	char		temp_port_store[6];
};


/*	here we are going to assemble all rules for
	the creation of the file for iptables-restore.

*/
typedef struct
{
	/* special rules */
	//char	

	/*
		mangle
	*/
	d_list		mangle_preroute;		/* list with rules */
	char		mangle_preroute_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		mangle_input;			/* list with rules */
	char		mangle_input_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		mangle_forward;			/* list with rules */
	char		mangle_forward_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		mangle_output;			/* list with rules */
	char		mangle_output_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		mangle_postroute;		/* list with rules */
	char		mangle_postroute_policy;	/* policy for this chain: 0: accept, 1: drop */

	/*
		nat
	*/
	d_list		nat_preroute;			/* list with rules */
	char		nat_preroute_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		nat_postroute;			/* list with rules */
	char		nat_postroute_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		nat_output;			/* list with rules */
	char		nat_output_policy;		/* policy for this chain: 0: accept, 1: drop */

	/*
		filter
	*/
	d_list		filter_input;			/* list with rules */
	char		filter_input_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		filter_forward;			/* list with rules */
	char		filter_forward_policy;		/* policy for this chain: 0: accept, 1: drop */
	d_list		filter_output;			/* list with rules */
	char		filter_output_policy;		/* policy for this chain: 0: accept, 1: drop */

	/*
		extra filter (no policies)
	*/
	d_list		filter_antispoof;		/* list with rules */
	d_list		filter_blocklist;		/* list with rules */
	d_list		filter_blocktarget;		/* list with rules */
	d_list		filter_badtcp;			/* list with rules */
	d_list		filter_synlimittarget;		/* list with rules */
	d_list		filter_udplimittarget;		/* list with rules */

	/*
		special chains
	*/
	char		block;				/* the block target */
	char		synlimit;			/* synlimiting */

} RuleSet;


/*@null@*/
struct SHM_TABLE *shm_table;


/* counters */
int ipt_rulecount;
int ipt_input_count;
int ipt_output_count;
int ipt_forward_count;
int ipt_preroute_count;
int ipt_postroute_count;


/* pluginlist */
d_list	PluginList;
/* list of ipaddresses to be blocked */
d_list	BlockList;


/* semaphore id */
int sem_id;

/* pointer to the environment */
extern char **environ;


struct ReloadRules_
{
	char	main;
	char	synlimit;
	char	udplimit;
	char	networkprotect;
	char	blocklist;
	char	systemprotect;
	char	interfaceprotect;

} ReloadRules;


/*************************************************************************************************************************\
 ******************************************************* FUNCTIONS *******************************************************
\*************************************************************************************************************************/

/* rules.c */
int analyze_interface_protect_rules(const int, Rules *, Zones *, Services *, Interfaces *);
int analyze_network_protect_rules(const int, Rules *, Zones *, Services *, Interfaces *);
int analyze_normal_rules(const int, Rules *, Zones *, Services *, Interfaces *);
int analyze_all_rules(const int, Rules *, Zones *, Services *, Interfaces *);

int create_all_rules(const int, Rules *, Zones *, Interfaces *, d_list *, IptCap *, int);

int pre_rules(const int, /*@null@*/RuleSet *, Interfaces *, IptCap *);
int post_rules(const int, /*@null@*/RuleSet *, IptCap *, int);

int update_synlimit_rules(const int, /*@null@*/RuleSet *, IptCap *);
int update_udplimit_rules(const int, /*@null@*/RuleSet *, IptCap *);
int create_block_rules(const int, /*@null@*/RuleSet *, d_list *);

int create_network_protect_rules(const int, /*@null@*/RuleSet *, Zones *, IptCap *);
int create_interface_protect_rules(const int, Interfaces *);
int create_system_protectrules(const int, struct vuurmuur_config *);
int create_normal_rules(const int, /*@null@*/RuleSet *, Rules *, IptCap *, char *);

int create_rule(const int, /*@null@*/RuleSet *, struct RuleCache_ *, IptCap *);
int remove_rule(const int debuglvl, int chaintype, int first_ipt_rule, int rules);

int create_rule_input(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_output(const int,  /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_forward(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_masq(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_snat(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_portfw(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);
int create_rule_redirect(const int, /*@null@*/RuleSet *, struct RuleCreateData_ *, struct RuleCache_ *, IptCap *);

int clear_all_iptables_rules(const int debuglvl);


/* misc.c */
int vuurmuur_check_lib_version(int debuglvl, int major, int minor, int sub, int min_lib_major, int min_lib_minor, int min_lib_sub);

/* main.c */
// none ;-)

/* reload.c */
int apply_changes(const int, Services *, Zones *, Interfaces *, Rules *, d_list *, IptCap *, struct rgx_ *);

int reload_services(const int, Services *, regex_t *);
int reload_services_check(const int, struct ServicesData_ *);

int reload_zonedata(const int, Zones *, Interfaces *, struct rgx_ *);
int reload_zonedata_check(const int, Zones *, Interfaces *, struct ZoneData_ *, struct rgx_ *);

int reload_interfaces(const int, Interfaces *);
int reload_interfaces_check(const int, struct InterfaceData_ *iface_ptr);

int check_for_changed_dynamic_ips(const int debuglvl, Interfaces *interfaces);

/* ruleset */
int ruleset_add_rule_to_set(const int, d_list *, char *, char *, unsigned long long, unsigned long long);
int load_ruleset(const int, Rules *, Zones *, Interfaces *, Services *, d_list *, IptCap *, struct vuurmuur_config *);



#endif
