/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

/* we put this here, because we only use it here in main. */
static int sigint_count = 0;
static int sighup_count = 0;
static void catch_sigint(int signo);
static void catch_sighup(int signo);


int
main(int argc, char *argv[])
{
	Interfaces		interfaces;
	Services		services;
	Zones			zones;
	Rules			rules;
	IptCap			iptcap;

	int			retval=0,
				optch,
				result=0,
				loop=0,
				nodaemon=0,
				debuglvl = 0;
	uid_t			user;
	gid_t			group;
	static char		optstring[] = "hd:bVlvnc:L:C";
	int			reload_shm = 0,
				reload_dyn = 0;
//	struct stat		stat_buf;
	int			shm_id;
	char			*shmp;
	union semun		semarg;
	ushort			seminit[] = { 1,0 };

	int			wait_time = 0;	// time in seconds we have waited for an VR_RR_RESULT_ACK when using SHM-IPC
	struct rgx_		reg;		// regexes
	/* clear all the iptables rules? */
	char			clear_rules = 0;
	/* interfaces data structure */
	int			dynamic_wait_time = 0;	/* for checking the dynamic ipaddresses */

	// registering signals we use
	if(signal(SIGINT, &catch_sigint) == SIG_ERR)
	{
		fprintf(stdout, "Error: couldn't attach the signal SIGINT to the signal handler.\n");
		retval=-1;
	}
	if(signal(SIGHUP, &catch_sighup) == SIG_ERR)
	{
		fprintf(stdout, "Error: couldn't attach the signal SIGHUP to the signal handler.\n");
		retval=-1;
	}

	libvuurmuur_verbose = 0;
	bash_out = 0;
	conf.loglevel_cmdline = 0;
	shm_table = NULL;
	sem_id = 0;
	strlcpy(conf.configfile_location, VUURMUUR_CONFIGFILE, sizeof(conf.configfile_location));

	/* Process commandline options */
	while((optch = getopt(argc, argv, optstring)) != -1 )
	{
		switch(optch)
		{
			case 'c' :
				/* config file */
				if(libvuurmuur_verbose == 1)
					fprintf(stdout, "Using this configfile: %s\n", optarg);

				if(strlcpy(conf.configfile_location, optarg, sizeof(conf.configfile_location)) > sizeof(conf.configfile_location))
				{
					fprintf(stdout, "Error: configfile (-c): argument too long (max: %d).\n", sizeof(conf.configfile_location)-1);
					exit(EXIT_FAILURE);
				}
				break;

			case 'd' :
				/* debugging */
				fprintf(stdout, "vuurmuur: debugging enabled.\n");

				/* convert the debug string and check the result */
				debuglvl = atoi(optarg);
				if(debuglvl < 0 || debuglvl > HIGH)
				{
					fprintf(stdout, "Error: illegal debug level: %d (max: %d).\n", debuglvl, HIGH);
					exit(EXIT_FAILURE);
				}

				fprintf(stdout, "vuurmuur: debug level: %d\n", debuglvl);
				break;

			case 'L' :

				conf.loglevel_cmdline = 1;

				/* loglevel */
				if(strlcpy(conf.loglevel, optarg, sizeof(conf.loglevel)) > sizeof(conf.loglevel))
				{
					fprintf(stdout, "Error: loglevel (-L): argument too long (max: %d).\n", sizeof(conf.loglevel)-1);
					exit(EXIT_FAILURE);
				}

				break;

			case 'b' :

				/* bash output */
				bash_out = 1;
				fprintf(stdout, "#!/bin/sh\n");
				fprintf(stdout, "# Firewall generated by Vuurmuur %d.%d.%d, Copyright (C) 2002-2005 by Victor Julien\n", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
				break;

			case 'h' :
				/* help */
				fprintf(stdout, "Usage: vuurmuur [OPTION]\n");
				fprintf(stdout, "\n");
				fprintf(stdout, "Options:\n");
				fprintf(stdout, "-b gives a bashscript output\n");
				fprintf(stdout, "-d [1 - 3] enables debugging, 1 low, 3 high\n");
				fprintf(stdout, "-h gives this help\n");
				fprintf(stdout, "-V gives the version\n");
				fprintf(stdout, "-l vuurmuur starts and goes into daemon-mode.\n");
				fprintf(stdout, "-v verbose mode.\n");
				fprintf(stdout, "-n for use with -l, it goes into the loop without daemonizing.\n");
				fprintf(stdout, "-C clear all iptables rules and set policy to ACCEPT. Use with care!\n");
				fprintf(stdout, "\n");
				exit(0);

			case 'V' :

				/* print version */
				fprintf(stdout, "Vuurmuur %d.%d.%d\n", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
				fprintf(stdout, "Copyright (C) 2002-2005 by Victor Julien\n");
				exit(0);

			case 'l' :

				/* looping */
				loop=1;
				break;

			case 'v' :

				/* verbose */
				fprintf(stdout, "verbose output\n");
				libvuurmuur_verbose = 1;
				break;

			case 'n' :

				/* dont daemonize */
				fprintf(stdout, "no daemon\n");
				nodaemon=1;
				break;

			case 'C' :

				/* clear all rules */
				fprintf(stdout, "Clearing all rules...\n");
				clear_rules = 1;
				break;

			default:
				//fprintf(stdout, "Error: unknown option '-%c'. See -h for valid options.\n", optch);
				break;
		}
	}

//TODO: do an options sanity check (eg. bash and loop dont play together)


	/* check if were already running, but not if we want bash output */
	if(bash_out == 0)
	{
		if(check_pidfile(PIDFILE) == -1)
			exit(EXIT_FAILURE);
	}

/*
	// check if the binary is a symlink
	if(lstat(argv[0], &stat_buf) == -1)
	{
		fprintf(stdout, "Error: lstat failed on the executable.\n");
		return(-1);
	}
	else
	{
		if(S_ISLNK(stat_buf.st_mode) == 1)
		{
			fprintf(stdout, "For security reasons Vuurmuur will not run when called from a symbolic link.\n");
			return(-1);
		}
	}
*/

	/* init the print functions: all to stdout */
	vrprint.logger = "vuurmuur";
	vrprint.error = libvuurmuur_stdoutprint_error;
	vrprint.warning = libvuurmuur_stdoutprint_warning;
	vrprint.info = libvuurmuur_stdoutprint_info;
	vrprint.debug = libvuurmuur_stdoutprint_debug;

	/* get userid and groupid */
	user = getuid();
	group = getgid();

	if(debuglvl >= LOW)
	{
		fprintf(stdout, "main: user=%d, group=%d\n", (int)user, (int)group);
	}

	/*	exit if the user is not root. BUG/TODO - we could enable this program also for non-root users,
		but for example only with bashout on and loop off...
	*/
	if(user > 0 || group > 0)
	{
		fprintf(stdout, "Error: you are not root! Exitting.\n");
		exit(EXIT_FAILURE);
	}

	/* check for the version of the library */
	result = vuurmuur_check_lib_version(debuglvl, VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB, MIN_LIBVUURMUUR_VERSION_MAJOR, MIN_LIBVUURMUUR_VERSION_MINOR, MIN_LIBVUURMUUR_VERSION_SUB);
	if(result < 0)
		exit(EXIT_FAILURE);

	/* init the config */
	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "main: initializing config... calling init_config()");

	result = init_config(debuglvl, conf.configfile_location, &conf);
	if(result >= VR_CNF_OK)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug("Debug", "main: initializing config complete and succesful.");
	}
	else
	{
		fprintf(stdout, "Initializing config failed.\n");
		exit(EXIT_FAILURE);
	}

	/* after the config we can remove the rules if we need to */
	if(clear_rules)
	{
		if(clear_all_iptables_rules(debuglvl) < 0)
		{
			fprintf(stdout, "Error: clearing iptablesrules failed.\n");
			exit(EXIT_FAILURE);
		}

		exit(EXIT_SUCCESS);
	}


	/* now we know the logfile locations, so init the log functions */
	vrprint.error = libvuurmuur_logprint_error;
	vrprint.warning = libvuurmuur_logprint_warning;
	vrprint.info = libvuurmuur_logprint_info;
	vrprint.debug = libvuurmuur_logprint_debug;


	/* check capabilities */
	if(!bash_out)
	{
		if(check_iptcaps(debuglvl, &iptcap, conf.load_modules) < 0)
		{
			fprintf(stdout, "Error: checking for iptables-capabilities failed. Please see error.log.\n");
			exit(EXIT_FAILURE);
		}
	}

	/* setup regexes */
	if(setup_rgx(1, &reg) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up regular expressions failed.");
		exit(EXIT_FAILURE);
	}

	/* load the backends */
	result = load_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		fprintf(stdout, "Error: loading backends failed\n");
		exit(EXIT_FAILURE);
	}

	/* print some nice info about me being the coolest of 'm all ;-) */
	(void)vrprint.info("Info", "This is Vuurmuur %d.%d.%d", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
	(void)vrprint.info("Info", "Copyright (C) 2002-2005 by Victor Julien");

	/* set chain couters to zero */
	ipt_rulecount = 0;
	ipt_input_count = 0;
	ipt_output_count = 0;
	ipt_forward_count = 0;
	ipt_preroute_count = 0;
	ipt_postroute_count = 0;

	/* load the services into memory */
	(void)vrprint.info("Info", "Loading services...");
	result = init_services(debuglvl, &services, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading services succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all services where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "initializing services failed.");
		exit(EXIT_FAILURE);
	}
	//services_print_list(&ServicesList);

	/* load the interfaces into memory */
	result = init_interfaces(debuglvl, &interfaces);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading interfaces succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all interfaces where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Initializing interfaces failed.");
		exit(EXIT_FAILURE);
	}

	/* load the zonedata into memory */
	(void)vrprint.info("Info", "Loading zones...");
	result = init_zonedata(debuglvl, &zones, &interfaces, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading zonedata succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all zones where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Initializing zones failed.");
		exit(EXIT_FAILURE);
	}
	//zonedata_print_list(&ZonedataList);


	/* load the blockfile if any */
	if(conf.use_blocklist)
	{
		/* call it with load_ips == 1 */
		if(blocklist_read_file(debuglvl, &zones, &BlockList, 1) < 0)
			conf.use_blocklist = 0;
	}


	/* load the rulesfile into memory */
	(void)vrprint.info("Info", "Loading rulesfile...");
	result = rules_init_list(debuglvl, &rules, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading rulesfile succesfull.");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "initializing the rules failed.");
		exit(EXIT_FAILURE);
	}

	/* analyzing the rules */
	if(analyze_all_rules(debuglvl, &rules, &zones, &services, &interfaces) != 0)
	{
		(void)vrprint.error(-1, "Error", "analizing the rules failed.");
		exit(EXIT_FAILURE);
	}

	if(debuglvl >= LOW)
		rules_print_list(&rules);

	/* now create the rules */
	if(conf.old_rulecreation_method || bash_out)
	{
		if(create_all_rules(debuglvl, &rules, &zones, &interfaces, &BlockList, &iptcap, 1) != 0)
		{
			(void)vrprint.error(-1, "Error", "creating rules failed.");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		if(load_ruleset(debuglvl, &rules, &zones, &interfaces, &services, &BlockList, &iptcap, &conf) < 0)
		{
			(void)vrprint.error(-1, "Error", "creating rules failed.");
			exit(EXIT_FAILURE);
		}
	}


	// ** analyzing and creating rules done ** //


	/* enter the infinite loop... */
	if(loop)
	{
		/* if we going to enter the loop, make sure we dont are in bash-mode */
		if(bash_out == 0)
		{
			/* leave console */
			if(nodaemon == 0)
			{
				if(daemon(1,1) < 0)
				{
					(void)vrprint.error(-1, "Error", "failed to daemonize: %s.", strerror(errno));
					exit(EXIT_FAILURE);
				}
				else
					(void)vrprint.info("Info", "Entered daemon mode: obtained a new PID (%ld).", getpid());
			}

			shm_id = shmget(IPC_PRIVATE, sizeof(*shm_table), 0600);
			if(shm_id < 0)
			{
				(void)vrprint.error(-1, "Error", "unable to create shared memory: %s.", strerror(errno));
				exit(EXIT_FAILURE);
			}
			/* for some reason on my machine the shm_id is zero when vuurmuur is started at boot
			   if we sleep for some time and retry it works */
			else if(shm_id == 0)
			{
				/* sleep 3 seconds before trying again */
				(void)sleep(3);

				shm_id = shmget(IPC_PRIVATE, sizeof(*shm_table), 0600);
				if(shm_id < 0)
				{
					(void)vrprint.error(-1, "Error", "Unable to create shared memory: %s (retry).", strerror(errno));
					exit(EXIT_FAILURE);
				}
				else if(shm_id == 0)
				{
					(void)vrprint.info("Info", "Still no valid shm_id. Giving up.");
				}
				else
				{
					(void)vrprint.info("Info", "Creating shared memory successfull: shm_id: %d (retry).", shm_id);
				}
			}
			else
			{
				(void)vrprint.info("Info", "Creating shared memory successfull: shm_id: %d.", shm_id);
			}

			/* now attach to the shared mem */
			if(shm_id > 0)
			{
				shmp = shmat(shm_id, 0, 0);
				if(shmp == (char *)(-1))
				{
					(void)vrprint.error(-1, "Error", "unable to attach to shared memory: %s.", strerror(errno));
					exit(EXIT_FAILURE);
				}
				else
				{
					shm_table = (struct SHM_TABLE *)shmp;
					(void)vrprint.info("Info", "Attaching to shared memory successfull.");
				}

				/* if all went well we create a semaphore */
				if(shm_table)
				{
					sem_id = semget(IPC_PRIVATE, 2, 0600);
					if(sem_id == -1)
					{
						(void)vrprint.error(-1, "Error", "Unable to create semaphore: %s.", strerror(errno));
						exit(EXIT_FAILURE);
					}
					else
					{
						(void)vrprint.info("Info", "Creating a semaphore success: %d", sem_id);
					}

					semarg.array = seminit;
					if(semctl(sem_id, 0, SETALL, semarg) == -1)
					{
						(void)vrprint.error(-1, "Error", "Unable to initialize semaphore: %s.", strerror(errno));
						exit(EXIT_FAILURE);
					}
					else
					{
						(void)vrprint.info("Info", "Initializeing the semaphore successfull.");
					}

					/* now initialize the shared mem */
					if(LOCK(sem_id))
					{
						shm_table->sem_id = sem_id;
						shm_table->backend_changed = 0;
						shm_table->reload_result = VR_RR_READY;

						UNLOCK(sem_id);
					}
				}
			}

			/* create a pidfile */
			result = create_pidfile(PIDFILE, shm_id);
			if(result < 0)
			{
				(void)vrprint.error(-1, "Error", "Unable to create pidfile.");
/* TODO: is this really that serious? */
				exit(EXIT_FAILURE);
			}

			(void)vrprint.info("Info", "Entering the loop... (interval %d seconds)", LOOP_INT);

			while(retval == 0 && sigint_count == 0)
			{
				if(LOCK(sem_id))
				{
					if(shm_table->configtool.connected == 1)
					{
						(void)vrprint.info("Info", "Configtool connected: %s.", shm_table->configtool.name);
						shm_table->configtool.connected=2;
					}
					else if(shm_table->configtool.connected == 3)
					{
						(void)vrprint.info("Info", "Configtool disconnected: %s.", shm_table->configtool.name);
						shm_table->configtool.connected=0;
					}

					if(shm_table->backend_changed == 1)
					{
						(void)vrprint.info("Info", "IPC-SHM: backend changed.");
						reload_shm = 1;
						shm_table->backend_changed = 0;

						/* start at 0% */
						shm_table->reload_progress = 0;
					}

					UNLOCK(sem_id);
				}

				/*	if we have one or more dynamic interfaces
					we check if there we're changes.
				*/
				if(conf.dynamic_changes_check && interfaces.dynamic_interfaces)
				{
					dynamic_wait_time++;

					if(dynamic_wait_time >= conf.dynamic_changes_interval)
					{
						(void)vrprint.debug(__FUNCTION__, "check the dynamic ipaddresses.");

						if(check_for_changed_dynamic_ips(debuglvl, &interfaces))
						{
							reload_dyn = 1;
						}

						dynamic_wait_time = 0;
					}
				}

				/*	well, we either recieved a SIGHUP or we want to reload trough an IPC command, or we
					have an interface with a changed ip.
				*/
				if(sighup_count == 1 || reload_shm == 1 || reload_dyn == 1)
				{
					/* apply changes */
					result = apply_changes(debuglvl, &services, &zones, &interfaces, &rules, &BlockList, &iptcap, &reg);
					if(result < 0)
					{
						(void)vrprint.error(-1, "Error", "applying changes failed.");
/* TODO: shouldn't we use retval here? */
						exit(EXIT_FAILURE);
					}
					else
					{
						sighup_count = 0;
						reload_dyn = 0;
					}

					/* if we are reloading because of an IPC command, we need to communicate with the caller */
					if(reload_shm == 1)
					{
						if(LOCK(sem_id))
						{
							/* finished so 100% */
							shm_table->reload_progress = 100;

							/* tell the caller about the reload result */
							if(result < 0)
							{
								shm_table->reload_result = VR_RR_ERROR;
							}
							else if(result == 0)
							{
								shm_table->reload_result = VR_RR_SUCCES;
							}
							else
							{
								shm_table->reload_result = VR_RR_NOCHANGES;
							}
							UNLOCK(sem_id);
						}
						reload_shm = 0;
					
						(void)vrprint.info("Info", "Waiting for an VR_RR_RESULT_ACK");

						result = 0;
						wait_time = 0;

						/* now wait max 30 seconds for an ACK from the caller */
						while(result == 0 && wait_time < 30)
						{
							if(LOCK(sem_id))
							{
								/* ah, we got one */
								if(shm_table->reload_result == VR_RR_RESULT_ACK)
								{
									shm_table->reload_result = VR_RR_READY;
									shm_table->reload_progress = 0;
									result = 1;

									(void)vrprint.info("Info", "We got an VR_RR_RESULT_ACK!");
								}
								UNLOCK(sem_id);
							}

							wait_time++;
							sleep(1);
						}

						/* damn, we didn't get one */
						if(result == 0)
						{
							(void)vrprint.info("Info", "We've waited for %d seconds for an VR_RR_RESULT_ACK, but got none. Setting to VR_RR_READY", wait_time);
							if(LOCK(sem_id))
							{
								shm_table->reload_result = VR_RR_READY;
								shm_table->reload_progress = 0;
								UNLOCK(sem_id);
							}
							else
							{
								(void)vrprint.info("Info", "Hmmmm, failed to set to ready. Did the client crash?");
							}
						}
						result = 0;
					}
				}

				sleep(LOOP_INT);
			}

			(void)vrprint.info("Info", "Destroying shared memory...");
			if(shmctl(shm_id, IPC_RMID, NULL) < 0)
			{
				(void)vrprint.error(-1, "Error", "destroying shared memory failed: %s.", strerror(errno));
				retval = -1;
			}
			else
			{
				if(debuglvl >= LOW)
					(void)vrprint.debug(__FUNCTION__, "shared memory destroyed.");
			}

			/* destroy semaphore */
			if(semctl(sem_id, 0, IPC_RMID, semarg) == -1)
			{
				(void)vrprint.error(-1, "Error", "failed to remove semaphore.");
				retval = -1;
			}

			/* remove the pidfile */
			if(remove_pidfile(PIDFILE) < 0)
			{
				(void)vrprint.error(-1, "Error", "unable to remove pidfile: %s.", strerror(errno));
				retval = -1;
			}

			(void)vrprint.info("Info", "Loop shutting down...");
		}
		else
		{
			fprintf(stdout, "# loop mode not supported when using -b (bash output).\n");
		}
	}

	/* unload the backends */
	result = unload_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		fprintf(stdout, "Error: unloading backends failed.\n");
		exit(EXIT_FAILURE);
	}


	/*
		Destroy the linked-lists
	*/

	/* destroy the ServicesList */
	destroy_serviceslist(debuglvl, &services);

	/* destroy the ZonedataList */
	destroy_zonedatalist(debuglvl, &zones);

	/* destroy the InterfacesList */
	destroy_interfaceslist(debuglvl, &interfaces);

	/* destroy the QuerydataList */
	if(rules_cleanup_list(debuglvl, &rules) < 0)
		retval = -1;

	d_list_cleanup(debuglvl, &BlockList);

	/* cleanup regexes */
	(void)setup_rgx(0, &reg);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "main: ** end **, return = %d", retval);

	return(retval);
}

static void catch_sigint(int signo)
{
	sigint_count = 1;
}

static void catch_sighup(int signo)
{
	sighup_count = 1;
}
