/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "main.h"


typedef struct
{
	char	newline;
	char	*word;

	int	line_num;

} helpword;


static void
free_helpword(void *ptr)
{
	helpword	*hw = NULL;
	
	if(!ptr)
		return;
		
	hw = (helpword *)ptr;
		
	if(hw->word != NULL)
		free(hw->word);
		
	free(hw);
	
	return;
}


int
read_helpline(const int debuglvl, d_list *help_list, char *line)
{
	char		oneword[512] = "";
	int		i = 0;
	int		k = 0;
	helpword	*hw = NULL;

	for(i = 0, k = 0; i < strlen(line); i++)
	{
		if(line[i] == ' ' || line[i] == '\n')
		{
			oneword[k] = '\0';
			k = 0;

			/* only add a word to the list if it really contains characters */
			if(strlen(oneword) > 0)
			{
				/* get some mem for the word struct */
				if(!(hw = malloc(sizeof(helpword))))
				{
					(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
					return(-1);
				}
				hw->word = NULL;
				hw->newline = 0;
				hw->line_num = 0;

				if(!(hw->word = malloc(strlen(oneword) + 1)))
				{
					(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
					return(-1);
				}
				strlcpy(hw->word, oneword, strlen(oneword) + 1);

				if(d_list_append(debuglvl, help_list, hw) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "append to list %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
			}

			/* the newline is a special word */
			if(	(i == 0 && line[i] == '\n') ||

				(i > 0  && line[i] == '\n' &&
				(line[i-1] == '.' ||
				 line[i-1] == '?' ||
				 line[i-1] == '!' ||
				 line[i-1] == ':' ||
				(line[i-1] == ',' && i == (strlen(line) - 1) ))
				)
			)
			{
				/* get some mem for the word struct */
				if(!(hw = malloc(sizeof(helpword))))
				{
					(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
					return(-1);
				}
				hw->word = NULL;
				hw->newline = 1;
				hw->line_num = 0;

				if(d_list_append(debuglvl, help_list, hw) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "append to list %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
			}
		}
		else
		{
			oneword[k] = line[i];
			k++;
		}
	}

	return(0);
}


int
read_helpfile(const int debuglvl, d_list *help_list, char *part)
{
	char		line[512] = "";
	FILE		*fp = NULL;
	char		inrange = 0;

	/* safety */
	if(!help_list || !part)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* setup the list */
	if(d_list_setup(debuglvl, help_list, free_helpword) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setup the helplist failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* open the file */
	if(!(fp = fopen(vccnf.helpfile_location, "r")))
	{
		(void)vrprint.error(-1, "Error", "opening '%s' failed: %s (in: %s:%d).", vccnf.helpfile_location, strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	while(fgets(line, sizeof(line), fp) != NULL)
	{
		if(inrange)
		{
			if(strcmp(line, ":[END]:\n") == 0)
			{
				inrange = 0;
				break;
			}
		}

		if(inrange)
		{
			if(read_helpline(debuglvl, help_list, line) < 0)
				return(-1);
		}
		else
		{
			if(strncmp(line, part, strlen(part)) == 0)
				inrange = 1;
		}
	}
	
	fclose(fp);
	return(0);
}


static void
set_lines(const int debuglvl, d_list *help_list, int width)
{
	int		line_width = 0,
			line_num = 1,
			words = 0;
	helpword	*hw = NULL,
			*next_hw = NULL;
	d_list_node	*d_node = NULL,
			*next_d_node = NULL;


	/* safety */
	if(!help_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return;
	}

	for(d_node = help_list->top; d_node; d_node = d_node->next, words++)
	{
		if(!(hw = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return;
		}

		if(hw->word != NULL && !hw->newline)
		{
			/* get the next word */
			if((next_d_node = d_node->next))
			{
				if(!(next_hw = next_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return;
				}

				/* middle in sentence, so we add a space to the word */
				if(next_hw->word != NULL && next_hw->newline == 0)
				{
					/* don't add this on the current line */
					if((strlen(hw->word) + 1) >= (width - line_width))
					{
						/* set line_width to the size of this word */
						line_width = strlen(hw->word) + 1;

						line_num++;
						hw->line_num = line_num;
					}
					/* this word fits on the current line */
					else
					{
						/* add this word to the line_width */
						line_width = line_width + strlen(hw->word) + 1;

						hw->line_num = line_num;
					}
				}
				/* end of sentence, so no trailing space */
				else if(next_hw->word == NULL && next_hw->newline == 1)
				{
					/* don't add this on the current line */
					if(strlen(hw->word) >= (width - line_width))
					{
						/* set line_width to the size of this word */
						line_width = strlen(hw->word);

						line_num++;
						hw->line_num = line_num;
					}
					/* this word fits on the current line */
					else
					{
						/* add this word to the line_width */
						line_width = line_width + strlen(hw->word);

						hw->line_num = line_num;
					}
				}
				else
				{
					/* undefined state */
					(void)vrprint.error(-1, "Internal Error", "undefined state (in: %s:%d).", __FUNCTION__, __LINE__);
					return;
				}
			}
			/* end of text, doc has no trailing new-line */
			else
			{
				/* don't add this on the current line */
				if(strlen(hw->word) >= (width - line_width))
				{
					/* set line_width to the size of this word */
					line_width = strlen(hw->word);

					line_num++;
					hw->line_num = line_num;
				}
				/* this word fits on the current line */
				else
				{
					/* add this word to the line_width */
					line_width = line_width + strlen(hw->word);

					hw->line_num = line_num;
				}
			}
		}
		else if(hw->word == NULL && hw->newline)
		{
			hw->line_num = line_num;
			line_num++;

			line_width = 0;
		}
		else
		{
			/* undefined state */
			(void)vrprint.error(-1, "Internal Error", "undefined state (in: %s:%d).", __FUNCTION__, __LINE__);
			return;
		}
	}

	return;
}


void
print_list(const int debuglvl, d_list *list, char *title, int height, int width, int starty, int startx)
{
	WINDOW		*boxwin = NULL,
			*printwin = NULL;
	PANEL		*panel[2];
	int		ch;

	helpword	*hw = NULL,
			*next_hw = NULL;
	d_list_node	*d_node = NULL,
			*next_d_node = NULL;
	int		start_print = 1,
			end_print = 1;
	char		done = 0;
	int		i = 0;

	end_print = height - 2;

	if(!(boxwin = create_newwin(height, width, starty, startx, title, COLOR_PAIR(CP_RED_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s).", __FUNCTION__);
		return;
	}
	if(!(panel[0] = new_panel(boxwin)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return;
	}
	if(!(printwin = newwin(height-2, width-4, starty+1, startx+2)))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s).", __FUNCTION__);
		return;
	}
	(void)wbkgd(printwin, COLOR_PAIR(CP_BLACK_WHITE));
	if(!(panel[1] = new_panel(printwin)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return;
	}
	keypad(printwin, TRUE);

	while(!done)
	{
		werase(printwin);

		if(list->len == 0)
		{
			wprintw(printwin, "The requested helptext was not found.\n");
		}

		/* print the text */
		for(d_node = list->top; d_node; d_node = d_node->next)
		{
			if(!(hw = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return;
			}

			if(hw->line_num >= start_print && hw->line_num <= end_print)
			{
				if(hw->word != NULL)
				{
					if((next_d_node = d_node->next))
					{
						if(!(next_hw = next_d_node->data))
						{
							(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
							return;
						}
					

						/* end of sentence */
						if(next_hw->word == NULL && next_hw->newline == 1)
						{
							wprintw(printwin, "%s", hw->word);
						}
						/* middle in sentence */
						else if(next_hw->line_num > hw->line_num)
						{
							wprintw(printwin, "%s\n", hw->word);
						}
						else
						{
							wprintw(printwin, "%s ", hw->word);
						}
					}
					/* end of text */
					else
					{
						wprintw(printwin, "%s", hw->word);
					}
				}
				/* new line */
				else
				{
					wprintw(printwin, "\n");
				}
			}
		}

		update_panels();
		doupdate();

		/* get user input */
		ch = wgetch(printwin);

		switch(ch)
		{
			case KEY_DOWN:

				if(list->len > 0)
				{
					if(!(hw = list->bot->data))
					{
						(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
						return;
					}

					if(end_print < hw->line_num)
					{
						start_print++;
						end_print++;
					}
				}
				break;

			case KEY_UP:

				if(list->len > 0)
				{
					if(start_print > 1)
					{
						start_print--;
						end_print--;
					}
				}
				break;
		
			case KEY_PPAGE:

				if(list->len > 0)
				{
					i = (height - 2)/3;

					while((start_print - i) < 1)
						i--;

					start_print = start_print - i;
					end_print = end_print - i;
				}

				break;

			case KEY_NPAGE:

				if(list->len > 0)
				{
					i = (height - 2)/3;

					if(!(hw = list->bot->data))
					{
						(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
						return;
					}

					while((end_print + i) > hw->line_num)
						i--;

					start_print = start_print + i;
					end_print = end_print + i;
				}

				break;

			default:

				done = 1;
				break;
		}
	}

	del_panel(panel[0]);
	del_panel(panel[1]);
	destroy_win(printwin);
	destroy_win(boxwin);

	update_panels();
	doupdate();
	return;
}


void
print_help(const int debuglvl, char *part)
{
	d_list		HelpList;
	int		max_height = 0,
			max_width = 0,
			height = 0,
			width = 0,
			startx = 0,
			starty = 0;

	/* get screensize */
	getmaxyx(stdscr, max_height, max_width);

	width  = 72;
	height = max_height - 6;
	startx = max_width - width - 5;
	starty = 3;

	/* read the helpfile */
	if(read_helpfile(debuglvl, &HelpList, part) < 0)
	{
		(void)vrprint.error(-1, "Error", "helpfile could not be opened.");
		return;
	}
	
	set_lines(debuglvl, &HelpList, width - 4);

	print_list(debuglvl, &HelpList, "Help", height, width, starty, startx);

	d_list_cleanup(debuglvl, &HelpList);
}


void
print_status(const int debuglvl)
{
	int		max_height = 0,
			max_width = 0,
			height = 0,
			width = 0,
			startx = 0,
			starty = 0;

	/* get screensize */
	getmaxyx(stdscr, max_height, max_width);

	width  = 72;
	height = max_height - 6;
	startx = (max_width - width) / 2;
	starty = 3;

	/* should not happen */
	if(VuurmuurStatus.StatusList.len == 0)
	{
		(void)read_helpline(debuglvl, &VuurmuurStatus.StatusList, "No problems were detected in the current setup.\n");
	}

	set_lines(debuglvl, &VuurmuurStatus.StatusList, width - 4);

	/* print the status list */
	print_list(debuglvl, &VuurmuurStatus.StatusList, "Status", height, width, starty, startx);
}


int
setup_statuslist(const int debuglvl)
{
	/* initialize */
	memset(&VuurmuurStatus, 0, sizeof(VuurmuurStatus));
	memset(&VuurmuurStatus.StatusList, 0, sizeof(VuurmuurStatus.StatusList));

	VuurmuurStatus.vuurmuur = 1;
	VuurmuurStatus.vuurmuur_log = 1;
	
	VuurmuurStatus.zones = 1;
	VuurmuurStatus.interfaces = 1;
	VuurmuurStatus.services = 1;
	VuurmuurStatus.rules = 1;

	VuurmuurStatus.shm = 1;
	VuurmuurStatus.backend = 1;
	VuurmuurStatus.config = 1;
	VuurmuurStatus.settings = 1;
	VuurmuurStatus.system = 1;

	/* setup the status list */
	if(d_list_setup(debuglvl, &VuurmuurStatus.StatusList, free_helpword) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setup the statuslist failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}
