/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

#define FIELDS_PER_BAR	7

#define MIN_ACTIVE	4
#define MIN_NUMFIELD	4
#define MIN_ACTION	7
#define	MIN_SERVICE	9
#define	MIN_FROM	18
#define	MIN_TO		18
#define	MIN_OPTIONS	14

#define MAX_ACTIVE	4
#define MAX_NUMFIELD	4
#define MAX_ACTION	10
#define	MAX_SERVICE	32
#define	MAX_FROM	96	// TODO
#define	MAX_TO		96	// TODO
#define	MAX_OPTIONS	256	// TODO


/*	rulebar

	Container for pointers to fields.
*/
typedef struct
{
	int	bar_num;

	/* pointers the the fields */
	FIELD	*active;
	FIELD	*num_field;
	FIELD	*action;
	FIELD	*service;
	FIELD	*from;
	FIELD	*to;
	FIELD	*options;
}  rulebar;


struct RuleBarForm_
{
	d_list	RuleBar_list;

	int	bars;

	int	max_bars_on_screen;
	int	filtered_rules;

	int	printable_rules;
	int	scroll_offset;

	/* for regex filtering */
	regex_t	filter_reg;
	int	use_filter;
	
	/* some more filtering */
	char	show_only_forward,
		show_only_input,
		show_only_output;

	/* field sizes */
	size_t	active_size;
	size_t	num_field_size;
	size_t	action_size;
	size_t	service_size;
	size_t	from_size;
	size_t	to_size;
	size_t	options_size;
};


static int SetupRuleBarForm(const int, struct RuleBarForm_ *, int, Rules *, int);
static int move_rule(const int, int, int, Rules *, struct RuleBarForm_ *);
static int MatchFilter_RuleBar(struct RuleData_ *rule_ptr, regex_t *reg, char only_in, char only_out, char only_forward);
static int Toggle_RuleBar(const int debuglvl, rulebar *bar, Rules *rules);
static int draw_rules(const int, Rules *, struct RuleBarForm_ *);
static int Enter_RuleBar(const int, rulebar *, Rules *, Zones *, Interfaces *, Services *, struct rgx_ *);


static int
SetupRuleBarForm(const int debuglvl, struct RuleBarForm_ *rbform, int max_bars_on_screen, Rules *rules, int screen_width)
{
	int			sum,
				i;

	/* safety checks */
	if(!rbform || !rules || max_bars_on_screen <= 0 || screen_width <= 0)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: SetupRuleBarForm).");
		return(-1);
	}

	/* init bars */
	rbform->bars = 0;
	rbform->max_bars_on_screen = max_bars_on_screen;

	rbform->filtered_rules = 0;
	rbform->printable_rules = 0;
	rbform->scroll_offset = 0;
	
	rbform->use_filter = 0;
	rbform->show_only_input = 0;
	rbform->show_only_output = 0;
	rbform->show_only_forward = 0;

	/* setup list */
	if(d_list_setup(debuglvl, &rbform->RuleBar_list, free) < 0)
		return(-1);

	/* calculate field sizes */
	rbform->active_size    = MIN_ACTIVE;
	rbform->num_field_size = MIN_NUMFIELD;
	rbform->action_size    = MIN_ACTION;
	rbform->service_size   = MIN_SERVICE;
	rbform->from_size      = MIN_FROM;
	rbform->to_size        = MIN_TO;
	rbform->options_size   = MIN_OPTIONS;

	sum =	rbform->active_size + rbform->num_field_size + rbform->action_size +
		rbform->service_size +rbform->from_size + rbform->to_size + rbform->options_size;
	if(sum > screen_width)
	{
		(void)vrprint.error(-1, "Error", "screen too small (sum: %d, width: %d).", sum, screen_width);
		return(-1);
	}

	while(sum <= screen_width)
	{
		for(i = 0; sum <= screen_width && i < 6; i++)
		{
			if(i == 0 && rbform->action_size < MAX_ACTION)
				rbform->action_size++;
			else if(i == 1 && rbform->service_size < MAX_SERVICE)
				rbform->service_size++;
			else if(i == 2 && rbform->from_size < MAX_FROM)
				rbform->from_size++;
			else if(i == 3 && rbform->to_size < MAX_TO)
				rbform->to_size++;
			else if(i == 5 && rbform->options_size)
				rbform->options_size++;

			sum =	rbform->active_size + rbform->num_field_size + rbform->action_size + rbform->service_size +
				rbform->from_size + rbform->to_size + rbform->options_size;
		}
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "SetupRuleBarForm succes.");

	return(0);
}


/*
	TODO: error checking
*/
static int
move_rule(const int debuglvl, int rule_num, int new_place, Rules *rules, struct RuleBarForm_ *rbform)
{
	int			retval = 0,
				i = 0;
	struct RuleData_	*query_ptr = NULL;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!rules || !rbform)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	status_print(status_win, "Moving rule...");

	/* santiy check for new_place */
	if(new_place > rules->list.len)
	{
		(void)vrprint.debug(__FUNCTION__, "new_place (%d) > rules_list->len (%d) so new_place = %d", new_place, rules->list.len, rules->list.len);
		new_place = rules->list.len;
	}
	else if(new_place <= 0)
		new_place = 1;


	for(d_node = rules->list.top; d_node ; d_node = d_node->next)
	{
		if(!(query_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(query_ptr->number == rule_num)
			break;
	}

	(void)vrprint.debug("Debug", "** move_rule: query_ptr found: i: %d (query_ptr: %s %s %s %s)", i, query_ptr->action, query_ptr->service, query_ptr->from, query_ptr->to);

//TODO
	rules_remove_rule_from_list(debuglvl, rules, rule_num, 1);

	query_ptr->number = new_place;
	(void)vrprint.debug("Debug", "move_rule: new_place: %d, query_ptr->number: %d", new_place, query_ptr->number);

	rules_insert_list(debuglvl, rules, new_place, query_ptr);

        if(debuglvl >= LOW)
		rules_print_list(rules);

	return(retval);
}



/*	Move_RuleBarForm

	display a screen TODO
*/
static int
MoveRuleBarForm(const int debuglvl, struct RuleBarForm_ *rbform, Rules *rules, int cur_rule)
{
	int		ch,
			quit=0;
	WINDOW		*move_win;
	PANEL		*panels[1];
	FIELD		**fields;
	FORM		*form;

	if(cur_rule <= 0 || !rules)
		return(-1);

	// create window, panel, fields, form and post it
	if(!(move_win = create_newwin(6, 40, (LINES-6)/2, (COLS-40)/2, "Move Rule", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "unable to create window.");
		return(-1);
	}
	keypad(move_win, TRUE);
	panels[0] = new_panel(move_win);

	fields = (FIELD **)calloc(1 + 1, sizeof(FIELD *));
	fields[0] = new_field(1, 5, 1, 28, 0, 0);
	set_field_type(fields[0], TYPE_INTEGER, 5, 1, 99999);
	fields[1] = NULL;

	set_field_back(fields[0], COLOR_PAIR(CP_WHITE_BLUE));
	field_opts_off(fields[0], O_AUTOSKIP);
	set_field_status(fields[0], FALSE);

	// Create the form and post it
	form = new_form(fields);
	// Set main window and sub window and post the form
	set_form_win(form, move_win);
	set_form_sub(form, derwin(move_win, 4, 36, 1, 1));
	post_form(form);

	// labels
	mvwprintw(move_win, 2, 2, "Enter the new rule number:");
	mvwprintw(move_win, 4, 2, "Cur: %d, Min: 1, Max: %d", cur_rule, rules->list.len);

	(void)vrprint.debug("Debug", "Move_RuleBarForm: cur_rule: %d, QuerydataList.len: %d.", cur_rule, rules->list.len);

	update_panels();
	doupdate();

	while(quit == 0)
	{
		// get user input
		ch = wgetch(move_win);

		switch(ch)
		{
			case 10: // enter
				form_driver(form, REQ_VALIDATION);
				
//TODO atoi sepparate
				if(move_rule(debuglvl, cur_rule, atoi(field_buffer(fields[0], 0)), rules, rbform) < 0)
					return(-1);
				quit = 1;
				break;

			case KEY_BACKSPACE:
			case 127:
				form_driver(form, REQ_PREV_CHAR);
				form_driver(form, REQ_DEL_CHAR);
				form_driver(form, REQ_END_LINE);
				break;

			case KEY_DC:
				form_driver(form, REQ_PREV_CHAR);
				form_driver(form, REQ_DEL_CHAR);
				form_driver(form, REQ_END_LINE);
				break;

			case KEY_F(10):
			case 27:
			case 'q':
				quit = 1;
				break;

			default:
				form_driver(form, ch);
				break;
		}
	}

	/*
		cleanup
	*/
	unpost_form(form);
	free_form(form);
	free_field(fields[0]);
	free(fields);

	del_panel(panels[0]);
	destroy_win(move_win);

	update_panels();
	doupdate();

	return(0);
}


rulebar *
CurrentBar(struct RuleBarForm_ *rbform, FORM *form)
{
	FIELD		*cur_field = NULL;
	rulebar		*cur_bar = NULL;
	d_list_node	*d_node = NULL;

	/* safety */
	if(!rbform || !form)
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
		return(NULL);
	}

	/* get the current field */
	cur_field = current_field(form);
	/* look for the current bar */
	for(d_node = rbform->RuleBar_list.top; d_node; d_node = d_node->next)
	{
		if(!(cur_bar = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(NULL);
		}

		if(cur_bar->active == cur_field)
			return(cur_bar);
	}

	(void)vrprint.error(-1, "Internal Error", "bar not found!");
	return(NULL);
}


static void
rulebar_setcolor(FIELD *active, FIELD *num_field, FIELD *action, FIELD *service, FIELD *from, FIELD *to, FIELD *options, chtype color)
{
	char	active_rule = 0;

	/* active */
	if(color == COLOR_BLUE)
		set_field_back(active, COLOR_PAIR(CP_WHITE_BLUE));
	else if(color == COLOR_WHITE)
		set_field_back(active, COLOR_PAIR(CP_BLUE_WHITE));
	else if(color == COLOR_RED)
		set_field_back(active, COLOR_PAIR(CP_RED_WHITE));

	if(strncmp(field_buffer(active, 0), "[x]", 3) == 0)
	{
		active_rule = 1;
	}

	/* num_field */
	if(color == COLOR_BLUE)
		set_field_back(num_field, COLOR_PAIR(CP_WHITE_BLUE));
	else if(color == COLOR_WHITE)
		set_field_back(num_field, COLOR_PAIR(CP_BLUE_WHITE));
	else if(color == COLOR_RED)
		set_field_back(num_field, COLOR_PAIR(CP_RED_WHITE));

	/* action */
	if(strncasecmp(field_buffer(action, 0), "drop", 4) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_RED_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "reject", 6) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_RED_BLUE));
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "accept", 6) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_GREEN_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_GREEN_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "log", 3) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "portfw", 6) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "snat", 4) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "masq", 4) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "redirect", 8) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else if(strncasecmp(field_buffer(action, 0), "queue", 5) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
		else if(color == COLOR_WHITE)
			set_field_back(action, COLOR_PAIR(CP_BLUE_WHITE));
		else if(color == COLOR_RED)
			set_field_back(action, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(action, COLOR_PAIR(CP_WHITE_BLUE));
	}


	/* service */
	if(color == COLOR_BLUE && active_rule)
		set_field_back(service, COLOR_PAIR(CP_CYAN_BLUE)|A_BOLD);
	else if(color == COLOR_WHITE)
		set_field_back(service, COLOR_PAIR(CP_CYAN_WHITE)|A_BOLD);
	else if(color == COLOR_RED)
		set_field_back(service, COLOR_PAIR(CP_RED_WHITE));
	else
		set_field_back(service, COLOR_PAIR(CP_WHITE_BLUE));

	/* from zone or firewall */
	if(strncasecmp(field_buffer(from, 0), "firewall", 8) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(from, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(from, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(from, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(from, COLOR_PAIR(CP_WHITE_BLUE));
	}
	else
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(from, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(from, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(from, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(from, COLOR_PAIR(CP_WHITE_BLUE));
	}

	// to
	if(strncasecmp(field_buffer(to, 0), "firewall", 8) == 0)
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(to, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(to, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(to, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(to, COLOR_PAIR(CP_WHITE_BLUE));

	}
	else
	{
		if(color == COLOR_BLUE && active_rule)
			set_field_back(to, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		else if(color == COLOR_WHITE)
			set_field_back(to, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);
		else if(color == COLOR_RED)
			set_field_back(to, COLOR_PAIR(CP_RED_WHITE));
		else
			set_field_back(to, COLOR_PAIR(CP_WHITE_BLUE));
	}

	// options field
	if(color == COLOR_BLUE)
		set_field_back(options, COLOR_PAIR(CP_WHITE_BLUE));
	else if(color == COLOR_WHITE)
		set_field_back(options, COLOR_PAIR(CP_BLUE_WHITE));
	else if(color == COLOR_RED)
		set_field_back(options, COLOR_PAIR(CP_RED_WHITE));
	else
		set_field_back(options, COLOR_PAIR(CP_WHITE_BLUE));
}

/*	Highlight_RuleBar

	Highlights the cursor, and clears the previous highlight.
*/
void
HighlightRuleBar(rulebar *bar)
{
	if(bar)
		rulebar_setcolor(bar->active, bar->num_field, bar->action, bar->service, bar->from, bar->to, bar->options, COLOR_RED);
	else
		(void)vrprint.debug("Debug", "Highlight_RuleBar: warning: no bar.");
}


/*

	Returncodes:
		 1: changed rule
		 0: no changes
		-1: error
*/
static int
Enter_RuleBar(const int debuglvl, rulebar *bar, Rules *rules, Zones *zones, Interfaces *interfaces, Services *services, struct rgx_ *reg)
{
	int	rule_num = 0,
		result = 0,
		retval = 0;

	/* safety */
	if(!bar || !rules || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	
	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "Enter_RuleBar: '%s'.", field_buffer(bar->num_field, 0));

	rule_num = atoi(field_buffer(bar->num_field, 0));
	if(rule_num <= 0)
	{
		(void)vrprint.error(-1, "Internal Error", "invalid rule_num: %d", rule_num);
		return(-1);
	}

	result = edit_rule(debuglvl, rules, zones, interfaces, services, rule_num, reg);
	if(result < 0)
	{
		/* editting failed so remove the rule again */
		if(rules_remove_rule_from_list(debuglvl, rules, rule_num, 1) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "removing rule failed (in: %s).", __FUNCTION__);
			return(-1);
		}
		
		retval = -1;
	}
	else if(result == 1)
	{
		retval = 1;
	}

	return(retval);
}


/*
*/
static int
Toggle_RuleBar(const int debuglvl, rulebar *bar, Rules *rules)
{
	int			rule_num = 0,
				retval = 0;
	int			i = 0;
	d_list_node		*d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "'%s'.", field_buffer(bar->num_field, 0));

	rule_num = atoi(field_buffer(bar->num_field, 0));
	if(rule_num <= 0)
	{
		(void)vrprint.error(-1, "Internal Error", "invalid rule_num: %d (in: %s:%d).", rule_num, __FUNCTION__, __LINE__);
		return(-1);
	}

	if(rules->list.len == 0)
	{
		(void)vrprint.error(-1, "Internal Error", "ruleslist is empty (in: %s)", __FUNCTION__);
		return(-1);
	}
	
	/* go to rulenum in the rules list to get the rule_ptr */
	if(!(d_node = rules->list.top))
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s)", __FUNCTION__);
		return(-1);
	}

	/* look for the rule_ptr */
	for(i = 1; i <= rule_num; i++)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s)", __FUNCTION__);
			return(-1);
		}

		d_node = d_node->next;
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "active: %s (%s %s %s %s)", rule_ptr->active ? "Yes" : "No", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);

	/* set the active */
	if(rule_ptr->active == 1)
		rule_ptr->active = 0;
	else
		rule_ptr->active = 1;

	return(retval);
}


/*	Set_RuleBar

	Sets the rulebar to the position 'pos'.
*/
void
Set_RuleBar(struct RuleBarForm_ *rbform, FORM *form, int pos)
{
	d_list_node		*d_node = NULL;
	rulebar			*cur_bar;
	int			i;

	for(i=1, d_node = rbform->RuleBar_list.top; d_node; i++, d_node = d_node->next)
	{
		cur_bar = d_node->data;

		if(i == pos)
			set_current_field(form, cur_bar->num_field);
	}
}

int
Insert_RuleBar(const int debuglvl, struct RuleBarForm_ *rbform, FIELD *active, FIELD *num_field, FIELD *action, FIELD *service, FIELD *from, FIELD *to, FIELD *options)
{
	rulebar		*bar;

	/* alloc mem */
	if(!(bar = malloc(sizeof(rulebar))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(-1);
	}

	bar->active = active;
	bar->num_field = num_field;
	bar->action = action;
	bar->service = service;
	bar->from = from;
	bar->to = to;
	bar->options = options;

	/* insert the bar into the list */
	if(!(d_list_append(debuglvl, &rbform->RuleBar_list, bar)))
	{
		(void)vrprint.error(-1, "Internal Error", "insert into list failed (in: Insert_RuleBar).");
		return(-1);
	}

	rbform->bars++;
	bar->bar_num = rbform->bars;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "Insert_RuleBar succes at %p (bars: %d, bar->num: %d).", bar, rbform->bars, bar->bar_num);

	return(0);
}


static int
MatchFilter_RuleBar(struct RuleData_ *rule_ptr, regex_t *reg, char only_in, char only_out, char only_forward)
{
	char	*options_ptr = NULL,
		rule_str[512] = "";
	int	option_len = 0,
		rule_len = 0,
		result = 0,
		retval = 0;
		
	if(only_in == 1)
	{
		if(strcmp(rule_ptr->to, "firewall") != 0)
			return(0);
	}
	else if(only_out == 1)
	{
		if(strcmp(rule_ptr->from, "firewall") != 0)
			return(0);
	}
	else if(only_forward == 1)
	{
		if(strcmp(rule_ptr->from, "firewall") == 0 || strcmp(rule_ptr->to, "firewall") == 0)
			return(0);
	}
	
	/* if we're not using a regex, we match here */
	if(!reg)
		return(1);

	if(!(options_ptr = rules_assemble_options_string(0, rule_ptr->opt, rule_ptr->action)))
		option_len = 0;
	else
		option_len = strlen(options_ptr);

	rule_len = strlen(rule_ptr->action)+1 + strlen(rule_ptr->service)+1 + strlen(rule_ptr->from)+1 + strlen(rule_ptr->to)+1 + option_len + 1;

	strlcpy(rule_str, rule_ptr->action, sizeof(rule_str));
	strlcat(rule_str, " ", sizeof(rule_str));
	strlcat(rule_str, rule_ptr->service, sizeof(rule_str));
	strlcat(rule_str, " ", sizeof(rule_str));
	strlcat(rule_str, rule_ptr->from, sizeof(rule_str));
	strlcat(rule_str, " ", sizeof(rule_str));
	strlcat(rule_str, rule_ptr->to, sizeof(rule_str));
	strlcat(rule_str, " ", sizeof(rule_str));
	if(options_ptr)
		strlcat(rule_str, options_ptr, sizeof(rule_str));

	/* now filter */
	result = regexec(reg, rule_str, 0, NULL, 0);
	if(result == 0)
		retval = 1;
	else
		retval = 0;

	return(retval);
}


static int
draw_rules(const int debuglvl, Rules *rules, struct RuleBarForm_ *rbform)
{
	struct RuleData_	*query_ptr;
	rulebar			*cur_bar;
	d_list_node		*d_node = NULL;
	d_list_node		*dl_node = NULL;

	int			draw_count=0,
				printable_lines=0,
				filtered_rules=0;

	char			number[MAX_NUMFIELD] = "",
				active[MAX_ACTIVE] = "",
				action[MAX_ACTION] = "",
				service[MAX_SERVICE] = "",
				from[MAX_FROM] = "",
				to[MAX_TO] = "",
				*option_str,
				options[MAX_OPTIONS] = "";

	int			i,
				x;

	for(	dl_node = rules->list.top, d_node = rbform->RuleBar_list.top;
		dl_node && draw_count < rbform->max_bars_on_screen && d_node;
		dl_node = dl_node->next
		)
	{
//TODO null check
		query_ptr = dl_node->data;

		/* test if the rules fits in the filter */
		if(	(!rbform->use_filter && MatchFilter_RuleBar(query_ptr, NULL,			rbform->show_only_input,
													rbform->show_only_output,
													rbform->show_only_forward)) || 
			(rbform->use_filter && MatchFilter_RuleBar(query_ptr, &rbform->filter_reg,	rbform->show_only_input,
													rbform->show_only_output,
													rbform->show_only_forward))
		)
		{
			printable_lines++;

			if(printable_lines > rbform->scroll_offset)
			{
				/* note if you change the [x] into something else, also change it in rulebar_setcolor */
				snprintf(active, rbform->active_size,    "%s",  query_ptr->active ? "[x]" : "[ ]");
				snprintf(number, rbform->num_field_size, "%2d", printable_lines + filtered_rules);
				snprintf(action, rbform->action_size,    "%s",  query_ptr->action);
				snprintf(service, rbform->service_size,  "%s",  query_ptr->service);
				snprintf(from, rbform->from_size,        "%s",  query_ptr->from);
				snprintf(to, rbform->to_size,            "%s",  query_ptr->to);

				if(!(option_str = rules_assemble_options_string(debuglvl, query_ptr->opt, query_ptr->action)))
					strcpy(options, "-");
				else
				{
					/* cut off: 'options:' */
					for(i=8, x=0; i < strlen(option_str) && x < rbform->options_size-1; i++, x++)
					{
						options[x] = option_str[i];
					}
					options[x]='\0';
				}

//TODO null check
				cur_bar = d_node->data;

				set_field_buffer(cur_bar->active, 0, active);
				set_field_buffer(cur_bar->num_field, 0, number);
				set_field_buffer(cur_bar->action, 0, action);
				set_field_buffer(cur_bar->service, 0, service);
				set_field_buffer(cur_bar->from, 0, from);
				set_field_buffer(cur_bar->to, 0, to);
				set_field_buffer(cur_bar->options, 0, options);

				free(option_str);

				draw_count++;

				rulebar_setcolor(cur_bar->active, cur_bar->num_field, cur_bar->action, cur_bar->service, cur_bar->from, cur_bar->to, cur_bar->options, COLOR_BLUE);

				d_node = d_node->next;
			}
			else
			{
				//vrprint.info("Debug", "draw_rules: dont draw: printable_lines: %d, draw_offset_down: %d", printable_lines, *draw_offset_down);
			}
		}
		else
		{
			filtered_rules++;
		}
	}

	/* clear the remaining bars (if any) */
	for(; draw_count < rbform->max_bars_on_screen && d_node; draw_count++, d_node = d_node->next)
	{
		if(!(cur_bar = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		set_field_buffer(cur_bar->active, 0, "");
		set_field_buffer(cur_bar->num_field, 0, "");
		set_field_buffer(cur_bar->action, 0, "");
		set_field_buffer(cur_bar->service, 0, "");
		set_field_buffer(cur_bar->from, 0, "");
		set_field_buffer(cur_bar->to, 0, "");
		set_field_buffer(cur_bar->options, 0, "");

		rulebar_setcolor(cur_bar->num_field, cur_bar->active, cur_bar->action, cur_bar->service, cur_bar->from, cur_bar->to, cur_bar->options, COLOR_BLUE);
	}

	return(0);
}

/*	rules_form

	Returncodes:
		0: ok
		-1: error
*/
int
rules_form(const int debuglvl, Rules *rules, Zones *zones, Interfaces *interfaces, Services *services, struct rgx_ *reg)
{
	WINDOW			*rules_win;
	PANEL			*panels[1];
	FIELD   		**fields;
	FORM			*form;
	int			n_fields,
				rows,
				cols,
				ch,
				quit = 0,
				i,
				x,
				bars = 0,
				field_y,
				field_x,
				field_bar_num,
				current_bar_num=1,
				rules_changed=0; // 1 if rules are changed, 0 if not

	struct RuleBarForm_	*rbform;
	rulebar			*cur_bar = NULL;

	int			max_height = 0,
				max_width = 0,
				height = 0,
				width = 0,
				startx = 0,
				starty = 0,
				pgdn_offset=0,

				insert_rule_num = 0;

	char			*filter_ptr,
				*filter_string_regex;

	d_list_node		*d_node = NULL;
	struct RuleData_	*query_ptr = NULL;

	int			retval = 0,
				result = 0;

	/* safety */
	if(!rules || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	
	// get the dimentions of the screen
	getmaxyx(stdscr, max_height, max_width);

	// set windowsize and start position
	height = max_height-6;
	width = max_width; // = minimum screensize - 2
	startx = 0;
	starty = 3;

	// max bars on the screen
	bars = max_height - 10;

	/* Create the window to be associated with the menu */
	if(!(rules_win = create_newwin(height, width, starty, startx, "Rules Section", COLOR_PAIR(CP_WHITE_BLUE))))
	{
		(void)vrprint.error(-1, "Error", "unable to create rules window.");
		return(-1);
	}
	panels[0] = new_panel(rules_win);
	keypad(rules_win, TRUE);

	/* malloc the rbform struct */
	if(!(rbform = malloc(sizeof(struct RuleBarForm_))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(-1);
	}
	/* now set it up */
	if(SetupRuleBarForm(debuglvl, rbform, bars, rules, width-4) < 0)
		return(-1);

	// calloc and create the fields
	//
	//
	n_fields = bars * FIELDS_PER_BAR;
	fields = (FIELD **)calloc(n_fields + 1, sizeof(FIELD *));

	for(i = 1, field_bar_num = 0, field_y = 2; i <= bars; i++, field_y++)
	{
		for(x = 1, field_x = 0; x <= FIELDS_PER_BAR; x++)
		{
			/* num field */
			if(x == 1)
			{
				fields[field_bar_num] = new_field(1, rbform->active_size, field_y, field_x, 0, 1);
				field_x = field_x+rbform->active_size;
			}
			/* active field */
			else if(x == 2)
			{
				fields[field_bar_num] = new_field(1, rbform->num_field_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->num_field_size;
			}
			/* action field */
			else if(x == 3)
			{
				fields[field_bar_num] = new_field(1, rbform->action_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->action_size;
			}
			// service field
			else if(x == 4)
			{
				fields[field_bar_num] = new_field(1, rbform->service_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->service_size;
			}
			// from field
			else if(x == 5)
			{
				fields[field_bar_num] = new_field(1, rbform->from_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->from_size;
			}
			// to field
			else if(x == 6)
			{
				fields[field_bar_num] = new_field(1, rbform->to_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->to_size;
			}
			// options field
			else
			{
				fields[field_bar_num] = new_field(1, rbform->options_size, field_y, field_x, 0, 1);
				field_x=field_x+rbform->options_size;
			}

			// only the first field active
			if(x > 1)
				field_opts_off(fields[field_bar_num], O_ACTIVE);

			field_bar_num++;
		}

		// create & insert bar		num			action				service				from			to			options
		if(Insert_RuleBar(debuglvl, rbform,	fields[field_bar_num-x+1],
							fields[field_bar_num-x+1+1],
							fields[field_bar_num-x+2+1],
							fields[field_bar_num-x+3+1],
							fields[field_bar_num-x+4+1],
							fields[field_bar_num-x+5+1],
							fields[field_bar_num-x+6+1]) < 0)
		{
			(void)vrprint.error(-1, "Error", "Insert_RuleBar failed");
			return(-1);
		}
	}
	fields[n_fields] = NULL;

	//
	// set field attr
	//
	for(i = 0; i < n_fields; i++)
	{
		// set field options
		set_field_back(fields[i], COLOR_PAIR(CP_WHITE_BLUE));
		field_opts_off(fields[i], O_AUTOSKIP);

		// set status to false
		set_field_status(fields[i], FALSE);
	}

	// Create the form and post it
	form = new_form(fields);
	// Calculate the area required for the form
	scale_form(form, &rows, &cols);
	// Set main window and sub window
	set_form_win(form, rules_win);
	set_form_sub(form, derwin(rules_win, rows, cols, 1, 2));
	post_form(form);

	// print labels, must be after the form is posted
	wattron(rules_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
	mvwprintw(rules_win, 1, 2+rbform->active_size, "Nr.");
	mvwprintw(rules_win, 1, 2+rbform->num_field_size+rbform->active_size, "Action");
	mvwprintw(rules_win, 1, 2+rbform->num_field_size+rbform->active_size+rbform->action_size, "Service");
	mvwprintw(rules_win, 1, 2+rbform->num_field_size+rbform->active_size+rbform->action_size+rbform->service_size, "Source");
	mvwprintw(rules_win, 1, 2+rbform->num_field_size+rbform->active_size+rbform->action_size+rbform->service_size+rbform->from_size, "Destination");
	mvwprintw(rules_win, 1, 2+rbform->num_field_size+rbform->active_size+rbform->action_size+rbform->service_size+rbform->from_size+rbform->to_size, "Options");
	wattroff(rules_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);

	// horizontal bar below the labels
	mvwhline(rules_win,  2, 1, ACS_HLINE, width-2);

	update_panels();
	doupdate();
	wrefresh(rules_win);

	status_print(status_win, "Ready.");

	while(quit == 0)
	{
		/* count the number of lines that are filtered */
		if(rbform->use_filter || rbform->show_only_input || rbform->show_only_output || rbform->show_only_forward)
		{
			rbform->filtered_rules = 0;

			for(d_node = rules->list.top; d_node; d_node = d_node->next)
			{
//TODO null check
				query_ptr = d_node->data;

				if((rbform->use_filter && !(MatchFilter_RuleBar(query_ptr, &rbform->filter_reg, rbform->show_only_input,
														rbform->show_only_output,
														rbform->show_only_forward))) ||
				  (!rbform->use_filter && !(MatchFilter_RuleBar(query_ptr, NULL, 		rbform->show_only_input,
														rbform->show_only_output,
														rbform->show_only_forward)))
				)
				{
					rbform->filtered_rules++;
				}
			}
		}
		else
			rbform->filtered_rules = 0;

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "filtered_rules: %d", rbform->filtered_rules);

		rbform->printable_rules = rules->list.len - rbform->filtered_rules;
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "printable_rules: %d (current_bar_num: %d)", rbform->printable_rules, current_bar_num);

		/* if we filter, position the bar at the top of the list if needed */
		if(	rbform->use_filter ||
			rbform->show_only_forward ||
			rbform->show_only_input ||
			rbform->show_only_output)
		{
			if(current_bar_num > rbform->printable_rules)
			{
				Set_RuleBar(rbform, form, 1);
				rbform->scroll_offset = 0;
			}
		}

		// set current bar
		cur_bar = CurrentBar(rbform, form);
		current_bar_num = cur_bar->bar_num;

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "current_bar_num: %d", current_bar_num);

		// print the rules
		if(draw_rules(debuglvl, rules, rbform) < 0)
			return(-1);


		// highlight the current bar
		HighlightRuleBar(cur_bar);

		// get user input
		ch = wgetch(rules_win);

		switch(ch)
		{
			/* to the bottom */
			case 360: /* end */
				if(rbform->printable_rules > rbform->max_bars_on_screen)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "360 (end): rbform->printable_rules > rbform->max_bars_on_screen (%d > %d).", rbform->printable_rules, rbform->max_bars_on_screen);

					rbform->scroll_offset = rbform->printable_rules - rbform->max_bars_on_screen;
					Set_RuleBar(rbform, form, rbform->max_bars_on_screen);
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "360 (end): rbform->printable_rules <= rbform->max_bars_on_screen (%d <= %d).", rbform->printable_rules, rbform->max_bars_on_screen);

					rbform->scroll_offset = 0;
					Set_RuleBar(rbform, form, rbform->printable_rules);
				}
				break;

			/* to the top */
			case 262: /* home key */
				rbform->scroll_offset = 0;

				Set_RuleBar(rbform, form, 1);
				break;

			/* half a page up */
			case 339: /* page up */
				for(i=0; i < (rbform->max_bars_on_screen/2); i++)
				{
					if(current_bar_num > 1)
					{
						form_driver(form, REQ_PREV_FIELD);

						current_bar_num--;
					}
					else
					{
						if(rbform->scroll_offset > 0)
							rbform->scroll_offset--;
					}
				}
				break;

			/* half a page down */
			case 338: /* page down */
				pgdn_offset = 0;

				for(i=0; i < (rbform->max_bars_on_screen/2); i++)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "338 (pgdn): current_bar_num : %d, rbform->max_bars_on_screen: %d, rbform->printable_rules: %d, rbform->scroll_offset: %d, atoi(field_buffer(cur_bar->num_field,0)): %d, pgdn_offset: %d.", current_bar_num, rbform->max_bars_on_screen, rbform->printable_rules, rbform->scroll_offset, atoi(field_buffer(cur_bar->num_field,0)), pgdn_offset);

					// make sure we dont move to the next field if we:
					//	1. scroll
					//	2. are at the end of a list that is shorter than the number of bars on screen
					if(current_bar_num < rbform->max_bars_on_screen && atoi(field_buffer(cur_bar->num_field,0)) + pgdn_offset < rbform->printable_rules)
					{
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "338 (pgdn): current_bar_num < rbform->max_bars_on_screen (%d < %d).", current_bar_num, rbform->max_bars_on_screen);
							(void)vrprint.debug("Debug", "338 (pgdn): current_bar_num < printable_rules (%d < %d).", current_bar_num, rbform->printable_rules);
						}

						form_driver(form, REQ_NEXT_FIELD);
						current_bar_num++;

						pgdn_offset++;
					}
					else if(current_bar_num == rbform->printable_rules || current_bar_num + rbform->scroll_offset == rbform->printable_rules)
					{
						// just do'in nothin'
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "338 (pgdn): current_bar_num == rbform->printable_rules (%d == %d), OR", current_bar_num, rbform->printable_rules);
							(void)vrprint.debug("Debug", "338 (pgdn): atoi(field_buffer(cur_bar->num_field,0)) + pgdn_offset == rbform->printable_rules (%d + %d == %d).", atoi(field_buffer(cur_bar->num_field,0)), pgdn_offset, rbform->printable_rules);
						}
					}
					else
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "338 (pgdn): rbform->scroll_offset: %d.", rbform->scroll_offset);

						rbform->scroll_offset++;
					}
				}

				break;

			/* one up */
			case KEY_UP:
				if(current_bar_num > 1)
					form_driver(form, REQ_PREV_FIELD);
				else
				{
					if(rbform->scroll_offset > 0)
						rbform->scroll_offset--;
				}

				form_driver(form, REQ_BEG_LINE);
				break;

			/* one down */
			case KEY_DOWN:
				// make sure we dont move to the next field if we:
				//	1. scroll
				//	2. are at the end of a list that is shorter than the number of bars on screen
				if(current_bar_num < rbform->max_bars_on_screen && current_bar_num < rbform->printable_rules)
				{
					if(debuglvl >= HIGH)
					{
						(void)vrprint.debug("Debug", "KEY_DOWN: current_bar_num < rbform->max_bars_on_screen (%d < %d).", current_bar_num, rbform->max_bars_on_screen);
						(void)vrprint.debug("Debug", "KEY_DOWN: current_bar_num < printable_rules (%d < %d).", current_bar_num, rbform->printable_rules);
					}

					form_driver(form, REQ_NEXT_FIELD);
				}
				else if(current_bar_num == rbform->printable_rules || atoi(field_buffer(cur_bar->num_field,0)) == rbform->printable_rules)
				{
					// do nothing, just sit here
					if(debuglvl >= HIGH)
					{
						(void)vrprint.debug("Debug", "KEY_DOWN: current_bar_num == printable_rules (%d == %d), OR", current_bar_num, rbform->printable_rules);
						(void)vrprint.debug("Debug", "KEY_DOWN: atoi(field_buffer(cur_bar->num_field,0)) == rbform->printable_rules (%d == %d)", atoi(field_buffer(cur_bar->num_field,0)), rbform->printable_rules);
					}
				}
				else
				{
					if(debuglvl >= HIGH)
					{
						(void)vrprint.debug("Debug", "KEY_DOWN: current_bar_num >= rbform->max_bars_on_screen (%d >= %d).", current_bar_num, rbform->max_bars_on_screen);
						(void)vrprint.debug("Debug", "KEY_DOWN: current_bar_num >= printable_rules (%d >= %d).", current_bar_num, rbform->printable_rules);
					}

					rbform->scroll_offset++;
				}

				form_driver(form, REQ_BEG_LINE);
				break;

			case 32: /* spacebar */

				Toggle_RuleBar(debuglvl, cur_bar, rules);
				rules_changed = 1;
				break;

			/* edit the rule */
			case 10: /* enter key */
			case KEY_RIGHT:
			case 'e':
			case 'E':

				draw_top_menu_edit_rule(top_win);

				result = Enter_RuleBar(debuglvl, cur_bar, rules, zones, interfaces, services, reg);
				if(result == 1)
				{
					rules_changed = 1;
				}
				/* we have removed the rule from the list, so now we need to make sure that
				   the screen is updated properly */
				else if(result == -1)
				{
					/*	if we remove the last rule in a none scrolling list make sure the bar is set to
						the last rule, otherwise we can scroll of the screen */
					if(current_bar_num > rbform->printable_rules - 1)
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "KEY_DC: current_bar_num > printable_rules - 1 (%d > %d).", current_bar_num, rbform->printable_rules - 1);
						
						form_driver(form, REQ_PREV_FIELD);
					}
					else
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "KEY_DC: current_bar_num <= printable_rules - 1 (%d <= %d).", current_bar_num, rbform->printable_rules - 1);
					}
				}

				draw_top_menu_rules(top_win);
				break;

			case KEY_DC: /* delete key */
			case 'd':
			case 'D':

				/* delete the rule */
//TODO detacht atoi part
				result = delete_rule(debuglvl, rules, atoi(field_buffer(cur_bar->num_field, 0)), 1);
				if(result == 0)
					status_print(status_win, "Delete rule cancelled.");
				else if(result == 1)
				{
					(void)vrprint.info("Info", "rule %d removed.", atoi(field_buffer(cur_bar->num_field, 0)));

					rules_changed = 1;

					/* decrease the scroll_offset so we don't scroll of the list */
					if(rbform->scroll_offset > 0)
					{
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "KEY_DC: scroll_offset > 0 (%d) decreasing.", rbform->scroll_offset);
						}
						rbform->scroll_offset--;
					}
					else
					{
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "KEY_DC: scroll_offset <= 0 (%d) doing nothing.", rbform->scroll_offset);
						}
					}

					/*
						if we remove the last rule in a none scrolling list make sure the bar is set to
						the last rule
					*/
					if(current_bar_num > rbform->printable_rules - 1)
					{
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "KEY_DC: current_bar_num > printable_rules - 1 (%d > %d).", current_bar_num, rbform->printable_rules - 1);
						}
						form_driver(form, REQ_PREV_FIELD);
					}
					else
					{
						if(debuglvl >= HIGH)
						{
							(void)vrprint.debug("Debug", "KEY_DC: current_bar_num <= printable_rules - 1 (%d <= %d).", current_bar_num, rbform->printable_rules - 1);
						}
					}
				}
				else
				{
					retval = -1;
					quit = 1;
				}

				break;

			/* insert a new rule */
			case KEY_IC: /* insert key */
			case 'i':
			case 'I':

				/* insert a new rule into the list */
				insert_rule_num = atoi(field_buffer(cur_bar->num_field, 0));
				if(insert_rule_num == 0)
					insert_rule_num = 1;

				if(insert_new_rule(debuglvl, rules, insert_rule_num) >= 0)
				{
					draw_top_menu_edit_rule(top_win);

					/* now edit the rule */
					if(edit_rule(debuglvl, rules, zones, interfaces, services, insert_rule_num, reg) < 0)
					{
						/* editting failed so remove the rule again */
						if(rules_remove_rule_from_list(debuglvl, rules, insert_rule_num, 1) < 0)
						{
							(void)vrprint.error(-1, "Internal Error", "removing rule failed (in: %s).", __FUNCTION__);
							return(-1);
						}
					}

					draw_top_menu_rules(top_win);
				}

				rules_changed = 1;
				break;

			/* move a rule */
			case 'm':
			case 'M':

//TODO: first atoi, then argument to func
				if(MoveRuleBarForm(debuglvl, rbform, rules, atoi(field_buffer(cur_bar->num_field, 0))) < 0)
					return(-1);

				rules_changed = 1;

				status_print(status_win, "Ready.");
				break;

			/* filter */
			case 'f':
			case 'F':

				if((filter_ptr = input_box(32, "Enter filter", "Set the filter (leave empty for no filter)")))
				{
					/* first clear the old regex (if we have one) */
					if(rbform->use_filter == 1)
					{
						regfree(&rbform->filter_reg);
					}

					/* first construct the regex string */
					if(!(filter_string_regex = malloc(sizeof(filter_ptr)+1+4)))
						return(-1);
					snprintf(filter_string_regex, (sizeof(filter_ptr)+1+4), ".*%s.*", filter_ptr);

					rbform->use_filter = 1;

					/* compiling the regex */
					if(regcomp(&rbform->filter_reg, filter_string_regex, REG_EXTENDED) != 0)
					{
						(void)vrprint.warning("Warning", "Setting up the regular expression with regcomp failed. Disabling filter.");
						rbform->use_filter = 0;
					}

					status_print(status_win, "Filter in place: '%s' (press 'f' and then just 'enter' to clear).", filter_ptr);

					free(filter_ptr);
					filter_ptr = NULL;
					free(filter_string_regex);
				}
				else
				{
					/* if a filter was in place, clear regex */
					if(rbform->use_filter == 1)
					{
						/* clear regex */
						regfree(&rbform->filter_reg);
					}
					rbform->use_filter = 0;

					status_print(status_win, "Filter removed.");
				}
				break;

			/* quit */
			case 27:
			case KEY_F(10):
			case 'q':
			case 'Q':
			case KEY_LEFT:

				quit = 1;
				break;

			case 'o':
			case 'O':

				if(rbform->show_only_forward == 0)
				{
					rbform->show_only_forward = 1;
					rbform->show_only_input   = 0;
					rbform->show_only_output  = 0;
				}
				else
					rbform->show_only_forward = 0;

				status_print(status_win, "Showing %s rules.", rbform->show_only_forward ? "only forward" : "all");
				break;

			case 'n':
			case 'N':

				if(rbform->show_only_input == 0)
				{
					rbform->show_only_forward = 0;
					rbform->show_only_input   = 1;
					rbform->show_only_output  = 0;
				}
				else
					rbform->show_only_input   = 0;

				status_print(status_win, "Showing %s rules.", rbform->show_only_input ? "only input" : "all");
				break;

			case 'u':
			case 'U':

				if(rbform->show_only_output == 0)
				{
					rbform->show_only_forward = 0;
					rbform->show_only_input   = 0;
					rbform->show_only_output  = 1;
				}
				else
					rbform->show_only_output  = 0;

				status_print(status_win, "Showing %s rules.", rbform->show_only_output ? "only output" : "all");
				break;

			case KEY_F(12):
			case 'h':
			case 'H':

				print_help(debuglvl, ":[VUURMUUR:RULES]:");
				break;
		}

		//cur_bar = CurrentBar(rbform, form);
	}

	// if the rules are changed, save the changes. But only if retval != -1.
	if(rules_changed && retval != -1)
	{
		if(rules_write_file(debuglvl, rules, conf.rules_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "writing rulesfile failed.");
			retval = -1;
		}
	}

	d_list_cleanup(debuglvl, &rbform->RuleBar_list);
	free(rbform);

	unpost_form(form);
	free_form(form);

	for(i = 0; i < n_fields; i++)
	{
		free_field(fields[i]);
	}
	free(fields);

	del_panel(panels[0]);
	delwin(rules_win);

	return(retval);
}

int
delete_rule(const int debuglvl, Rules *rules, int rule_num, int call_confirm)
{
	int	retval=0,
		result=0;
	int	remove_rule=0;

	if(call_confirm == 1)
	{
		/* first ask the user to confirm */
		result = confirm("Delete rule", "Are you sure?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0);
		if(result == 1)
			remove_rule = 1;
		else
			remove_rule = 0;
	}
	else
		remove_rule = 1;

	if(remove_rule == 1)
	{
		retval=1;

		/* okay, here we remove */
		if(rules_remove_rule_from_list(debuglvl, rules, rule_num, 1) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "removing rule failed (in: delete_rule).");
			retval = -1;
		}
	}
                
        if(debuglvl >= LOW)
		rules_print_list(rules);
                  
	return(retval);
}

//
int
insert_new_rule(const int debuglvl, Rules *rules, int rule_num)
{
	int			retval = 0;
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!rules)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "rule_num: %d", rule_num);

	/* inserting into an empty rules list */
	if(rule_num == 0)
		rule_num = 1;

	/* claim memory */
	if(!(rule_ptr = rule_malloc()))
		return(-1);

	/* set rule to standard */
	strcpy(rule_ptr->action, "Drop"); /* default we drop */
	strcpy(rule_ptr->service, "");
	strcpy(rule_ptr->from, "");
	strcpy(rule_ptr->to, "");
	rule_ptr->active = 1;
	
	/* only setup the options if we are going to change one or more */
	if(vccnf.newrule_log || vccnf.newrule_loglimit)
	{
		if(!(rule_ptr->opt = ruleoption_malloc(debuglvl)))
		{
			free(rule_ptr);
			return(-1);
		}

		/* default log and limit the log */
		rule_ptr->opt->rule_log = vccnf.newrule_log;
		rule_ptr->opt->loglimit = vccnf.newrule_loglimit;
		rule_ptr->opt->logburst = vccnf.newrule_logburst;
	}

	/* set the rule number */
	rule_ptr->number = rule_num;

	/* handle the rules list is empty */
	if(rules->list.len == 0)
	{
		/* insert at 1 of course */
		rule_ptr->number = 1;

		/* init the list */
		//if(d_list_setup(debuglvl, rules_list, NULL) < 0)
		//	return(-1);

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "** insert_new_rule: rule_num: %d, query_ptr->number: %d", rule_num, rule_ptr->number);

		if(rules_insert_list(debuglvl, rules, rule_ptr->number, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "failed to insert in list (in: %s:%d).", __FUNCTION__, __LINE__);
			retval = -1;
		}
	}
	/* handle in a non-empty list */
	else
	{
		if(rules_insert_list(debuglvl, rules, rule_num, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "failed to insert in list (in: %s:%d).", __FUNCTION__, __LINE__);
			retval=-1;
		}
	}


	if(debuglvl >= LOW)
		rules_print_list(rules);


	if(retval == 0)
		(void)vrprint.info("Info", "new rule inserted.");

	return(retval);
}


// returns 0: no change, or 1: change
int
edit_rule(const int debuglvl, Rules *rules, Zones *zones, Interfaces *interfaces, Services *services, int rule_num, struct rgx_ *reg)
{
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL;
	int			retval = 0;

	/* safety */
	if(!reg || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}
	
	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: called with rule_num = %d", __FUNCTION__, rule_num);

	if(rule_num == 0)
		rule_num = 1;

	if(rules->list.len == 0)
	{
		(void)vrprint.error(-1, "Internal Error", "list is empty (in: %s)", __FUNCTION__);
		return(-1);
	}
	
	/* go to rulenum in the rules list to get the rule_ptr */
	if(!(d_node = rules->list.top))
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s)", __FUNCTION__);
		return(-1);
	}

	/* look for the rule_ptr */
	for(; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d)", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(rule_ptr->number == rule_num)
			break;
	}

	if(rule_ptr != NULL)
	{
		if(strcasecmp(rule_ptr->action, "protect") == 0)
		{
			(void)vrprint.error(-1, "Internal Error", "edit_rule can no longer be used for editting protect rules (in: %s)", __FUNCTION__);
			return(-1);
		}
		else
		{
			retval = edit_rule_normal(debuglvl, zones, interfaces, services, rule_ptr, rule_num, reg);
		}
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Rule not found (in: edit_rule).");
		retval = -1;
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "returning retval = %d.", retval);

	return(retval);
}


struct RuleFlds_
{
	FIELD	*action_label_fld_ptr,
		*action_fld_ptr,
		*queue_label_fld_ptr,
		*queue_fld_ptr,
		*service_label_fld_ptr,
		*service_fld_ptr,
		*fromzone_label_fld_ptr,
		*fromzone_fld_ptr,
		*tozone_label_fld_ptr,
		*tozone_fld_ptr,
		*log_label_fld_ptr,
		*log_fld_ptr,
		*logprefix_label_fld_ptr,
		*logprefix_fld_ptr,
		*loglimit_label_fld_ptr,
		*loglimit_fld_ptr,
		*interface_label_fld_ptr,
		*interface_fld_ptr,
		*reject_label_fld_ptr,
		*reject_fld_ptr,
		*redirect_label_fld_ptr,
		*redirect_fld_ptr,
		*listen_label_fld_ptr,
		*listen_fld_ptr,
		*remote_label_fld_ptr,
		*remote_fld_ptr,
		*markiptstate_label_fld_ptr,
		*markiptstate_fld_ptr,
		*comment_label_fld_ptr,
		*comment_fld_ptr;
} RuleFlds;


/*	edit_rule_fields_to_rule

	Returncodes:
		 1: changes stored
		 0: no changes
		-1: error
*/
static int
edit_rule_fields_to_rule(const int debuglvl, FIELD **fields, int n_fields, struct RuleData_ *query_ptr, struct rgx_ *reg)
{
	int	i = 0,
		z = 0,
		retval = 0;

	if(!fields || !query_ptr || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check for changed fields */
	for(i = 0; i < n_fields; i++)
	{
		if(field_status(fields[i]) == TRUE)
		{
			if(fields[i] == RuleFlds.action_fld_ptr)
			{
				/* action */
				for(z = 0; field_buffer(fields[i], 0)[z] != ' '; z++)
				{
					query_ptr->action[z] = field_buffer(fields[i], 0)[z];
				}
				query_ptr->action[z] = '\0';

				if(!(copy_field2buf(	query_ptr->action,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->action))))
					return(-1);

				retval = 1;
			}
			else if(fields[i] == RuleFlds.service_fld_ptr)
			{
				/* service */
				if(!(copy_field2buf(	query_ptr->service,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->service))))
					return(-1);

				retval = 1;
			}
			else if(fields[i] == RuleFlds.fromzone_fld_ptr)
			{
				/* from */
				if(!(copy_field2buf(	query_ptr->from,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->from))))
					return(-1);

				retval = 1;
			}
			else if(fields[i] == RuleFlds.tozone_fld_ptr)
			{
				/* to */
				if(!(copy_field2buf(	query_ptr->to,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->to))))
					return(-1);

				retval = 1;
			}
			else if(fields[i] == RuleFlds.reject_fld_ptr)
			{
				/* option rejecttype */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(!(copy_field2buf(	query_ptr->opt->reject_type,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->opt->reject_type))))
					return(-1);

				if(strcmp(query_ptr->opt->reject_type, "") == 0)
					query_ptr->opt->reject_option = 0;
				else
					query_ptr->opt->reject_option = 1;

				retval = 1;
			}
			else if(fields[i] == RuleFlds.redirect_fld_ptr)
			{
				char	port_one[6];

				/* option redirect port */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(!(copy_field2buf(	port_one,
							field_buffer(fields[i], 0),
							sizeof(port_one))))
					return(-1);

				query_ptr->opt->redirectport = atoi(port_one);
				if(query_ptr->opt->redirectport <= 0 || query_ptr->opt->redirectport > 65535)
				{
					(void)vrprint.warning("Warning", "redirectport must be 1-65535.");
					query_ptr->opt->redirectport = 0;
				}

				retval = 1;
			}
			else if(fields[i] == RuleFlds.listen_fld_ptr)
			{
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				/* first clear the list */
				if(query_ptr->opt->listenport == 1 && query_ptr->opt->ListenportList.len > 0)
					d_list_cleanup(debuglvl, &query_ptr->opt->ListenportList);

				/* if the first char is a whitespace, we asume the field is empty */
				if(field_buffer(fields[i], 0)[0] == ' ')
				{
					query_ptr->opt->listenport = 0;
				}
				else
				{
					/* add the ports to the list */
					if(portopts_to_list(debuglvl, field_buffer(fields[i], 0), &query_ptr->opt->ListenportList) < 0)
						query_ptr->opt->listenport = 0;
					else
					{
						if(query_ptr->opt->ListenportList.len == 0)
							query_ptr->opt->listenport = 0;
						else
							query_ptr->opt->listenport = 1;
					}
				}

				retval=1;
			}
			else if(fields[i] == RuleFlds.remote_fld_ptr)
			{
				//status_print(status_win, "remoteport: '%s'", field_buffer(fields[i], 0));

				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				/* first clear the list */
				if(query_ptr->opt->remoteport == 1 && query_ptr->opt->RemoteportList.len > 0)
					d_list_cleanup(debuglvl, &query_ptr->opt->RemoteportList);

				/* if the first char is a whitespace, we asume the field is empty */
				if(field_buffer(fields[i], 0)[0] == ' ')
				{
					query_ptr->opt->remoteport = 0;
				}
				else
				{
					/* add the ports to the list */
					if(portopts_to_list(debuglvl, field_buffer(fields[i], 0), &query_ptr->opt->RemoteportList) < 0)
						query_ptr->opt->remoteport = 0;
					else
					{
						if(query_ptr->opt->RemoteportList.len == 0)
							query_ptr->opt->remoteport = 0;
						else
							query_ptr->opt->remoteport = 1;
					}
				}

				retval=1;
			}
			else if(fields[i] == RuleFlds.logprefix_fld_ptr)
			{
				int	last_char = 0;

				/* options */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				for(z = 0; z < 12; z++) /* 12 is max prefix length */
				{
					query_ptr->opt->logprefix[z] = field_buffer(fields[i], 0)[z];

					/* make sure that we place the NULL after the last char: no trailing spaces. */
					if(query_ptr->opt->logprefix[z] != ' ')
						last_char = z+1;
				}
				query_ptr->opt->logprefix[last_char] = '\0';

				if(strcmp(query_ptr->opt->logprefix, "") == 0)
					query_ptr->opt->rule_logprefix = 0;
				else
					query_ptr->opt->rule_logprefix = 1;

				retval = 1;
			}
			else if(fields[i] == RuleFlds.comment_fld_ptr)
			{
				int	last_char = 0;

				/* first check if the commentfield is valid */
				if(validate_commentfield(debuglvl, field_buffer(fields[i], 0), reg->comment) == 0)
				{
					/* options */
					if(query_ptr->opt == NULL)
					{
						if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
							return(-1);
						}
					}

					for(z = 0; z < sizeof(query_ptr->opt->comment) && field_buffer(fields[i], 0)[z] != '\n' && field_buffer(fields[i], 0)[z] != '\0'; z++) /* 12 is max prefix length */
					{
						query_ptr->opt->comment[z] = field_buffer(fields[i], 0)[z];

						/* make sure that we place the NULL after the last char: no trailing spaces. */
						if(query_ptr->opt->comment[z] != ' ')
							last_char = z + 1;
					}
					query_ptr->opt->comment[last_char] = '\0';

					if(strcmp(query_ptr->opt->comment, "") == 0)
						query_ptr->opt->rule_comment = 0;
					else
						query_ptr->opt->rule_comment = 1;

					retval = 1;
				}
			}
			else if(fields[i] == RuleFlds.loglimit_fld_ptr)
			{
				char	limit_str[4];

				/* option redirect port */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(!(copy_field2buf(	limit_str,
							field_buffer(fields[i], 0),
							sizeof(limit_str))))
					return(-1);

				query_ptr->opt->loglimit = atoi(limit_str);
				if(query_ptr->opt->loglimit < 0 || query_ptr->opt->loglimit > 999)
				{
					(void)vrprint.warning("Warning", "loglimit must be 0-999.");
					query_ptr->opt->loglimit = 0;
				}

				retval = 1;
			}
			else if(fields[i] == RuleFlds.log_fld_ptr)
			{
				/* log */

				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(strncmp(field_buffer(fields[i], 0), "X", 1) == 0)
					query_ptr->opt->rule_log = 1;
				else
					query_ptr->opt->rule_log = 0;

				retval=1;
			}
			else if(fields[i] == RuleFlds.markiptstate_fld_ptr)
			{
				/* markiptstate */

				/* if needed alloc the opt struct */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(strncmp(field_buffer(fields[i], 0), "X", 1) == 0)
					query_ptr->opt->markiptstate = 1;
				else
					query_ptr->opt->markiptstate = 0;

				retval=1;
			}
			else if(fields[i] == RuleFlds.queue_fld_ptr)
			{
				/* queue */

				/* if needed alloc the opt struct */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(strncmp(field_buffer(fields[i], 0), "X", 1) == 0)
					query_ptr->opt->queue = 1;
				else
					query_ptr->opt->queue = 0;

				retval=1;
			}
			else if(fields[i] == RuleFlds.interface_fld_ptr)
			{
				/* option interface */
				if(query_ptr->opt == NULL)
				{
					if(!(query_ptr->opt = ruleoption_malloc(debuglvl)))
					{
						(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
						return(-1);
					}
				}

				if(!(copy_field2buf(	query_ptr->opt->interface,
							field_buffer(fields[i], 0),
							sizeof(query_ptr->opt->interface))))
					return(-1);

				retval = 1;
			}
		}  
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "returning retval = %d.", retval);

	return(retval);
}


/*	edit_rule_normal

	Returncodes:
		 0: ok, no changes
		 1: ok, changes
		-1: error

	TODO: split this beast up
*/
int
edit_rule_normal(const int debuglvl, Zones *zones, Interfaces *interfaces, Services *services, struct RuleData_ *rule_ptr, int rule_num, struct rgx_ *reg)
{
	PANEL 			*my_panels[1];
	WINDOW			*edit_win;
	FIELD			**fields,
				*cur = NULL,
				*prev = NULL;

	FORM			*form;
	int			ch,
				rows,
				cols,
				i,
				retval=0,
				quit=0,
				not_defined=0;
	int			n_fields = 0,
				field_num = 0;
	char			redirect_port[6]   = "",
				loglimit_string[4] = "";
	int			height,
				width,
				startx,
				starty,
				max_height,
				max_width;
	char			*action_choices[] = {	"Accept",
							"Drop",
							"Reject",
							"Log",
							"Portfw",
							"Redirect",
							"Snat",
							"Masq",
							"Queue" },
				*action_ptr=NULL,
				*reject_types[] = {	"icmp-net-unreachable",
							"icmp-host-unreachable",
							"icmp-proto-unreachable",
							"icmp-port-unreachable",
							"icmp-net-prohibited",
							"icmp-host-prohibited",
							"tcp-reset" },
				*reject_ptr=NULL;
	int			action_choices_n = 9,
				reject_types_n = 7;
	char			**zone_choices,
				**choices,
				*choice_ptr,
				**service_choices;
	int			zone_choices_n=0,
				service_choices_n=0,
				n_choices = 0;
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL,
				*network_ptr = NULL;
	struct ServicesData_	*service_ptr = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	int			action_type=0;	// int for easy action manipulation
	
	int			result = 0;
	struct RuleCache_	tmp_ruledata;
	char			window_title[32] = "";

	/* is this screen in advanced mode or not? */
	char			advanced_mode = vccnf.advanced_mode;
	char			zonename[MAX_HOST_NET_ZONE] = "";

	/* safety */
	if(!rule_ptr || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* clear tmp_ruledata for the initial */
	memset(&tmp_ruledata, 0, sizeof(tmp_ruledata));
	memset(&RuleFlds, 0, sizeof(struct RuleFlds_));

	/* set to keep first */
	rule_ptr->status = ST_CHANGED;

	/* get the dimentions of the screen */
	getmaxyx(stdscr, max_height, max_width);

	/* set windowsize and start position */
	height = 18;
	width  = 78; /* = minimum screensize - 2 */
	startx = 3;
	starty = 1;

	/* init the action_type */
	action_type = actiontoi(rule_ptr->action);

	/* set number of fields */
	n_fields = 30;
	if(!(fields = (FIELD **)calloc(n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/*
		create the fields
	*/

	/* action label */
	RuleFlds.action_label_fld_ptr = (fields[field_num] = new_field(1, 8, 1, 1, 0, 0));
	set_field_buffer(RuleFlds.action_label_fld_ptr, 0, "Action");
	field_opts_off(RuleFlds.action_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.action_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.action_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* action */
	RuleFlds.action_fld_ptr = (fields[field_num] = new_field(1, 16, 1, 10, 0, 0));
	set_field_buffer(RuleFlds.action_fld_ptr, 0, rule_ptr->action);
	set_field_back(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* portfw/redirect queue label */
	RuleFlds.queue_label_fld_ptr = (fields[field_num] = new_field(1, 10, 3, 10, 0, 0));
	set_field_buffer(RuleFlds.queue_label_fld_ptr, 0, "Queue  [ ]");
	field_opts_off(RuleFlds.queue_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.queue_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.queue_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* portfw/redirect queue */
	RuleFlds.queue_fld_ptr = (fields[field_num] = new_field(1, 1, 3, 18, 0, 0));
	set_field_back(RuleFlds.queue_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.queue_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* enable */
	if(rule_ptr->opt != NULL && rule_ptr->opt->queue == 1)
		set_field_buffer(RuleFlds.queue_fld_ptr, 0, "X");

	/* queue starts disabled */
	field_opts_off(RuleFlds.queue_fld_ptr, O_VISIBLE);
	field_opts_off(RuleFlds.queue_label_fld_ptr, O_VISIBLE);


	/* service label */
	RuleFlds.service_label_fld_ptr = (fields[field_num] = new_field(1, 8, 5, 1, 0, 0));
	set_field_buffer(RuleFlds.service_label_fld_ptr, 0, "Service");
	field_opts_off(RuleFlds.service_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.service_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.service_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* service */
	RuleFlds.service_fld_ptr = (fields[field_num] = new_field(1, 32, 5, 10, 0, 0));
	set_field_buffer(RuleFlds.service_fld_ptr, 0, rule_ptr->service);
	set_field_back(RuleFlds.service_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.service_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* from zone label */
	RuleFlds.fromzone_label_fld_ptr = (fields[field_num] = new_field(1, 8, 9, 1, 0, 0));
	set_field_buffer(RuleFlds.fromzone_label_fld_ptr, 0, "From");
	field_opts_off(RuleFlds.fromzone_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.fromzone_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.fromzone_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* from zone */
	RuleFlds.fromzone_fld_ptr = (fields[field_num] = new_field(1, 48, 9, 10, 0, 0));
	set_field_buffer(RuleFlds.fromzone_fld_ptr, 0, rule_ptr->from);
	set_field_back(RuleFlds.fromzone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.fromzone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* to zone label */
	RuleFlds.tozone_label_fld_ptr = (fields[field_num] = new_field(1, 8, 11, 1, 0, 0));
	set_field_buffer(RuleFlds.tozone_label_fld_ptr, 0, "To");
	field_opts_off(RuleFlds.tozone_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.tozone_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.tozone_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* to zone */
	RuleFlds.tozone_fld_ptr = (fields[field_num] = new_field(1, 48, 11, 10, 0, 0));
	set_field_buffer(RuleFlds.tozone_fld_ptr, 0, rule_ptr->to);
	set_field_back(RuleFlds.tozone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.tozone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* comment label */
	RuleFlds.comment_label_fld_ptr = (fields[field_num] = new_field(1, 8, 14, 1, 0, 0));
	set_field_buffer(RuleFlds.comment_label_fld_ptr, 0, "Comment");
	field_opts_off(RuleFlds.comment_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.comment_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.comment_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* comment */
	RuleFlds.comment_fld_ptr = (fields[field_num] = new_field(1, 63, 14, 10, 0, 0));
	if(rule_ptr->opt != NULL && rule_ptr->opt->rule_comment == 1)
		set_field_buffer(RuleFlds.comment_fld_ptr, 0, rule_ptr->opt->comment);
	set_field_back(RuleFlds.comment_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.comment_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* log label */
	RuleFlds.log_label_fld_ptr = (fields[field_num] = new_field(1, 10, 1, 29, 0, 0));
	set_field_buffer(RuleFlds.log_label_fld_ptr, 0, "Log  [ ]");
	field_opts_off(RuleFlds.log_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.log_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.log_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* log */
	RuleFlds.log_fld_ptr = (fields[field_num] = new_field(1, 1, 1, 35, 0, 0));

	/* enable */
	if(rule_ptr->opt != NULL && rule_ptr->opt->rule_log == 1)
		set_field_buffer(RuleFlds.log_fld_ptr, 0, "X");

	set_field_back(RuleFlds.log_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.log_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;


	/* log prefix label */
	RuleFlds.logprefix_label_fld_ptr = (fields[field_num] = new_field(1, 8, 1, 39, 0, 0));
	set_field_buffer(RuleFlds.logprefix_label_fld_ptr, 0, "Prefix");
	field_opts_off(RuleFlds.logprefix_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.logprefix_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.logprefix_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* log prefix */
	RuleFlds.logprefix_fld_ptr = (fields[field_num] = new_field(1, 12, 1, 48, 0, 0));
	if(rule_ptr->opt != NULL)
		set_field_buffer(RuleFlds.logprefix_fld_ptr, 0, rule_ptr->opt->logprefix);
	set_field_back(RuleFlds.logprefix_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.logprefix_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* limit label */
	RuleFlds.loglimit_label_fld_ptr = (fields[field_num] = new_field(1, 8, 1, 62, 0, 0));
	set_field_buffer(RuleFlds.loglimit_label_fld_ptr, 0, "Limit");
	field_opts_off(RuleFlds.loglimit_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.loglimit_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.loglimit_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* limit */
	RuleFlds.loglimit_fld_ptr = (fields[field_num] = new_field(1, 3, 1, 70, 0, 0));
	if(rule_ptr->opt != NULL)
	{
		if(rule_ptr->opt->loglimit > 0)
		{
			snprintf(loglimit_string, sizeof(loglimit_string), "%d", rule_ptr->opt->loglimit);
			set_field_buffer(RuleFlds.loglimit_fld_ptr, 0, loglimit_string);
		}
	}
	set_field_back(RuleFlds.loglimit_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.loglimit_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;


	/* interface label */
	RuleFlds.interface_label_fld_ptr = (fields[field_num] = new_field(1, 9, 3, 29, 0, 0));
	set_field_buffer(RuleFlds.interface_label_fld_ptr, 0, "Interface");
	field_opts_off(RuleFlds.interface_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.interface_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.interface_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* interface */
	RuleFlds.interface_fld_ptr = (fields[field_num] = new_field(1, MAX_INTERFACE, 3, 40, 0, 0));
	if(rule_ptr->opt != NULL)
		set_field_buffer(RuleFlds.interface_fld_ptr, 0, rule_ptr->opt->interface);

	set_field_back(RuleFlds.interface_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.interface_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* interface starts disabled */
	field_opts_off(RuleFlds.interface_label_fld_ptr, O_VISIBLE);
	field_opts_off(RuleFlds.interface_fld_ptr, O_VISIBLE);


	/* Reject type label */
	RuleFlds.reject_label_fld_ptr = (fields[field_num] = new_field(1, 12, 3, 29, 0, 0));
	set_field_buffer(RuleFlds.reject_label_fld_ptr, 0, "Reject type");
	field_opts_off(RuleFlds.reject_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.reject_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.reject_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* Reject type */
	RuleFlds.reject_fld_ptr = (fields[field_num] = new_field(1, 23, 3, 48, 0, 0));

	if(rule_ptr->opt != NULL && rule_ptr->opt->reject_option == 1)
		set_field_buffer(RuleFlds.reject_fld_ptr, 0, rule_ptr->opt->reject_type);

	set_field_back(RuleFlds.reject_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.reject_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* reject starts disabled */
	field_opts_off(RuleFlds.reject_label_fld_ptr, O_VISIBLE);
	field_opts_off(RuleFlds.reject_fld_ptr, O_VISIBLE);


	/* Redirectport label */
	RuleFlds.redirect_label_fld_ptr = (fields[field_num] = new_field(1, 14, 5, 45, 0, 0));
	set_field_buffer(RuleFlds.redirect_label_fld_ptr, 0, "Redirect port");
	field_opts_off(RuleFlds.redirect_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.redirect_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.redirect_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* Redirectport */
	RuleFlds.redirect_fld_ptr = (fields[field_num] = new_field(1, 11, 5, 61, 0, 0));
	if(rule_ptr->opt != NULL && (rule_ptr->opt->redirectport > 0 && rule_ptr->opt->redirectport <= 65535))
	{
		snprintf(redirect_port, sizeof(redirect_port), "%d", rule_ptr->opt->redirectport);
		set_field_buffer(RuleFlds.redirect_fld_ptr, 0, redirect_port);
	}
	set_field_back(RuleFlds.redirect_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.redirect_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* start disabled */
	field_opts_off(RuleFlds.redirect_fld_ptr, O_VISIBLE);
	field_opts_off(RuleFlds.redirect_label_fld_ptr, O_VISIBLE);


	/* listenport (portfw) label */
	RuleFlds.listen_label_fld_ptr = (fields[field_num] = new_field(1, 12, 5, 45, 0, 0));
	set_field_buffer(RuleFlds.listen_label_fld_ptr, 0, "Listenport");
	field_opts_off(RuleFlds.listen_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.listen_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.listen_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* listenport */
	RuleFlds.listen_fld_ptr = (fields[field_num] = new_field(1, 14, 5, 58, 0, 0));
	set_field_back(RuleFlds.listen_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.listen_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* start disabled and set the field to dynamic */
	field_opts_off(RuleFlds.listen_fld_ptr, O_VISIBLE|O_STATIC);
	field_opts_off(RuleFlds.listen_label_fld_ptr, O_VISIBLE);

	/* this is needed after declaring the field dynamic */
	if(rule_ptr->opt != NULL && rule_ptr->opt->listenport == 1)
		set_field_buffer(RuleFlds.listen_fld_ptr, 0, list_to_portopts(debuglvl, &rule_ptr->opt->ListenportList, NULL));


	/* remoteport (portfw) label */
	RuleFlds.remote_label_fld_ptr = (fields[field_num] = new_field(1, 12, 7, 45, 0, 0));
	set_field_buffer(RuleFlds.remote_label_fld_ptr, 0, "Remoteport");
	field_opts_off(RuleFlds.remote_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.remote_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.remote_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* remoteport - total field size: 64 -> 50 offscreen */
	RuleFlds.remote_fld_ptr = (fields[field_num] = new_field(1, 14, 7, 58, 0, 0));
	set_field_back(RuleFlds.remote_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_fore(RuleFlds.remote_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
	field_num++;

	/* start disabled and set the field to dynamic */
	field_opts_off(RuleFlds.remote_fld_ptr, O_VISIBLE|O_STATIC);
	field_opts_off(RuleFlds.remote_label_fld_ptr, O_VISIBLE);

	/* this is needed after declaring the field dynamic */
	if(rule_ptr->opt != NULL && rule_ptr->opt->remoteport == 1)
		set_field_buffer(RuleFlds.remote_fld_ptr, 0, list_to_portopts(debuglvl, &rule_ptr->opt->RemoteportList, NULL));


	/* markiptstate (queue) label */
	RuleFlds.markiptstate_label_fld_ptr = (fields[field_num] = new_field(1, 22, 3, 29, 0, 0));
	set_field_buffer(RuleFlds.markiptstate_label_fld_ptr, 0, "Mark IptState      [ ]");
	field_opts_off(RuleFlds.markiptstate_label_fld_ptr, O_ACTIVE);
	set_field_back(RuleFlds.markiptstate_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.markiptstate_label_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* markiptstate */
	RuleFlds.markiptstate_fld_ptr = (fields[field_num] = new_field(1, 1, 3, 49, 0, 0));
	set_field_back(RuleFlds.markiptstate_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	set_field_fore(RuleFlds.markiptstate_fld_ptr, COLOR_PAIR(CP_BLUE_WHITE));
	field_num++;

	/* enable markiptstate option */
	if(rule_ptr->opt != NULL && rule_ptr->opt->markiptstate == 1)
		set_field_buffer(RuleFlds.markiptstate_fld_ptr, 0, "X");

	/* start disabled  */
	field_opts_off(RuleFlds.markiptstate_fld_ptr, O_VISIBLE);
	field_opts_off(RuleFlds.markiptstate_label_fld_ptr, O_VISIBLE);


	/* terminate the fields-array */
	fields[n_fields] = NULL;
	
	if(n_fields != field_num)
		(void)vrprint.error(-1, "Internal Error","oops! n_fields: %d, field_num: %d.", n_fields, field_num);

	/* create the window, panel, form */
	snprintf(window_title, sizeof(window_title), "Edit Rule: %d", rule_ptr->number);
	if(!(edit_win = create_newwin(height, width, startx, starty, window_title, COLOR_PAIR(CP_BLUE_WHITE))))
		return(-1);

	if(!(my_panels[0] = new_panel(edit_win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel.");
		return(-1);
	}
	keypad(edit_win, TRUE);

	if(!(form = new_form(fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form.");
		return(-1);
	}
	scale_form(form, &rows, &cols);
	set_form_win(form, edit_win);
	set_form_sub(form, derwin(edit_win, rows, cols, 1, 2));
	post_form(form);

	/* set cursor position */
	pos_form_cursor(form);

	update_panels();
	doupdate();

	/*
		loop through to get user requests
	*/
	while(quit == 0)
	{
		/* hide/disable fields we don't need */
		if(action_type != AT_REJECT)
		{
			field_opts_off(RuleFlds.reject_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.reject_label_fld_ptr, O_VISIBLE);
		}
		if(action_type != AT_REDIRECT)
		{
			field_opts_off(RuleFlds.redirect_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.redirect_label_fld_ptr, O_VISIBLE);
		}
		if((action_type != AT_REDIRECT && action_type != AT_PORTFW) || !advanced_mode)
		{
			field_opts_off(RuleFlds.interface_label_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.interface_fld_ptr, O_VISIBLE);

			field_opts_off(RuleFlds.queue_label_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.queue_fld_ptr, O_VISIBLE);
		}
		if(action_type != AT_PORTFW || !advanced_mode)
		{
			field_opts_off(RuleFlds.listen_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.listen_label_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.remote_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.remote_label_fld_ptr, O_VISIBLE);
		}
		if(action_type != AT_QUEUE || !advanced_mode)
		{
			field_opts_off(RuleFlds.markiptstate_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.markiptstate_label_fld_ptr, O_VISIBLE);
		}
		if(action_type == AT_LOG)
		{
			field_opts_off(RuleFlds.log_fld_ptr, O_ACTIVE);
		}
		if(!advanced_mode)
		{
			field_opts_off(RuleFlds.loglimit_fld_ptr, O_VISIBLE);
			field_opts_off(RuleFlds.loglimit_label_fld_ptr, O_VISIBLE);
		}

		/* show/enable fields we need */
		if(action_type == AT_REJECT)
		{
			field_opts_on(RuleFlds.reject_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.reject_label_fld_ptr, O_VISIBLE);
		}
		else if(action_type == AT_REDIRECT)
		{
			field_opts_on(RuleFlds.redirect_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.redirect_label_fld_ptr, O_VISIBLE);

			if(advanced_mode)
			{
				field_opts_on(RuleFlds.interface_label_fld_ptr, O_VISIBLE);
				field_opts_on(RuleFlds.interface_fld_ptr, O_VISIBLE);

				field_opts_on(RuleFlds.queue_label_fld_ptr, O_VISIBLE);
				field_opts_on(RuleFlds.queue_fld_ptr, O_VISIBLE);
			}
		}
		else if(action_type == AT_PORTFW && advanced_mode)
		{
			field_opts_on(RuleFlds.listen_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.listen_label_fld_ptr, O_VISIBLE);

			field_opts_on(RuleFlds.remote_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.remote_label_fld_ptr, O_VISIBLE);

			field_opts_on(RuleFlds.interface_label_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.interface_fld_ptr, O_VISIBLE);

			field_opts_on(RuleFlds.queue_label_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.queue_fld_ptr, O_VISIBLE);
		}
		else if(action_type == AT_QUEUE && advanced_mode)
		{
			/* first the label, because it overlaps the normal field */
			field_opts_on(RuleFlds.markiptstate_label_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.markiptstate_fld_ptr, O_VISIBLE);
		}

		if(action_type != AT_LOG)
		{
			field_opts_on(RuleFlds.log_fld_ptr, O_ACTIVE);
		}
		if(advanced_mode)
		{
			field_opts_on(RuleFlds.loglimit_fld_ptr, O_VISIBLE);
			field_opts_on(RuleFlds.loglimit_label_fld_ptr, O_VISIBLE);
		}

		/* do some nice coloring of the action field */
		if(action_type == AT_ACCEPT)
		{
			set_field_back(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_GREEN));
			set_field_fore(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_GREEN)|A_BOLD);
		}
		else if(action_type == AT_DROP || action_type == AT_REJECT)
		{
			set_field_back(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_RED));
			set_field_fore(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_RED)|A_BOLD);
		}
		else if(action_type == AT_LOG)
		{
			set_field_back(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
			set_field_fore(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		}
		else
		{
			set_field_back(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE));
			set_field_fore(RuleFlds.action_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);
		}

		/* color firewall zones */
		if(strncasecmp(field_buffer(RuleFlds.fromzone_fld_ptr, 0), "firewall", 8) == 0)
			set_field_fore(RuleFlds.fromzone_fld_ptr, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else
			set_field_fore(RuleFlds.fromzone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);

		if(strncasecmp(field_buffer(RuleFlds.tozone_fld_ptr, 0), "firewall", 8) == 0)
			set_field_fore(RuleFlds.tozone_fld_ptr, COLOR_PAIR(CP_YELLOW_BLUE)|A_BOLD);
		else
			set_field_fore(RuleFlds.tozone_fld_ptr, COLOR_PAIR(CP_WHITE_BLUE)|A_BOLD);

		prev = cur;
		cur = current_field(form);

		draw_field_active_mark(cur, prev, edit_win, form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		/*
			now give some help message in the status win
		*/
		if(cur == RuleFlds.action_fld_ptr)
			status_print(status_win, "Press SPACE to select an action.");
		else if(cur == RuleFlds.service_fld_ptr)
			status_print(status_win, "Press SPACE to select a service.");
		else if(cur == RuleFlds.fromzone_fld_ptr || cur == RuleFlds.tozone_fld_ptr)
			status_print(status_win, "Press SPACE to select a host, group or network, or the firewall.");
		else if(cur == RuleFlds.logprefix_fld_ptr)
			status_print(status_win, "Enter a text to be included in the log message.");
		else if(cur == RuleFlds.reject_fld_ptr)
			status_print(status_win, "Press SPACE to select a reject type.");
		else if(cur == RuleFlds.redirect_fld_ptr)
			status_print(status_win, "Enter a portnumber to redirect to.");
		else if(cur == RuleFlds.listen_fld_ptr)
			status_print(status_win, "Enter a comma-sepparated list of ports for the firewall to listen on.");
		else if(cur == RuleFlds.remote_fld_ptr)
			status_print(status_win, "Enter a comma-sepparated list of ports for the firewall to forward to.");
		else if(cur == RuleFlds.comment_fld_ptr)
			status_print(status_win, "Enter a optional comment.");
		else if(cur == RuleFlds.loglimit_fld_ptr)
			status_print(status_win, "Maximum number of logs per second (to prevent DoS), 0 for no limit.");
		else if(cur == RuleFlds.log_fld_ptr)
			status_print(status_win, "Press SPACE to toggle logging of this rule.");
		else if(cur == RuleFlds.markiptstate_fld_ptr)
			status_print(status_win, "Press SPACE to toggle marking the state of this rule (snort_inline).");
		else if(cur == RuleFlds.queue_fld_ptr)
			status_print(status_win, "Press SPACE to toggle queue'ing of this rule.");
		else if(cur == RuleFlds.interface_fld_ptr)
			status_print(status_win, "Press SPACE to select an interface to limit this rule to.");

		ch = wgetch(edit_win);
		not_defined = 0;

		if(	cur == RuleFlds.logprefix_fld_ptr ||
			cur == RuleFlds.redirect_fld_ptr  ||
			cur == RuleFlds.listen_fld_ptr    ||
			cur == RuleFlds.loglimit_fld_ptr  ||
			cur == RuleFlds.remote_fld_ptr    ||
			cur == RuleFlds.comment_fld_ptr)
		{
			if(nav_field_simpletext(form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == RuleFlds.markiptstate_fld_ptr  ||
			cur == RuleFlds.queue_fld_ptr         ||
			cur == RuleFlds.log_fld_ptr)
		{
			if(nav_field_toggleX(form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined == 1)
		{
			switch(ch)
			{
				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab

					form_driver(form, REQ_NEXT_FIELD);
					form_driver(form, REQ_END_LINE);
					break;

				case KEY_UP:

					form_driver(form, REQ_PREV_FIELD);
					form_driver(form, REQ_END_LINE);
					break;

				case 32: /* space */

					if(cur == RuleFlds.action_fld_ptr)
					{
						/* ask the user about the new action */
						if((action_ptr = selectbox("Action", "Select action", action_choices_n, action_choices, 1)))
						{
							set_field_buffer(cur, 0, action_ptr);
							action_type = actiontoi(action_ptr);
							free(action_ptr);
						}
					}
					else if(cur == RuleFlds.fromzone_fld_ptr || cur == RuleFlds.tozone_fld_ptr)
					{
						for(zone_choices_n = 0, d_node = zones->list.top; d_node; d_node = d_node->next)
						{
							if(!(zone_ptr = d_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "zone_ptr == NULL! (in: edit_rule_normal).");
								return(-1);
							}

							if(zone_ptr->type != TYPE_ZONE && zone_ptr->type != TYPE_FIREWALL)
							{
								zone_choices_n++;
							}
						}
						zone_choices_n++;

						if(!(zone_choices = calloc(zone_choices_n + 1, MAX_HOST_NET_ZONE)))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
							return(-1);
						}

						for(i = zone_choices_n - 1, d_node = zones->list.bot; d_node ; d_node = d_node->prev)
						{
							if(!(zone_ptr = d_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "zone_ptr == NULL! (in: edit_rule_normal).");
								return(-1);
							}

							if(zone_ptr->type != TYPE_ZONE && zone_ptr->type != TYPE_FIREWALL)
							{
								zone_choices[i] = zone_ptr->name;

								i--;
							}
						}
						zone_choices[0] = "firewall";

						/* get the zone */
						if((choice_ptr = selectbox("Select", "Select a host, group or network", zone_choices_n, zone_choices, 2)))
						{
							set_field_buffer(cur, 0, choice_ptr);
							free(choice_ptr);
							choice_ptr = NULL;
							
						}

						free(zone_choices);
						zone_choices = NULL;

					}
					else if(cur == RuleFlds.service_fld_ptr)
					{
						service_choices_n = services->list.len;

						if(!(service_choices = calloc(service_choices_n + 1, sizeof(char *))))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
							return(-1);
						}

						for(i = 0, d_node = services->list.top; d_node && i < service_choices_n; d_node = d_node->next, i++)
						{
							if(!(service_ptr = d_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "service_ptr == NULL! (in: edit_rule_normal).");
								return(-1);
							}

							service_choices[i] = service_ptr->name;
						}
						service_choices[i] = NULL;

						/* get the service */
						if((choice_ptr = selectbox("Select", "Select a service", service_choices_n, service_choices, 3)))
						{
							set_field_buffer(cur, 0, choice_ptr);
							free(choice_ptr);
							choice_ptr = NULL;
						}

						free(service_choices);
						service_choices = NULL;
					}
					else if(cur == RuleFlds.reject_fld_ptr)
					{
						if((reject_ptr = selectbox("Reject type", "Select reject type", reject_types_n, reject_types, 1)))
						{
							set_field_buffer(cur, 0, reject_ptr);
							free(reject_ptr);
						}
					}
					else if(cur == RuleFlds.interface_fld_ptr)
					{
						if(field_buffer(RuleFlds.fromzone_fld_ptr, 0)[0] == '\0' || field_buffer(RuleFlds.fromzone_fld_ptr, 0)[0] == ' ')
						{
							(void)vrprint.warning("Warning", "no from zone, please select one first.");
						}
						else if(strncasecmp(field_buffer(RuleFlds.fromzone_fld_ptr, 0), "firewall", 8) == 0)
						{
							(void)vrprint.warning("Warning", "the source of a portfw-rule cannot be 'firewall'.");
						}
						else
						{
							/* copy the from field to the zonename buffer */
							if(!(copy_field2buf(	zonename,
										field_buffer(RuleFlds.fromzone_fld_ptr, 0),
										sizeof(zonename))))
								return(-1);

							/* get the zone */
							if(!(zone_ptr = search_zonedata(debuglvl, zones, zonename)))
							{
								(void)vrprint.error(-1, "Error", "zone '%s' not found (in: %s:%d).", zonename, __FUNCTION__, __LINE__);
							}
							else
							{
								/* the interfaces are attached to the network, so get the network */
								if(zone_ptr->type == TYPE_NETWORK)
								{
									network_ptr = zone_ptr;
								}
								else if(zone_ptr->type == TYPE_HOST || zone_ptr->type == TYPE_GROUP)
								{
									network_ptr = zone_ptr->network_parent;
								}
								else
								{
									(void)vrprint.error(-1, "Internal Error", "wrong zone type '%d'  (in: %s:%d).", zone_ptr->type, __FUNCTION__, __LINE__);
									return(-1);
								}

								/* check if there are interfaces defined to choose from */
								n_choices = network_ptr->InterfaceList.len + 1;

								/* get some mem */
								if(!(choices = calloc(n_choices + 1, MAX_INTERFACE)))
								{
									(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
									return(-1);
								}

								/* load the interfaces */
								for(i = n_choices-1, d_node = network_ptr->InterfaceList.top; d_node ; d_node = d_node->next)
								{
									if(!(iface_ptr = d_node->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
										free(choices);
										return(-1);
									}

									choices[i] = iface_ptr->name;
									i--;
								}
								choices[i] = "Any";

								/* ask the user to select an interface */
								if(!(choice_ptr = selectbox("New interface", "Select an interface (Any for no filter)", n_choices, choices, 1)))
								{
									/* no choice */
								}
								else
								{
									/* any means empty the field */
									if(strcmp(choice_ptr, "Any") == 0)
										set_field_buffer(RuleFlds.interface_fld_ptr, 0, "");
									else
										set_field_buffer(RuleFlds.interface_fld_ptr, 0, choice_ptr);

									free(choice_ptr);
								}

								/* cleanup */
								free(choices);
							}
						}
					}
					else
					{
						form_driver(form, ch);
					}
					break;

				case KEY_LEFT:
				case 'q':
				case 'Q':
				case KEY_F(10):
				case 27:

					result = edit_rule_fields_to_rule(debuglvl, fields, n_fields, rule_ptr, reg);
					if(result == 1)
					{
						/* check */
						if(rules_analyze_rule(debuglvl, rule_ptr, &tmp_ruledata, services, zones, interfaces) < 0)
						{
							/* clear tmp_ruledata for the next use */
							bzero(&tmp_ruledata, sizeof(tmp_ruledata));

							/* ask the user if he/she want to look at the rule again */
							if(!(confirm("An error was detected in the rule.", "Do you want to look at it again? (no will delete the rule)", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 1)))
							{
								retval = -1;
								quit = 1;
							}
							else
							{
								/* we're not quiting yet! */
								quit = 0;
								retval = 0;
							}
						}
						else
						{
							quit = 1;
							retval = 1;
						}
					}
					else if(result == 0)
					{
						/* no change */
						quit = 1;
					}
					else
					{
						/* error */
						retval = -1;
					}
					
					break;

				case KEY_F(12):
				case 'h':
				case 'H':

					print_help(debuglvl, ":[VUURMUUR:RULES:EDIT]:");
					break;

				/* enable advanced mode */
				case KEY_F(5):
				case 'a':
				case 'A':

					if(!advanced_mode)
						advanced_mode = 1;
					else
						advanced_mode = 0;
					break;
			}
		}
	}

	/* Un post form and free the memory */
	unpost_form(form);
	free_form(form);
	for(i=0; i < n_fields; i++)
	{
		free_field(fields[i]);
	}
	free(fields);

	del_panel(my_panels[0]);
	destroy_win(edit_win);

	update_panels();
	doupdate();

	status_print(status_win, "Ready.");

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "returning retval = %d.", retval);

	return(retval);
}
