/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


struct ServicesSection_
{
	PANEL   *panel[1];
	WINDOW  *win;
	MENU    *menu;
	ITEM    **items;

	int     list_items;

	struct EditService_
	{
		WINDOW  *win;
		PANEL	*panel[1];

		FIELD   **fields;
		int	n_fields;
		FORM    *form;

		/* portrange list */
		MENU	*menu;
		ITEM	**items;
		int	n_items;

		d_list	item_list;
		d_list	item_number_list;
	
	} EditService;

	struct EditService_ EditServicePrt;
	
	char comment[241];

} ServicesSection;


/*	edit_serv_portranges_new_validate

	Validates the new portrange and inserts it into the list at the right place.
	
	Returncodes:
		0: ok
		-1: error
*/
static int
edit_serv_portranges_new_validate(const int debuglvl, struct ServicesData_ *ser_ptr, struct portdata *port_ptr)
{
	d_list_node		*d_node = NULL;
	struct portdata 	*portlist_ptr = NULL;
	int			insert_now = 0,
				insert_append = 0;

	/* safety */
	if(!port_ptr || !ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* if low and high are the same, only src is enough */
	if(port_ptr->src_low == port_ptr->src_high)
		port_ptr->src_high = 0;
	if(port_ptr->dst_low == port_ptr->dst_high)
	{
		if(port_ptr->protocol != 1)
			port_ptr->dst_high = 0;
	}

	/* check the protocol */
	if(port_ptr->protocol == 0 || port_ptr->protocol > 255)
	{
		(void)vrprint.error(-1, "Internal Error", "invalid protocol %d (in: %s).", port_ptr->protocol, __FUNCTION__);
		return(-1);
	}
	if(port_ptr->protocol != 1 && port_ptr->protocol != 6 && port_ptr->protocol != 17 && port_ptr->protocol != 47 && port_ptr->protocol != 50 && port_ptr->protocol != 51)
	{
		(void)vrprint.error(-1, "Internal Error", "unsupported protocol: %d. Currently supported protocols are: tcp (6), udp (17), icmp (1), ah (51), esp (50) and gre (47) (in: %s).", port_ptr->protocol, __FUNCTION__);
		return(-1);
	}

	/*	check the ports

		for tcp and udp
	*/
	if(port_ptr->protocol == 6 || port_ptr->protocol == 17)
	{
		/* no value is allowed to be higher than 65535, src_low and dst_low are not allowed to be smaller than 1 */
		if(	port_ptr->src_low > 65535 || port_ptr->src_high > 65535 ||
			port_ptr->dst_low > 65535 || port_ptr->dst_high > 65535 ||
			port_ptr->src_low < 1 || port_ptr->dst_low < 1
		)
		{
			/* this is an error because of wrong user input, so no function name */
			(void)vrprint.error(-1, "Error", "one of the ports is too low or too high. Valid port values for tcp and udp are 1-65535.");
			return(-1);
		}

		/* check if the ranges are ok */
		if(	(port_ptr->src_low > port_ptr->src_high && port_ptr->src_high > 0) ||
			(port_ptr->dst_low > port_ptr->dst_high && port_ptr->dst_high > 0)
		)
		{
			/* this is an error because of wrong user input, so no function name */
			(void)vrprint.error(-1, "Error", "Please make sure that the 'high'-port is actually higher than the 'low'-port.");
			return(-1);
		}
	}
	/*
		for icmp
	*/
	else if(port_ptr->protocol == 1)
	{
		if(port_ptr->dst_low > 255 || port_ptr->dst_high > 16)
		{
			/* this is an error because of wrong user input, so no function name */
			(void)vrprint.error(-1, "Error", "one of the values is too high. Valid icmp-types values are 1-255 (note that 41-255 are reserved). Valid icmp-codes are 0-16 (note that not all combinations of types and codes are valid. See http://www.iana.org/assignments/icmp-parameters for details).");
			return(-1);
		}
	}

	/* in an empty list we insert now */
	if(ser_ptr->PortrangeList.len == 0)
		insert_now = 1;
	else
	{
		/* else set the initial d_node */
		if(!(d_node = ser_ptr->PortrangeList.top))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}
	}

	/* now look for the place in the list to insert */
	while(!insert_now)
	{
		if(!(portlist_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* the order is: tcp, udp, icmp, gre */
		if(portlist_ptr->protocol == 17)
		{
			if(port_ptr->protocol == 6)
				insert_now = 1;
		}
		else if(portlist_ptr->protocol == 1)
		{
			if(port_ptr->protocol == 6)
				insert_now = 1;
			if(port_ptr->protocol == 17)
				insert_now = 1;
		}
		else if(portlist_ptr->protocol == 47)
		{
			if(port_ptr->protocol == 6)
				insert_now = 1;
			if(port_ptr->protocol == 17)
				insert_now = 1;
			if(port_ptr->protocol == 1)
				insert_now = 1;
			if(port_ptr->protocol == 47)
			{
				/* this is an error because of wrong user input, so no function name */
				(void)vrprint.error(-1, "Error", "only one GRE portrange is allowed.");
				return(-1);
			}
		}
		else if(portlist_ptr->protocol == 50)
		{
			if(port_ptr->protocol == 6)
				insert_now = 1;
			if(port_ptr->protocol == 17)
				insert_now = 1;
			if(port_ptr->protocol == 1)
				insert_now = 1;
			if(port_ptr->protocol == 47)
				insert_now = 1;
			if(port_ptr->protocol == 50)
			{
				/* this is an error because of wrong user input, so no function name */
				(void)vrprint.error(-1, "Error", "only one ESP portrange is allowed.");
				return(-1);
			}
		}
		else if(portlist_ptr->protocol == 51)
		{
			if(port_ptr->protocol == 6)
				insert_now = 1;
			if(port_ptr->protocol == 17)
				insert_now = 1;
			if(port_ptr->protocol == 1)
				insert_now = 1;
			if(port_ptr->protocol == 47)
				insert_now = 1;
			if(port_ptr->protocol == 50)
				insert_now = 1;
			if(port_ptr->protocol == 51)
			{
				/* this is an error because of wrong user input, so no function name */
				(void)vrprint.error(-1, "Error", "only one AH portrange is allowed.");
				return(-1);
			}
		}

		if(!insert_now)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "edit_serv_portranges_new_validate: don't insert at this run.");

			if(d_list_node_is_bot(debuglvl, d_node))
			{
				/* if we reach the bot, insert now */
				insert_now = 1;
				insert_append = 1;
				break;
			}
			else
				d_node = d_node->next;
		}
	}

	/*
		insert now
	*/
	if(insert_now)
	{
		/*	for appending at the bot we call d_list_append because d_list_insert_before is unable to
			insert at the bot.
		*/
		if(!insert_append)
		{
			if(d_list_insert_before(debuglvl, &ser_ptr->PortrangeList, d_node, port_ptr) == NULL)
			{
				(void)vrprint.error(-1, "Internal Error",  "inserting port_ptr into list failed.");
				return(-1);
			}
		}
		else
		{
			if(d_list_append(debuglvl, &ser_ptr->PortrangeList, port_ptr) == NULL)
			{
				(void)vrprint.error(-1, "Internal Error",  "appending port_ptr into list failed.");
				return(-1);
			}
		}

		ser_ptr->status = ST_CHANGED;

		/* save the portranges */
		if(services_save_portranges(debuglvl, ser_ptr) < 0)
		{
			(void)vrprint.error(-1, "Error", "saving the portranges failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}

	return(0);
}


static int
edit_serv_portranges_new_tcpudp(const int debuglvl, struct portdata *port_ptr)
{
	WINDOW		*new_portrange_win;
	PANEL		*my_panels[1];
	FIELD		**fields;
	FORM		*my_form;
	int		height,
			width,
			startx = 0,
			starty = 0,
			max_height,
			max_width,
			ch,
			i,
			rows,
			cols,
			retval=0,
			quit=0;
	int		not_defined = 0;

	/* set window dimentions */
	height = 16;
	width = 34;

	getmaxyx(stdscr, max_height, max_width);

	/* place in the center of the screen */
	starty = (max_height - height) / 2;
	startx = (max_width  - width)  / 2;

	/* create window and panel */
	if(port_ptr->protocol == 6)
		new_portrange_win = create_newwin(height, width, starty, startx, "New TCP Portrange", COLOR_PAIR(CP_BLUE_WHITE));
	else
		new_portrange_win = create_newwin(height, width, starty, startx, "New UDP Portrange", COLOR_PAIR(CP_BLUE_WHITE));
	if(!new_portrange_win)
	{
		(void)vrprint.error(-1, "Error", "unable to create window.");
		return(-1);
	}

	if(!(my_panels[0] = new_panel(new_portrange_win)))
	{
		(void)vrprint.error(-1, "Error", "unable to create panel.");
		destroy_win(new_portrange_win);
		return(-1);
	}

	keypad(new_portrange_win, TRUE);

	fields = (FIELD **)calloc(4 + 1, sizeof(FIELD *));

	fields[0] = new_field(1, 5, 3, 0, 0, 0);
	set_field_buffer(fields[0], 0, "1024");
	fields[1] = new_field(1, 5, 3, 6, 0, 0);
	set_field_buffer(fields[1], 0, "65535");
	fields[2] = new_field(1, 5, 3, 14, 0, 0);
	fields[3] = new_field(1, 5, 3, 20, 0, 0);

	for(i = 0; i < 4; i++)
	{
		set_field_back(fields[i], COLOR_PAIR(3));
		field_opts_off(fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(fields[i], FALSE);
	}

	/* create form */
	my_form = new_form(fields);
	scale_form(my_form, &rows, &cols);
	set_form_win(my_form, new_portrange_win);
	set_form_sub(my_form, derwin(new_portrange_win, rows, cols, 1, 2));
	post_form(my_form);

	mvwprintw(new_portrange_win, 2, 2,"Source        Destination");
	mvwprintw(new_portrange_win, 3, 2,"Low   High    Low   High");

	mvwprintw(new_portrange_win, 6, 8,"<F10> to quit");
	mvwprintw(new_portrange_win, 8, 2,"If you need only one port use");
	mvwprintw(new_portrange_win, 9, 2,"the low port. In that case you");
	mvwprintw(new_portrange_win, 10, 2,"can leave the high port empty.");
	mvwprintw(new_portrange_win, 13, 2,"Valid values:");
	mvwprintw(new_portrange_win, 14, 2," 1 - 65535");

	pos_form_cursor(my_form);

	/* go to the 3rd field, so we focus on dst_low */
	form_driver(my_form, REQ_NEXT_FIELD);
	form_driver(my_form, REQ_NEXT_FIELD);

	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(new_portrange_win);

		not_defined = 0;

		if(nav_field_simpletext(my_form, ch) < 0)
			not_defined = 1;

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case 'q':
				case 'Q':

					quit = 1;
					break;

				case KEY_DOWN:
				case 10: // enter

					form_driver(my_form, REQ_NEXT_FIELD);
					form_driver(my_form, REQ_END_LINE);
					break;
				case KEY_UP:
					// Go to previous field
					form_driver(my_form, REQ_PREV_FIELD);
					form_driver(my_form, REQ_END_LINE);
					break;

				case KEY_BACKSPACE:
					form_driver(my_form, REQ_PREV_CHAR);
					form_driver(my_form, REQ_DEL_CHAR);
					form_driver(my_form, REQ_END_LINE);
					break;

				case 127:
				case KEY_DC:
					form_driver(my_form, REQ_PREV_CHAR);
					form_driver(my_form, REQ_DEL_CHAR);
					form_driver(my_form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:SERVICE:EDIT:PORTRANGE:TCPUDP]:");
					break;

				default:
					// If this is a normal character, it gets printed
					form_driver(my_form, ch);
					break;
			}
		}
	}

	/* store input in pointer */
	port_ptr->src_low  = atoi(field_buffer(fields[0], 0));
	port_ptr->src_high = atoi(field_buffer(fields[1], 0));
	port_ptr->dst_low  = atoi(field_buffer(fields[2], 0));
	port_ptr->dst_high = atoi(field_buffer(fields[3], 0));

	/* cleanup */
	unpost_form(my_form);
	free_form(my_form);

	for(i = 0; i < 4; i++)
	{
		free_field(fields[i]);
	}
	free(fields);

	del_panel(my_panels[0]);
	destroy_win(new_portrange_win);

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_serv_portranges_new_icmp_choose_type(void)
{
	WINDOW	*win;
	PANEL	*panel[1];
	MENU	*menu;
	ITEM	**items;
	int	n_items;

	int	i,
		retval=0,
		height=0,
		width=0,
		max_height=0,
		max_width=0,
		startx=0,
		starty=0,
		quit=0,
		ch=0;

	char	**itemnames;
	char	**itemnumbers;
	char	*name = NULL,
		*select_ptr;
		
	size_t	name_size = 32;	/* max size of icmp name */

	int	icmp_type=0,
		icmp_type_has_code=0,
		icmp_type_num=0,
		type_cnt=0;

	// get screensize
	getmaxyx(stdscr, max_height, max_width);

	// count the number of icmp types (maybe this could be a fixed number?)
	while(list_icmp_types(&icmp_type, &icmp_type_has_code, &icmp_type_num) == 1)
		type_cnt++;

	// set number of menu items
	n_items = type_cnt;

	// get memory
	if(!(items = (ITEM **)calloc(n_items + 1, sizeof(ITEM *))))
		return(-1);
	if(!(itemnames = calloc(n_items + 1, 32)))
		return(-1);
	if(!(itemnumbers = calloc(n_items + 1, 32)))
		return(-1);

	// reset
	icmp_type = 0;
	icmp_type_has_code = 0;
	icmp_type_num = 0;
	type_cnt = 0;

	// init
	while(list_icmp_types(&icmp_type, &icmp_type_has_code, &icmp_type_num) == 1)
	{
		status_print(status_win, "%d", type_cnt);

		// set itemnames and itemnumbers
		if(!(name = malloc(name_size)))
			return(-1);

		if(get_icmp_name_short(icmp_type, -1, name, name_size, 0) < 0)
			return(-1);
		itemnames[type_cnt] = name;

		if(!(name = malloc(name_size)))
			return(-1);

		snprintf(name, name_size, "%d", icmp_type);
		itemnumbers[type_cnt] = name;

		items[type_cnt] = new_item(itemnumbers[type_cnt], itemnames[type_cnt]);

		//status_print(status_win, "%s, %s, %d", itemnumbers[type_cnt], itemnames[type_cnt], type_cnt);

		// update cnt
		type_cnt++;
	}
	items[n_items] = (ITEM *)NULL;
	menu = new_menu((ITEM **)items);

	// set window dimentions
	height = n_items+4;
	if(height > (max_height - 6))
		height = max_height - 6;

	width = 32+8+2;

	startx = (max_width -  width )/2;
	starty = (max_height - height)/2;

	win = newwin(height, width, starty, startx);
	wbkgd(win, COLOR_PAIR(5));
	keypad(win, TRUE);
	box(win, 0, 0);
	print_in_middle(win, 1, 0, width, "ICMP Types", COLOR_PAIR(5));
	wrefresh(win);

	panel[0] = new_panel(win);
	update_panels();

	set_menu_win(menu, win);
	set_menu_sub(menu, derwin(win, height-4, width-2, 3, 1));

	set_menu_format(menu, height-4, 1);

	mvwaddch(win, 2, 0, ACS_LTEE);
	mvwhline(win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(win, 2, width-1, ACS_RTEE);

	set_menu_back(menu, COLOR_PAIR(5));
	set_menu_fore(menu, COLOR_PAIR(3));

	post_menu(menu);
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(win);
		switch(ch)
		{
			case KEY_DOWN:
				menu_driver(menu, REQ_DOWN_ITEM);
				break;

			case KEY_UP:
				menu_driver(menu, REQ_UP_ITEM);
				break;

			case KEY_RIGHT:
			case 32: // space
			case 10: // enter
			{
				ITEM *cur;
				cur = current_item(menu);
				// select
				select_ptr = malloc(1+strlen((char *)item_name(cur)));
				if(!select_ptr)
					return(-1);
				
				strcpy(select_ptr, item_name(cur));
				retval = atoi(select_ptr);
				free(select_ptr);

				// quit
				quit=1;
				break;
			}

			case 27:
			case KEY_LEFT:
			case KEY_F(10):
			case 'q':
			case 'Q':

				quit=1;
				retval=-1;
				break;

		}
	}

	// cleanup
	unpost_menu(menu);
	free_menu(menu);
	for(i = 0; i < n_items; ++i)
		free_item(items[i]);

	free(items);
	free(itemnumbers);
	free(itemnames);

	del_panel(panel[0]);

	destroy_win(win);

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_serv_portranges_new_icmp_choose_code(int icmp_type)
{
	WINDOW	*win;
	PANEL	*panel[1];
	MENU	*menu;
	ITEM	**items;
	int	n_items;

	int	i,
		retval=0,
		height=0,
		width=0,
		max_height=0,
		max_width=0,
		startx=0,
		starty=0,
		quit=0,
		ch=0;

	char	**itemnames;
	char	**itemnumbers;
	char	*name = NULL,
		*select_ptr;
		
	size_t	name_size = 32;

	int	icmp_code=0,
		icmp_code_num=0,
		code_cnt=0;

	/* get screensize */
	getmaxyx(stdscr, max_height, max_width);
	
	/* count the number of icmp types (maybe this could be a fixed number?) */
	while(list_icmp_codes(icmp_type, &icmp_code, &icmp_code_num) == 1)
		code_cnt++;

	if(code_cnt == 0)
	{
		(void)vrprint.warning("Warning", "No ICMP Codes with selected ICMP Type (%d).", icmp_type);
		return(0);
	}

	/* set number of menu items */
	n_items = code_cnt;

	// get memory
	items = (ITEM **)calloc(n_items + 1, sizeof(ITEM *));
	if(!items) return(-2);
	itemnames = calloc(n_items + 1, 32);
	if(!itemnames) return(-2);
	itemnumbers = calloc(n_items + 1, 32);
	if(!itemnumbers) return(-2);

	// reset
	icmp_code_num = 0;
	code_cnt = 0;
	icmp_code = 0;

	// init
	while(list_icmp_codes(icmp_type, &icmp_code, &icmp_code_num) == 1)
	{
		/* set itemnames and itemnumbers */
		if(!(name = malloc(name_size)))
			return(-2);

		get_icmp_name_short(icmp_type, icmp_code, name, name_size, 1);
		itemnames[code_cnt] = name;

		if(!(name = malloc(name_size)))
			return(-1);

		snprintf(name, name_size, "%d", icmp_code);
		itemnumbers[code_cnt] = name;

		items[code_cnt] = new_item(itemnumbers[code_cnt], itemnames[code_cnt]);

		/* update cnt */
		code_cnt++;
	}
	items[n_items] = (ITEM *)NULL;
	menu = new_menu((ITEM **)items);

	// set window dimentions
	height = n_items+4;
	if(height > (max_height - 6))
		height = max_height - 6;
		
	width = 32+8+2;

	startx = (max_width - width ) /2;
	starty = (max_height - height)/2;

	// create win & panel & set attribs
	win = newwin(height, width, starty, startx);
	panel[0] = new_panel(win);
	box(win, 0, 0);
	wbkgd(win, COLOR_PAIR(5));
	keypad(win, TRUE);
	print_in_middle(win, 1, 0, width, "ICMP Codes", COLOR_PAIR(5));
	wrefresh(win);
	update_panels();

	set_menu_win(menu, win);
	set_menu_sub(menu, derwin(win, height-4, width-2, 3, 1));
	set_menu_format(menu, height-4, 1);

	mvwaddch(win, 2, 0, ACS_LTEE);
	mvwhline(win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(win, 2, width-1, ACS_RTEE);

	set_menu_back(menu, COLOR_PAIR(5));
	set_menu_fore(menu, COLOR_PAIR(3));

	post_menu(menu);
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(win);
		switch(ch)
		{
			case KEY_DOWN:
				menu_driver(menu, REQ_DOWN_ITEM);
				break;
			case KEY_UP:
				menu_driver(menu, REQ_UP_ITEM);
				break;
			case KEY_RIGHT:
			case 32: // space
			case 10: // enter
			{
				ITEM *cur;
				cur = current_item(menu);
				// select
				select_ptr = malloc(1+strlen((char *)item_name(cur)));
				if(!select_ptr) return(-2);

				strcpy(select_ptr, item_name(cur));
				retval = atoi(select_ptr);
				free(select_ptr);
				// quit
				quit=1;
				break;
			}
			case 27:
			case KEY_LEFT:
			case KEY_F(10):
			case 'q':
			case 'Q':
				quit=1;
				retval=-1;
				break;
		}
	}

	// cleanup
	unpost_menu(menu);
	free_menu(menu);
	for(i = 0; i < n_items; ++i)
		free_item(items[i]);

	free(items);
	free(itemnumbers);
	free(itemnames);

	del_panel(panel[0]);

	destroy_win(win);

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_serv_portranges_new_icmp(const int debuglvl, struct portdata *port_ptr)
{
	WINDOW		*new_portrange_win;
	PANEL		*my_panels[1];
	FIELD		**fields;
	FORM		*my_form;
	int		height,
			width,
			startx,
			starty,
			max_height,
			max_width,
			ch,
			i,
			j,
			k,
			l,
			rows,
			cols,
			result = 0,
			retval = 0,
			quit = 0;
	char		icmp_type[4], icmp_code[4];
	FIELD		*cur_field, *type_field;

	// set window dimentions
	height = 15;
	width  = 34;
	startx = 0;
	starty = 0;

	getmaxyx(stdscr, max_height, max_width);

	// place in the center of the screen
	starty = (max_height - height) / 2;
	startx = (max_width - width) / 2;
	
	// create window and panel
	new_portrange_win = create_newwin(height, width, starty, startx, "New ICMP Portrange", COLOR_PAIR(5));
	my_panels[0] = new_panel(new_portrange_win);
	keypad(new_portrange_win, TRUE);

	fields = (FIELD **)calloc(2 + 1, sizeof(FIELD *));

	i=0, j=0, k=0, l=3;
	fields[0] = new_field(1, 3, 2, 9, 0, 1);
	set_field_buffer(fields[0], 1, "typ");
	fields[1] = new_field(1, 3, 2, 16, 0, 1);
	set_field_buffer(fields[1], 1, "cod");

	for(i = 0; i < 2; i++)
	{
		set_field_back(fields[i], COLOR_PAIR(3));
		field_opts_off(fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(fields[i], FALSE);
	}

	// create form
	my_form = new_form(fields);
	scale_form(my_form, &rows, &cols);
	set_form_win(my_form, new_portrange_win);
	set_form_sub(my_form, derwin(new_portrange_win, rows, cols, 1, 2));
	post_form(my_form);

	mvwprintw(new_portrange_win, 2, 10, "Type");
	mvwprintw(new_portrange_win, 2, 17, "Code");

	mvwprintw(new_portrange_win, 5, 3,  "<SPACE> to select type/code");
	mvwprintw(new_portrange_win, 7, 5,  "If you want for example");
	mvwprintw(new_portrange_win, 9, 5,  "Destination unreachable:");
	mvwprintw(new_portrange_win, 10, 5, "port unreachable use:");
	mvwprintw(new_portrange_win, 11, 5, "type: 3, and code: 3");
	mvwprintw(new_portrange_win, 13, 8,  "<F10> to quit");
	
	pos_form_cursor(my_form);

	update_panels();
	doupdate();

	while(quit == 0) // F10 exits
	{
		cur_field = current_field(my_form);

		ch = wgetch(new_portrange_win);

		switch(ch)
		{
			case 32: //space

				if(strcmp(field_buffer(cur_field, 1), "typ") == 0)
				{
					result = edit_serv_portranges_new_icmp_choose_type();
					if(result >= 0)
					{
						snprintf(icmp_type, sizeof(icmp_type), "%d", result);
						set_field_buffer(cur_field, 0, icmp_type);
					}
				}
				else
				{
					type_field = fields[0];
					strcpy(icmp_type, field_buffer(type_field, 0));

					result = edit_serv_portranges_new_icmp_choose_code(atoi(icmp_type));
					if(result >= 0)
					{
						snprintf(icmp_code, sizeof(icmp_code), "%d", result);
						set_field_buffer(cur_field, 0, icmp_code);
					}
				}
				break;

			case KEY_DOWN:
			case 10:	// enter
			case 9:		// tab
				// Go to next field
				form_driver(my_form, REQ_NEXT_FIELD);
				// Go to the end of the present buffer
				// Leaves nicely at the last character
				form_driver(my_form, REQ_END_LINE);
				break;
			
			case KEY_UP:
				// Go to previous field
				form_driver(my_form, REQ_PREV_FIELD);
				form_driver(my_form, REQ_END_LINE);
				break;
			
			case KEY_BACKSPACE:
				form_driver(my_form, REQ_PREV_CHAR);
				form_driver(my_form, REQ_DEL_CHAR);
				form_driver(my_form, REQ_END_LINE);
				break;
			
			case 127:
			case KEY_DC:
				form_driver(my_form, REQ_PREV_CHAR);
				form_driver(my_form, REQ_DEL_CHAR);
				form_driver(my_form, REQ_END_LINE);
				break;
			

			case 27:
			case KEY_LEFT:
			case KEY_F(10):
			case 'q':
			case 'Q':
				quit=1;
				break;

			case KEY_F(12):
			case 'h':
			case 'H':
				print_help(debuglvl, ":[VUURMUUR:SERVICE:EDIT:PORTRANGE:ICMP]:");
				break;

			default:
				// If this is a normal character, it gets printed
				form_driver(my_form, ch);
				break;
		}
	}

	// store input in pointer
	port_ptr->src_low = 0;
	port_ptr->src_high = 0;
	port_ptr->dst_low = atoi(field_buffer(fields[0], 0));
	port_ptr->dst_high = atoi(field_buffer(fields[1], 0));

	// cleanup
	unpost_form(my_form);
	free_form(my_form);

	for(i = 0; i < 2; i++)
	{
		free_field(fields[i]);
	}
	free(fields);

	del_panel(my_panels[0]);
	destroy_win(new_portrange_win);

	update_panels();
	doupdate();

	return(retval);
}

static int
edit_serv_portranges_new(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int		retval=0;

	/* select protocol setup */
	char		*choice_ptr = NULL,
			*choices[]= { "TCP", "UDP", "ICMP", "GRE", "AH", "ESP" };
	int		n_choices = 6;

	struct portdata	*port_ptr = NULL;

	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* alloc a new portrange */
	if(!(port_ptr = malloc(sizeof(struct portdata))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/*
		get the new portrange protocol
	*/
	choice_ptr = selectbox("New portrange", "Select a Protocol", n_choices, choices, 1);
	if(choice_ptr != NULL)
	{
		if(strncmp(choice_ptr, "TCP", 3) == 0)
		{
			port_ptr->protocol = 6;

			if(edit_serv_portranges_new_tcpudp(debuglvl, port_ptr) < 0)
				retval = -1;
		}
		else if(strncmp(choice_ptr, "UDP", 3) == 0)
		{
			port_ptr->protocol = 17;

			if(edit_serv_portranges_new_tcpudp(debuglvl, port_ptr) < 0)
				retval = -1;
		}
		else if(strncmp(choice_ptr, "ICMP", 4) == 0)
		{
			port_ptr->protocol = 1;

			if(edit_serv_portranges_new_icmp(debuglvl, port_ptr) < 0)
				retval = -1;
		}
		else if(strncmp(choice_ptr, "GRE", 3) == 0)
		{
			/* gre has no ports */
			port_ptr->protocol = 47;
			port_ptr->src_low  =  0;
			port_ptr->src_high =  0;
			port_ptr->dst_low  =  0;
			port_ptr->dst_high =  0;
		}
		else if(strncmp(choice_ptr, "ESP", 3) == 0)
		{
			/* gre has no ports */
			port_ptr->protocol = 50;
			port_ptr->src_low  =  0;
			port_ptr->src_high =  0;
			port_ptr->dst_low  =  0;
			port_ptr->dst_high =  0;
		}
		else if(strncmp(choice_ptr, "AH", 2) == 0)
		{
			/* gre has no ports */
			port_ptr->protocol = 51;
			port_ptr->src_low  =  0;
			port_ptr->src_high =  0;
			port_ptr->dst_low  =  0;
			port_ptr->dst_high =  0;
		}
		else
		{
			(void)vrprint.error(-1, "Internal Error", "undefined protocol (%s).", choice_ptr);
			free(choice_ptr);
			return(-1);
		}

		/* free the choiceptr */
		free(choice_ptr);
	}

	if(retval == 0)
	{
		if(edit_serv_portranges_new_validate(debuglvl, ser_ptr, port_ptr) < 0)
			retval = -1;
	}

	return(retval);
}

/*	edit_serv_portranges_del

	Removes a portrange at place from a service.

	Returncodes:
		-1: error
		 0: not removed, user canceled
		 1: removed
*/
static int
edit_serv_portranges_del(const int debuglvl, int place, struct ServicesData_ *ser_ptr)
{
	int		i=0;
	d_list_node	*d_node = NULL;

	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* get user confimation */
	if(confirm("Delete portrange", "Are you sure?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 0)
		return(0);

	/* loop trough the list until we are at 'place'. */
	for(i = 1, d_node = ser_ptr->PortrangeList.top; d_node; d_node = d_node->next, i++)
	{
		/* here we are */
		if(place == i)
		{
			/* remove */
			if(d_list_remove_node(debuglvl, &ser_ptr->PortrangeList, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "unable to delete portrange '%d' from service '%s' (in: %s).", place, ser_ptr->name, __FUNCTION__);
				return(-1);
			}

			/* save */
			if(services_save_portranges(debuglvl, ser_ptr) < 0)
			{
				(void)vrprint.error(-1, "Error", "saveing portranges failed (in: %s).", __FUNCTION__);
				return(-1);
			}
			
			/* return 1 so the caller knows we removed it! */
			return(1);
		}
	}

	/* hmmm, this is really an error i think */
	return(0);
}


static int
edit_serv_portranges_init(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int			retval = 0;
	d_list_node		*d_node = NULL;
	int i=0;
	int			height = 30,
				width  = 64, // max width of host_name (32) + box (2) + 4 + 16
				startx = 5,
				starty = 5,
				max_height,
				max_width;
	d_list_node		*item_d_node = NULL,
				*number_d_node = NULL;
	struct portdata		*portrange_ptr = NULL;

	char			*port_string_ptr = NULL,
				*item_number_ptr = NULL,
				proto[5],
				src[12],
				dst[12],
				icmp_name[32];

	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* get number of items and calloc them */
	ServicesSection.EditServicePrt.n_items = ser_ptr->PortrangeList.len;
	ServicesSection.EditServicePrt.items = (ITEM **)calloc(ServicesSection.EditServicePrt.n_items + 1, sizeof(ITEM *));

	getmaxyx(stdscr, max_height, max_width);

	// get window height
	height = ServicesSection.EditServicePrt.n_items + 7; // 7 because: 3 above the list, 4 below
	if((height + 6) > max_height)
		height = max_height - 6;

	// place in the center of the screen
	starty = (max_height - height) / 2;
	startx = (max_width - width) / 2;

	// string item list
//TODO
	d_list_setup(debuglvl, &ServicesSection.EditServicePrt.item_list, free);
	item_d_node = ServicesSection.EditServicePrt.item_list.top;

	// number item list
	d_list_setup(debuglvl, &ServicesSection.EditServicePrt.item_number_list, free);
	number_d_node = ServicesSection.EditServicePrt.item_number_list.top;

	for(i = 0, d_node = ser_ptr->PortrangeList.top; d_node ; d_node = d_node->next, i++)
	{
		portrange_ptr = d_node->data;

		// item number
		if(!(item_number_ptr = malloc(5)))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
			return(-1);
		}

		snprintf(item_number_ptr, 5, "%3d", i+1);

		// string
		if(!(port_string_ptr = malloc(57)))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
			return(-1);
		}

		if(portrange_ptr->protocol == 1)
		{
			strcpy(proto, "ICMP");
			snprintf(src, sizeof(src), "TYPE=%2d", portrange_ptr->dst_low);
			snprintf(dst, sizeof(dst), "CODE=%2d", portrange_ptr->dst_high);

			get_icmp_name_short(portrange_ptr->dst_low, portrange_ptr->dst_high, icmp_name, sizeof(icmp_name), 0);
			snprintf(port_string_ptr, 57, "ICMP: T:%2d, C:%2d (%s)", portrange_ptr->dst_low, portrange_ptr->dst_high, icmp_name);
		}
		else if(portrange_ptr->protocol == 6)
			strcpy(proto, "TCP");
		else if(portrange_ptr->protocol == 17)
			strcpy(proto, "UDP");
		else if(portrange_ptr->protocol == 47)
		{
			/* no ports */
			snprintf(port_string_ptr, 57, "GRE : this protocol doesn't use ports.");
		}
		else if(portrange_ptr->protocol == 50)
		{
			/* no ports */
			snprintf(port_string_ptr, 57, "ESP : this protocol doesn't use ports.");
		}
		else if(portrange_ptr->protocol == 51)
		{
			/* no ports */
			snprintf(port_string_ptr, 57, "AH  : this protocol doesn't use ports.");
		}

		if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
		{
			if(portrange_ptr->src_high == 0)
				snprintf(src, sizeof(src), "%d", portrange_ptr->src_low);
			else
				snprintf(src, sizeof(src), "%d:%d", portrange_ptr->src_low, portrange_ptr->src_high);

			if(portrange_ptr->dst_high == 0)
				snprintf(dst, sizeof(dst), "%d", portrange_ptr->dst_low);
			else
				snprintf(dst, sizeof(dst), "%d:%d", portrange_ptr->dst_low, portrange_ptr->dst_high);

			snprintf(port_string_ptr, 57, "%-4s: %-12s -> %-12s", proto, src, dst);

		}

		/* load all into item array */
		ServicesSection.EditServicePrt.items[i] = new_item(item_number_ptr, port_string_ptr);

		/* store in list */
		if(d_list_append(debuglvl, &ServicesSection.EditServicePrt.item_list, port_string_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "Unable to insert in list (port_string_ptr).");
			return(-1);
		}
		if(d_list_append(debuglvl, &ServicesSection.EditServicePrt.item_number_list, item_number_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "Unable to insert in list (item_number_ptr).");
			return(-1);
		}

	}
	ServicesSection.EditServicePrt.items[ServicesSection.EditServicePrt.n_items] = (ITEM *)NULL;

	// create win and panel
	ServicesSection.EditServicePrt.win = newwin(height, width, starty, startx);
	wbkgd(ServicesSection.EditServicePrt.win, COLOR_PAIR(5));
	keypad(ServicesSection.EditServicePrt.win, TRUE);
	ServicesSection.EditServicePrt.panel[0] = new_panel(ServicesSection.EditServicePrt.win);

	ServicesSection.EditServicePrt.menu = new_menu((ITEM **)ServicesSection.EditServicePrt.items);
	set_menu_win(ServicesSection.EditServicePrt.menu, ServicesSection.EditServicePrt.win);
	set_menu_sub(ServicesSection.EditServicePrt.menu, derwin(ServicesSection.EditServicePrt.win, height-7, width-2, 3, 1));
	set_menu_format(ServicesSection.EditServicePrt.menu, height-7, 1);

	box(ServicesSection.EditServicePrt.win, 0, 0);
	print_in_middle(ServicesSection.EditServicePrt.win, 1, 0, width, "Portranges", COLOR_PAIR(5));
	mvwaddch(ServicesSection.EditServicePrt.win, 2, 0, ACS_LTEE);
	mvwhline(ServicesSection.EditServicePrt.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ServicesSection.EditServicePrt.win, 2, width-1, ACS_RTEE);

	set_menu_back(ServicesSection.EditServicePrt.menu, COLOR_PAIR(5));
	set_menu_fore(ServicesSection.EditServicePrt.menu, COLOR_PAIR(3));

	post_menu(ServicesSection.EditServicePrt.menu);

	mvwaddch(ServicesSection.EditServicePrt.win, height-4, 0, ACS_LTEE);
	mvwhline(ServicesSection.EditServicePrt.win, height-4, 1, ACS_HLINE, width-2);
	mvwaddch(ServicesSection.EditServicePrt.win, height-4, width-1, ACS_RTEE);

	mvwprintw(ServicesSection.EditServicePrt.win, height-3, 2, "Usage:  <INS> to insert a portrange");
	mvwprintw(ServicesSection.EditServicePrt.win, height-2, 2, "        <DEL> to remove a portrange");

	update_panels();
	doupdate();
	wrefresh(ServicesSection.EditServicePrt.win);

	return(retval);
}


static int
edit_serv_portranges_destroy(const int debuglvl)
{
	int	retval=0,
		i;

	// Un post form and free the memory
	unpost_menu(ServicesSection.EditServicePrt.menu);
	free_menu(ServicesSection.EditServicePrt.menu);

	for(i=0;i<ServicesSection.EditServicePrt.n_items;i++)
	{
		free_item(ServicesSection.EditServicePrt.items[i]);
	}
	free(ServicesSection.EditServicePrt.items);

	del_panel(ServicesSection.EditServicePrt.panel[0]);
	destroy_win(ServicesSection.EditServicePrt.win);

	d_list_cleanup(debuglvl, &ServicesSection.EditServicePrt.item_list);
	d_list_cleanup(debuglvl, &ServicesSection.EditServicePrt.item_number_list);

	update_panels();
	doupdate();

	return(retval);
}

static int
edit_serv_portranges(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int	quit=0,
		reload=0,
		result=0,
		ch,
		retval=0;

	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: edit_serv_portranges).");
		return(-1);
	}

	if(edit_serv_portranges_init(debuglvl, ser_ptr) < 0)
		return(-1);

	while(quit == 0)
	{
		if(reload == 1)
		{
			result = edit_serv_portranges_destroy(debuglvl);
			if(result < 0)
				return(-1);

			result = edit_serv_portranges_init(debuglvl, ser_ptr);
			if(result < 0)
				return(-1);

			// refresh screen
			update_panels();
			doupdate();
			reload=0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ServicesSection.EditServicePrt.win);
			switch(ch)
			{
				case 27: //esc
				case 'q':
				case 'Q':
				case KEY_F(10): //quit

					quit = 1;
					break;

				case KEY_IC:
				case 'i':
				case 'I':

					if(edit_serv_portranges_new(debuglvl, ser_ptr) < 0)
						retval= -1;
					else
						reload = 1;

					break;

				case KEY_DC:
				case 'd':
				case 'D':
				{
					ITEM *cur;
					if((cur = current_item(ServicesSection.EditServicePrt.menu)))
					{
						result = edit_serv_portranges_del(debuglvl, atoi((char *)item_name(cur)), ser_ptr);
						if(result < 0)
							retval=-1;
						else if(result == 1)
							reload=1;
					}

					break;
				}

				case KEY_DOWN:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ServicesSection.EditServicePrt.menu, REQ_LAST_ITEM); // page down
					break;
				case KEY_F(12):
				case 'h':
				case 'H':
					print_help(debuglvl, ":[VUURMUUR:SERVICE:EDIT:PORTRANGE]:");
					break;
			}
		}
	}

	result = edit_serv_portranges_destroy(debuglvl);
	if(result < 0)
		retval = -1;

	return(retval);

}


static int
edit_service_save(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int	retval=0,
		i = 0,
		result;
	char	helper[sizeof(ser_ptr->helper)] = "";

	// check for changed fields
	for(i=0; i < ServicesSection.EditService.n_fields; i++)
	{
		if(field_status(ServicesSection.EditService.fields[i]) == TRUE)
		{
			// active
			if(strncmp(field_buffer(ServicesSection.EditService.fields[i], 1), "act", 3) == 0)
			{
				ser_ptr->status = ST_CHANGED;
				if(strncasecmp(field_buffer(ServicesSection.EditService.fields[i], 0), "yes", 3) == 0)
				{
					ser_ptr->active = 1;
				}
				else if(strncasecmp(field_buffer(ServicesSection.EditService.fields[i], 0), "no", 2) == 0)
				{
					ser_ptr->active = 0;
				}
				else
				{
					ser_ptr->active = -1;
				}

				result=sf->tell(debuglvl, serv_backend, ser_ptr->name, "ACTIVE", ser_ptr->active ? "Yes" : "No", 1, TYPE_SERVICE);
				if(result < 0)
				{
					vrprint.error(-1, "Error", "Unable to write changes to backend (in: edit_service_save, active).");
					retval=-1;
				}
			}
			// broadcast
			else if(strncmp(field_buffer(ServicesSection.EditService.fields[i], 1), "bro", 3) == 0)
			{
				ser_ptr->status = ST_CHANGED;

				if(strncasecmp(field_buffer(ServicesSection.EditService.fields[i], 0), "yes", 3) == 0)
				{
					ser_ptr->broadcast = 1;
				}
				else if(strncasecmp(field_buffer(ServicesSection.EditService.fields[i], 0), "no", 2) == 0)
				{
					ser_ptr->broadcast = 0;
				}
				else
				{
					ser_ptr->broadcast = -1;
				}

				result=sf->tell(debuglvl, serv_backend, ser_ptr->name, "BROADCAST", ser_ptr->broadcast ? "Yes" : "No", 1, TYPE_SERVICE);
				if(result < 0){
					vrprint.error(-1, "Error", "Unable to write changes to backend (in: edit_service_save, broadcast).");
					retval = -1;
				}
			}
			/* helper field */
			else if(strncmp(field_buffer(ServicesSection.EditService.fields[i], 1), "helper", 6) == 0)
			{
				strlcpy(helper, ser_ptr->helper, sizeof(helper));

				if(!(copy_field2buf(	ser_ptr->helper,
							field_buffer(ServicesSection.EditService.fields[i], 0),
							sizeof(ser_ptr->helper))))
					return(-1);

				if(sf->tell(debuglvl, serv_backend, ser_ptr->name, "HELPER", ser_ptr->helper, 1, TYPE_SERVICE) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				(void)vrprint.info("Info", "service '%s' has been changed: protocol helper is set to '%s' (was: '%s').", ser_ptr->name, ser_ptr->helper, helper);
			}
			/* comment */
			else if(strncmp(field_buffer(ServicesSection.EditService.fields[i], 1), "comment", 7) == 0)
			{
				result=sf->tell(debuglvl, serv_backend, ser_ptr->name, "COMMENT", field_buffer(ServicesSection.EditService.fields[i], 0), 1, TYPE_SERVICE);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_service_save, comment).");
					retval = -1;
				}
			}

		}
	}
	return(retval);
}


static int
edit_service_init(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int		i,
			rows,
			cols,
			comment_y=0,
			comment_x=0;
	int		height, width, startx, starty, max_height, max_width;
	struct portdata *portrange_ptr = NULL;
	d_list_node	*d_node = NULL;

	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* get the screen dimentions */
	getmaxyx(stdscr, max_height, max_width);

	height = 20 + ser_ptr->PortrangeList.len;
	if(height > max_height - 6)
		height = max_height - 6;

	width = 54;

	startx = 3;
	starty = 24;

	/* 4 fields: active, broadcast, comment and helper */
	ServicesSection.EditService.n_fields = 4;
	if(!(ServicesSection.EditService.fields = (FIELD **)calloc(ServicesSection.EditService.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* active */
	ServicesSection.EditService.fields[0] = new_field(1, 3, 3, 1, 0, 1);
	set_field_buffer(ServicesSection.EditService.fields[0], 0, ser_ptr->active ? "Yes" : "No");
	set_field_buffer(ServicesSection.EditService.fields[0], 1, "act");

	/* broadcast */
	ServicesSection.EditService.fields[1] = new_field(1, 3, 6, 1, 0, 1);
	set_field_buffer(ServicesSection.EditService.fields[1], 0, ser_ptr->broadcast ? "Yes" : "No");
	set_field_buffer(ServicesSection.EditService.fields[1], 1, "bro");

	/* comment field size */
	comment_y = 5;
	comment_x = 48;

	/* load the comment from the backend */
	if(sf->ask(debuglvl, serv_backend, ser_ptr->name, "COMMENT", ServicesSection.comment, sizeof(ServicesSection.comment), TYPE_SERVICE, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	ServicesSection.EditService.fields[2] = new_field(comment_y, comment_x, 9, 1, 0, 1);
	set_field_buffer(ServicesSection.EditService.fields[2], 0, ServicesSection.comment);
	set_field_buffer(ServicesSection.EditService.fields[2], 1, "comment");

	/* active */
	ServicesSection.EditService.fields[3] = new_field(1, 32, 3, 17, 0, 1);
	set_field_buffer(ServicesSection.EditService.fields[3], 0, ser_ptr->helper);
	set_field_buffer(ServicesSection.EditService.fields[3], 1, "helper");

	
	ServicesSection.EditService.fields[4] = NULL;

	for(i = 0; i < ServicesSection.EditService.n_fields; i++)
	{
		// set field options
		set_field_back(ServicesSection.EditService.fields[i], COLOR_PAIR(3));
		field_opts_off(ServicesSection.EditService.fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(ServicesSection.EditService.fields[i], FALSE);
	}

	/* create window and panel */
	ServicesSection.EditService.win = create_newwin(height, width, startx, starty, "Edit Service", COLOR_PAIR(5));
	keypad(ServicesSection.EditService.win, TRUE);
	ServicesSection.EditService.panel[0] = new_panel(ServicesSection.EditService.win);

	/* create and post form */
	ServicesSection.EditService.form = new_form(ServicesSection.EditService.fields);
	scale_form(ServicesSection.EditService.form, &rows, &cols);
	set_form_win(ServicesSection.EditService.form, ServicesSection.EditService.win);
	set_form_sub(ServicesSection.EditService.form, derwin(ServicesSection.EditService.win, rows, cols, 1, 2));
	post_form(ServicesSection.EditService.form);

	/* print labels */
	mvwprintw(ServicesSection.EditService.win, 1, 2,  "Name: %s", ser_ptr->name);
	mvwprintw(ServicesSection.EditService.win, 3, 2,  "Active:");
	mvwprintw(ServicesSection.EditService.win, 3, 18, "Protocol Helper:");
	mvwprintw(ServicesSection.EditService.win, 6, 2,  "Broadcast:");
	mvwprintw(ServicesSection.EditService.win, 9, 2,  "Comment:");
	mvwprintw(ServicesSection.EditService.win, 16, 1, "Press <F6> to manage the portranges of this service.");
	
	if(height > 16+4)
	{
		mvwprintw(ServicesSection.EditService.win, 18, 2, "List of portranges:");
		if(ser_ptr->PortrangeList.len == 0)
			mvwprintw(ServicesSection.EditService.win, 19, 2, "  No portranges defined yet.");
		else
		{
			for(d_node = ser_ptr->PortrangeList.top, i = 1; d_node; d_node = d_node->next, i++)
			{
				if(!(portrange_ptr = d_node->data))
					return(-1);

				if(portrange_ptr->protocol == 6)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  TCP : ");
				else if(portrange_ptr->protocol == 17)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  UDP : ");
				else if(portrange_ptr->protocol == 1)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  ICMP: ");
				else if(portrange_ptr->protocol == 47)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  GRE : ");
				else if(portrange_ptr->protocol == 50)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  ESP : ");
				else if(portrange_ptr->protocol == 51)
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  AH  : ");
				else
				{
					mvwprintw(ServicesSection.EditService.win, 18+i, 2, "  %3d : ", portrange_ptr->protocol);
				}

				if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
				{
					if(portrange_ptr->src_high == 0)
						wprintw(ServicesSection.EditService.win, "%11d", portrange_ptr->src_low);
					else
						wprintw(ServicesSection.EditService.win, "%5d:%5d", portrange_ptr->src_low, portrange_ptr->src_high);

					wprintw(ServicesSection.EditService.win, " -> ");

					if(portrange_ptr->dst_high == 0)
						wprintw(ServicesSection.EditService.win, "%d", portrange_ptr->dst_low);
					else
						wprintw(ServicesSection.EditService.win, "%d:%d", portrange_ptr->dst_low, portrange_ptr->dst_high);
				}
				else if(portrange_ptr->protocol == 47 || portrange_ptr->protocol == 50 || portrange_ptr->protocol == 51)
				{
					wprintw(ServicesSection.EditService.win, "uses no ports.");
				}
				else if(portrange_ptr->protocol == 1)
				{
					wprintw(ServicesSection.EditService.win, "type: %d, code: %d.", portrange_ptr->dst_low, portrange_ptr->dst_high);
				}
				else
				{
					wprintw(ServicesSection.EditService.win, "unknown.");
				}

				if(18+i == height-1) /* -1 is for the border */
				{
					if((ser_ptr->PortrangeList.len - i) > 1)
					{
						mvwprintw(ServicesSection.EditService.win, 18+i+1, 2, "There are %d more portranges. Press F6 to manage.", ser_ptr->PortrangeList.len - i);
						break;
					}
				}
			}
		}
	}

	// position the cursor in the active field
	pos_form_cursor(ServicesSection.EditService.form);

	return(0);
}


static int
edit_service(const int debuglvl, Services *services, const char *name)
{
	int			ch,		/* for recording keystrokes */
				i,
				quit = 0,
				not_defined = 0,
				retval = 0;
	struct ServicesData_	*ser_ptr = NULL;
	FIELD			*cur = NULL,
				*prev = NULL;

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* search the service */
	if(!(ser_ptr = search_service(debuglvl, services, (char *)name)))
	{
		(void)vrprint.error(-1, "Internal Error", "service '%s' was not found in memory (in: %s).", name, __FUNCTION__);
		return(-1);
	}

	/* init */
	if(edit_service_init(debuglvl, ser_ptr) < 0)
		return(-1);

	// print (or not) initial warning about the group being empty.
	if(ser_ptr->PortrangeList.len == 0)
	{
		wattron(ServicesSection.EditService.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		mvwprintw(ServicesSection.EditService.win, 15, 9, " Warning: no port(range)s defined! ");
		wattroff(ServicesSection.EditService.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
	}

	pos_form_cursor(ServicesSection.EditService.form);
	cur = current_field(ServicesSection.EditService.form);
	wrefresh(ServicesSection.EditService.win);
	update_panels();
	doupdate();

	// Loop through to get user requests
	while(quit == 0)
	{
		draw_field_active_mark(cur, prev, ServicesSection.EditService.win, ServicesSection.EditService.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		ch = wgetch(ServicesSection.EditService.win);

		not_defined = 0;

		if(strncmp(field_buffer(cur, 1), "comment", 7) == 0)
		{
			if(nav_field_comment(ServicesSection.EditService.form, ch) < 0)
				not_defined = 1;
		}
		else if(strncmp(field_buffer(cur, 1), "helper", 6) == 0)
		{
			if(nav_field_simpletext(ServicesSection.EditService.form, ch) < 0)
				not_defined = 1;
		}
		else if(strncmp(field_buffer(cur, 1), "act", 3) == 0 ||
			strncmp(field_buffer(cur, 1), "bro", 3) == 0)
		{
			if(nav_field_yesno(ServicesSection.EditService.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined == 1)
		{
			switch(ch)
			{
				case KEY_F(6):
				case 'e':
				case 'E':
					/* F6 opens the portranges section */
//TODO
					edit_serv_portranges(debuglvl, ser_ptr);
					
					break;
				
				case 27: //esc
				case KEY_F(10):
				case 'q':
				case 'Q':
					quit=1;
					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					form_driver(ServicesSection.EditService.form, REQ_NEXT_FIELD);
					form_driver(ServicesSection.EditService.form, REQ_END_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ServicesSection.EditService.form, REQ_PREV_FIELD);
					form_driver(ServicesSection.EditService.form, REQ_END_LINE);
					break;

				case KEY_F(12):
				case 'h':
				case 'H':
					print_help(debuglvl, ":[VUURMUUR:SERVICE:EDIT]:");
					break;
			}
		}

		prev = cur;
		cur = current_field(ServicesSection.EditService.form);

		/* print or erase warning about the group being empty. */
		if(ser_ptr->PortrangeList.len == 0)
		{
			wattron(ServicesSection.EditService.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
			mvwprintw(ServicesSection.EditService.win, 15, 9, " Warning: no port(range)s defined! ");
			wattroff(ServicesSection.EditService.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		}
		else
			mvwprintw(ServicesSection.EditService.win, 15, 9, "                                   ");


		wrefresh(ServicesSection.EditService.win);
		pos_form_cursor(ServicesSection.EditService.form);
	}


	/* save */
		
	/* save the service */
	if(edit_service_save(debuglvl, ser_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving the service failed (in: %s).", __FUNCTION__);
		retval = -1;
	}
		
	/* save the portranges */
	if(services_save_portranges(debuglvl, ser_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving the portranges failed (in: %s).", __FUNCTION__);
		retval = -1;
	}


	/* cleanup */
	
	/* Un post form and free the memory */
	unpost_form(ServicesSection.EditService.form);
	free_form(ServicesSection.EditService.form);

	for(i=0; i < ServicesSection.EditService.n_fields;i++)
	{
		free_field(ServicesSection.EditService.fields[i]);
	}
	free(ServicesSection.EditService.fields);

	del_panel(ServicesSection.EditService.panel[0]);
	destroy_win(ServicesSection.EditService.win);

	/* clear comment */
	strcpy(ServicesSection.comment, "");

	update_panels();
	doupdate();
	
	return(retval);
}


static int
init_services_section(const int debuglvl, Services *services, int height, int width, int startx, int starty)
{
	int			retval=0,
				i;
	struct ServicesData_	*ser_ptr;
	d_list_node		*d_node = NULL;

	ServicesSection.list_items = services->list.len;
	ServicesSection.items = (ITEM **)calloc(ServicesSection.list_items + 1, sizeof(ITEM *));

	for(i = 0, d_node = services->list.top; d_node ; d_node = d_node->next)
	{
		if(!(ser_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "ser_ptr == NULL! (in: init_services_section).");
			return(-1);
		}

		// load all services into memory but special service 'all'.
		if(strcmp(ser_ptr->name, "all") != 0)
		{
			ServicesSection.items[i] = new_item(ser_ptr->name, NULL);
			i++;
		}
	}

	ServicesSection.items[ServicesSection.list_items] = (ITEM *)NULL;

	ServicesSection.win = newwin(height, width, startx, starty);
	wbkgd(ServicesSection.win, COLOR_PAIR(5));
	keypad(ServicesSection.win, TRUE);
	ServicesSection.panel[0] = new_panel(ServicesSection.win);

	ServicesSection.menu = new_menu((ITEM **)ServicesSection.items);
	set_menu_win(ServicesSection.menu, ServicesSection.win);
	set_menu_sub(ServicesSection.menu, derwin(ServicesSection.win, height-7, width-2, 3, 1));
	set_menu_format(ServicesSection.menu, height-8, 1);

	box(ServicesSection.win, 0, 0);
	print_in_middle(ServicesSection.win, 1, 0, width, "Services", COLOR_PAIR(5));
	mvwaddch(ServicesSection.win, 2, 0, ACS_LTEE);
	mvwhline(ServicesSection.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ServicesSection.win, 2, width-1, ACS_RTEE);

	set_menu_back(ServicesSection.menu, COLOR_PAIR(5));
	set_menu_fore(ServicesSection.menu, COLOR_PAIR(3));

	post_menu(ServicesSection.menu);

	mvwaddch(ServicesSection.win, height-5, 0, ACS_LTEE);
	mvwhline(ServicesSection.win, height-5, 1, ACS_HLINE, width-2);
	mvwaddch(ServicesSection.win, height-5, width-1, ACS_RTEE);

	mvwprintw(ServicesSection.win, height-4, 2, "<RET> to edit the service");
	mvwprintw(ServicesSection.win, height-3, 2, "<INS> to insert a new service");
	mvwprintw(ServicesSection.win, height-2, 2, "<DEL> to removes the service");

	return(retval);
}


static int
destroy_services_section(void)
{
	int retval=0, i;

	unpost_menu(ServicesSection.menu);
	free_menu(ServicesSection.menu);
	for(i = 0; i < ServicesSection.list_items; ++i)
		free_item(ServicesSection.items[i]);

	free(ServicesSection.items);

	del_panel(ServicesSection.panel[0]);
	destroy_win(ServicesSection.win);

	return(retval);
}


void
services_section(const int debuglvl, Services *services, struct rgx_ *reg)
{
	int result=0, quit=0, reload=0;
	int ch;
	int height, width, startx, starty;
	char *new_name_ptr;
	ITEM *cur = NULL;

	// todo
	height = LINES-8;
	width = 34;
	startx = 4;
	starty = 1;

	result = init_services_section(debuglvl, services, height, width, startx, starty);
	if(result < 0)
		return;

	update_panels();
	doupdate();

	while(quit == 0)
	{
		if(reload == 1)
		{
			//status_print(status_win, "Reloading services...");

			result = destroy_services_section();
			if(result < 0)
				return;

			result = init_services_section(debuglvl, services, height, width, startx, starty);
			if(result < 0)
				return;

			update_panels();
			doupdate();
			reload=0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ServicesSection.win);

			switch(ch)
			{
				case KEY_LEFT:
				case 27:
				case 'q':
				case 'Q':
				case KEY_F(10): //quit

					quit = 1;
					break;

				case KEY_IC: //insert
				case 'i':
				case 'I':

					new_name_ptr = input_box(32, "New Service", "Please enter the name of the new service:");
					if(new_name_ptr != NULL)
					{
						if(validate_servicename(debuglvl, new_name_ptr, reg->servicename) == 0)
						{
							result = new_service(debuglvl, services, new_name_ptr, TYPE_SERVICE);
							if(result == 0)
							{
								reload = 1;
								edit_service(debuglvl, services, new_name_ptr);
							}
							else
							{
								(void)vrprint.error(-1, "Error", "creating new service failed.");
							}
						}
						free(new_name_ptr);
					}
					break;

				case KEY_DC: // delete
				case 'd':
				case 'D':

					cur = current_item(ServicesSection.menu);
					if(cur)
					{
						if(confirm("Delete", "Are you sure?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{

							result = delete_service(debuglvl, services, (char *)item_name(cur), TYPE_SERVICE);
							if(result < 0)
							{
								(void)vrprint.error(-1, "Error", "Deleting %s failed.", (char *)item_name(cur));
							}

							reload=1;
						}
					}
					break;
				case KEY_DOWN:
					menu_driver(ServicesSection.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ServicesSection.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ServicesSection.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ServicesSection.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ServicesSection.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ServicesSection.menu, REQ_LAST_ITEM); // page down
					break;
				case KEY_RIGHT:
				case 10: // enter
				case 'e':
				case 'E':

					cur = current_item(ServicesSection.menu);
					if(cur)
					{
						draw_top_menu_edit_services(top_win);
						(void)edit_service(debuglvl, services, (char *)item_name(cur));
						draw_top_menu_main(debuglvl, top_win);
					}

					break;

				case KEY_F(12):
				case 'h':
				case 'H':

					print_help(debuglvl, ":[VUURMUUR:SERVICES]:");
					break;
			}
		}
	}

	destroy_services_section();

	update_panels();
	doupdate();
}
