/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

// minimun width = 13
// TODO: check maximum width of both values
int
confirm(char *title, char *text, chtype forecolor, chtype backcolor, int def)
{
	int	retval = 0;
	ITEM	**menu_items;
	MENU	*confirm_menu;
	PANEL 	*my_panels[1];
	WINDOW	*confirm_win;
	ITEM	*cur;

	int	height = 7,
		width  = 25,
		startx = 5,
		starty = 5,
		max_x  = 0,
		max_y  = 0;
	char	*choices[] = {"Yes", "No"};

	int 	n_choices = 2,
		ch,
		quit = 0,
		i;

	char	*print_title;


	/* safety */
	if(!title || !text)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem.");
		return(-1);
	}

	if(width < strlen(text))
		width = strlen(text) + 6;

	getmaxyx(stdscr, max_y, max_x);
	startx = (max_x - width) /2;
	starty = (max_y - height)/2;

	/* get some mem */
	if(!(print_title = malloc(strlen(title)+3)))
		return(-1);

	snprintf(print_title, strlen(title)+3, " %s ", title);

	// first display the menu
	menu_items = (ITEM **)calloc(n_choices + 1, sizeof(ITEM *));
	if(menu_items == NULL)
		return(-1);

	for(i = 0; i < n_choices; ++i)
	{
		menu_items[i] = new_item(choices[i], NULL);
	}
	menu_items[n_choices] = (ITEM *)NULL;

	confirm_menu = new_menu((ITEM **)menu_items);

	confirm_win = newwin(height, width, starty, startx);
	wbkgd(confirm_win, backcolor);
	keypad(confirm_win, TRUE);
	wrefresh(confirm_win);

	my_panels[0] = new_panel(confirm_win);

	set_menu_win(confirm_menu, confirm_win);
	set_menu_sub(confirm_menu, derwin(confirm_win, height-4, 10, 4, (width)/2-5));

	set_menu_format(confirm_menu, height-4, 2);

	box(confirm_win, 0, 0);
	print_in_middle(confirm_win, 0, 0, width, print_title, backcolor);
	print_in_middle(confirm_win, 2, 0, width, text, backcolor);

	set_menu_back(confirm_menu, backcolor);
	set_menu_fore(confirm_menu, forecolor);

	post_menu(confirm_menu);


	/* set the cursor to the 'no' position */
	if(!def)
	{
		menu_driver(confirm_menu, REQ_RIGHT_ITEM);
	}

	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(confirm_win);
		switch(ch)
		{
			case KEY_DOWN:
				menu_driver(confirm_menu, REQ_LEFT_ITEM);
				break;
			case KEY_UP:
				menu_driver(confirm_menu, REQ_RIGHT_ITEM);
				break;
			case KEY_LEFT:
				menu_driver(confirm_menu, REQ_LEFT_ITEM);
				break;
			case KEY_RIGHT:
				menu_driver(confirm_menu, REQ_RIGHT_ITEM);
				break;

			case 10: // enter
			{
				cur = current_item(confirm_menu);
				if(strncasecmp((char *)item_name(cur), "Yes", 3) == 0)
				{
					retval=1;
				}

				quit=1;
				break;
			}

			case 'y':
			case 'Y':
				retval=1;
				quit=1;
				break;

			case 'n':
			case 'N':
				retval=0;
				quit=1;
				break;

			case KEY_F(10):
				quit=1;
				break;
		}
	}

	unpost_menu(confirm_menu);
	free_menu(confirm_menu);
	for(i = 0; i < n_choices; ++i)
		free_item(menu_items[i]);

	free(menu_items);

	del_panel(my_panels[0]);

	destroy_win(confirm_win);

	free(print_title);
	update_panels();
	doupdate();
      
	return(retval);
}

char *
input_box(int length, char *title, char *description)
{
	WINDOW  *ib_win;
	PANEL   *my_panels[1];
	FIELD   **fields;
	FORM    *my_form;
	int height, width, startx, starty, max_height, max_width, ch=0, rows, cols, quit=0, i;
	char *result_ptr=NULL, *temp_ptr=NULL;

	// create a buffer with the size of 'length'
	if(!(temp_ptr = malloc(length)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return NULL;
	}
    
	// set the window size
	getmaxyx(stdscr, max_height, max_width);
	height = 8;
	if(length < 16)
	{
		width = 37; // minimum TODO: why 37?
	}
	else if(length+8 > max_width)
	{
		status_print(status_win, "Window too big!");
		free(temp_ptr);
		return NULL;
	}
	else
	{  
		width = length+8;
		if(strlen(title)+8 > width)
			width = strlen(title)+8;

		if(strlen(description)+8 > width)
			width = strlen(description)+8;
	}

	// print on the centre of the screen
	startx = (max_height-height)/2;
	starty = (max_width-width)/2;

	// create window
	ib_win = create_newwin(height, width, startx, starty, title, COLOR_PAIR(5));

	my_panels[0] = new_panel(ib_win);

	fields = (FIELD **)calloc(1 + 1, sizeof(FIELD *));

	fields[0] = new_field(1, length-1, 3, ((width-length)/2)-2, 0, 0);

	set_field_back(fields[0], COLOR_PAIR(3));
	field_opts_off(fields[0], O_AUTOSKIP);
	// set status to false
	set_field_status(fields[0], FALSE);

	my_form = new_form(fields);
	scale_form(my_form, &rows, &cols);
	keypad(ib_win, TRUE);
	set_form_win(my_form, ib_win);
	set_form_sub(my_form, derwin(ib_win, rows, cols, 1, 2));
	post_form(my_form);

	mvwprintw(ib_win, 2, 4, "%s", description);
	mvwprintw(ib_win, 6, 4, "Note: whitespaces not allowed.");
  
	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(ib_win);
		switch(ch)
		{
			case KEY_F(5): //help
				status_print(status_win, "FIXME: help not yet implemented");
				break;
			case 10: // enter
				// Go to next field
				form_driver(my_form, REQ_NEXT_FIELD);
				form_driver(my_form, REQ_END_LINE);
				quit=1;
				break;
			case KEY_BACKSPACE:
			case 127:
				form_driver(my_form, REQ_PREV_CHAR);
				form_driver(my_form, REQ_DEL_CHAR);
				form_driver(my_form, REQ_END_LINE);
				break;
			case KEY_DC:
				form_driver(my_form, REQ_PREV_CHAR);
				form_driver(my_form, REQ_DEL_CHAR);
				form_driver(my_form, REQ_END_LINE);
				break;
			default:
				// If this is a normal character, it gets printed
				form_driver(my_form, ch);
				break;
		}
	}

	//status_print(status_win, "data: '%s' (%d)", field_buffer(fields[0], 0), length);
	strlcpy(temp_ptr, field_buffer(fields[0], 0), length);

	// get the length of the entry
	for(i=0; temp_ptr[i] != ' ' && i < length-1; i++);

	if(!(result_ptr = malloc(i+1)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return NULL;
	}

	strncpy(result_ptr, temp_ptr, i);
	result_ptr[i]='\0';

	if(result_ptr[0] == '\0')
	{
		free(result_ptr);
		result_ptr = NULL;
	}

	free(temp_ptr);
  
	unpost_form(my_form);
	free_form(my_form);

	free_field(fields[0]);
	free(fields);

	del_panel(my_panels[0]);
	destroy_win(ib_win);

	update_panels();
	doupdate();

	return(result_ptr);
}

int vuumuurconf_print_warning(char *title, char *fmt, ...)
{
	int retval=0, size;
	va_list ap;
	char long_str[512];
	char temp_str[512];

	WINDOW  *err_win, *print_err_win;
	PANEL   *my_panels[1];
	int height, width, startx, starty, max_height, max_width, ch;

	snprintf(long_str, sizeof(long_str), "Warning: ");
	size=strlen(long_str);

	va_start(ap, fmt);
	vsprintf(temp_str, fmt, ap);

	size=size+strlen(temp_str);
	if(size < 512)
	{
		strcat(long_str, temp_str);
	}
	else
	{
		vrprint.error(-1, "Error", "string to long (size = %d, while max is %d)", size, 512);
		retval=-1;
	}

	va_end(ap);

	getmaxyx(stdscr, max_height, max_width);

	// window size
	height = 8;

	width = strlen(long_str)+13;
	if(width > max_width)
	{
		width = max_width - 10;
	}

	starty = (max_height - height) / 2;
	startx = (max_width - width) / 2;

	//status_print(status_win, "h %d, w %d, sx %d, sy: %d", height, width, startx, starty);
	err_win = create_newwin(height, width, starty, startx, title, COLOR_PAIR(CP_BLUE_WHITE));
	print_err_win = newwin(height-4, width-6, starty+2, startx+3);
	wbkgd(print_err_win, COLOR_PAIR(CP_BLUE_WHITE));
	my_panels[0] = new_panel(err_win);
	keypad(err_win, TRUE);

	wprintw(print_err_win, "%s", long_str);

	mvwprintw(err_win, height-2, 2, "Press any key to continue...");

	update_panels();
	doupdate();

	ch = wgetch(print_err_win);

	del_panel(my_panels[0]);
	destroy_win(print_err_win);
	destroy_win(err_win);

	update_panels();
	doupdate();

	return(retval);
}

int vuumuurconf_print_error(int error_no, char *title, char *fmt, ...)
{
	int retval=0, size;
	va_list ap;
	char long_str[512];
	char temp_str[512];

	WINDOW  *err_win, *print_err_win;
	PANEL   *my_panels[1];
	int height, width, startx, starty, max_height, max_width, ch;

	snprintf(long_str, sizeof(long_str), "Error: ");
	size=strlen(long_str);

	va_start(ap, fmt);
	vsprintf(temp_str, fmt, ap);

	size=size+strlen(temp_str);
	if(size < 512)
	{
		strcat(long_str, temp_str);
	}
	else
	{
		vrprint.error(-1, "Error", "string to long (size = %d, while max is %d)", size, 512);
		retval=-1;
	}

	va_end(ap);
  
	getmaxyx(stdscr, max_height, max_width);

	// window size
	height = 8;

	width = strlen(long_str)+13;
	if(width > max_width)
	{
		width = max_width - 10;
	}

	starty = (max_height - height) / 2;
	startx = (max_width - width) / 2;
	
	//status_print(status_win, "h %d, w %d, sx %d, sy: %d", height, width, startx, starty);
	err_win = create_newwin(height, width, starty, startx, title, COLOR_PAIR(6)|A_BOLD);
	print_err_win = newwin(height-4, width-6, starty+2, startx+3);
	wbkgd(print_err_win, COLOR_PAIR(CP_YELLOW_RED)|A_BOLD);
	my_panels[0] = new_panel(err_win);
	keypad(err_win, TRUE);

	wprintw(print_err_win, "%s (%d)", long_str, error_no);
	mvwprintw(err_win, height-2, 2, "Press any key to continue...");

//  mvwprintw(err_win, 2, 4, "%s (%d)", long_str, error_no);

//  mvwprintw(err_win, 4, 4, "mh %d mw %d", max_height, max_width);

//  mvwprintw(err_win, 2, 2, "   / \\");
//  mvwprintw(err_win, 3, 2, "  / | \\");
//  mvwprintw(err_win, 4, 2, " /  .  \\");
//  mvwprintw(err_win, 5, 2, " -------");
  
	update_panels();
	doupdate();

	ch = wgetch(print_err_win);

	del_panel(my_panels[0]);
	destroy_win(print_err_win);
	destroy_win(err_win);

	update_panels();
	doupdate();

	return(retval);
}

int
vuumuurconf_print_info(char *title, char *fmt, ...)
{
	int retval=0, size;
	va_list ap;
	char long_str[512];
	char temp_str[512];

	strcpy(long_str, "");
	size=strlen(long_str);

	va_start(ap, fmt);
	vsprintf(temp_str, fmt, ap);

	size=size+strlen(temp_str);
	if(size < 512)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		(void)vrprint.error(-1, "Error", "string to long (size = %d, while max is %d)", size, sizeof(long_str));
		return(-1);
	}

	va_end(ap);

	wprintw(mainlog_win, "%s\n", long_str);

	update_panels();
	doupdate();

	return(retval);
}

/*
	draws a box where the user can make a choice
	returns a pointer to the selection, or NULL in case
	of error or no selection

	the user must free the selection
*/
char *
selectbox(char *title, char *text, int n_choices, char **choices, unsigned int cols)
{
	ITEM	**menu_items;
	MENU	*confirm_menu;
	PANEL	*my_panels[1];
	WINDOW	*confirm_win;

	int	height,
		width,
		startx=0,
		starty=0,
		max_height=0,
		max_width=0;

	int	ch,
		quit=0,
		i;

	size_t	len = 0,
		item_len = 0,
		min_len = 0;

	char	*print_title,
		*select_ptr = NULL;

	size_t	size = 0;

	/*
		cols can never be 0 ofcourse
	*/
	if(cols == 0)
		cols = 1;

	/*
		get the screen size
	*/
	getmaxyx(stdscr, max_height, max_width);

	/*
		determine the minimal width of the window
	*/
	len = strlen(title);
	if(strlen(text) > len)
		len = strlen(text);

	for(i = 0; i < n_choices; i++)
	{
		if(strlen(choices[i]) > item_len)
		{
			item_len = strlen(choices[i]);
		}
	}

	if(item_len * cols + 8 > max_width)
		cols = 1;

	min_len = len + 8;

	if(cols == 1)
		width = item_len + 8;
	else
	{
		width = item_len * cols + 8;
	}

	if(width < min_len)
		width = min_len;

	/*
		set height, if it is to big we set starty
	*/
	height = 6 + (n_choices / cols);
	if(height > max_height - 8)
	{
		starty = 3;
		height = max_height - 8;
	}

	/*
		center the window
	*/
	starty = (max_height - height)/2;
	startx = (max_width - width)/2;

	print_title = malloc(strlen(title)+3);
	if(print_title == NULL)
		return(NULL);

	snprintf(print_title, strlen(title)+3, " %s ", title);

	if(!(menu_items = (ITEM **)calloc(n_choices + 1, sizeof(ITEM *))))
		return(NULL);

	for(i = 0; i < n_choices; ++i)
	{
		menu_items[i] = new_item(choices[i], NULL);
	}
	menu_items[n_choices] = (ITEM *)NULL;

	confirm_menu = new_menu((ITEM **)menu_items);

	confirm_win = newwin(height, width, starty, startx);
	wbkgd(confirm_win, COLOR_PAIR(5));
	keypad(confirm_win, TRUE);
	wrefresh(confirm_win);

	my_panels[0] = new_panel(confirm_win);

	set_menu_win(confirm_menu, confirm_win);
	set_menu_sub(confirm_menu, derwin(confirm_win, height-5, width-3, 4, 2));

	set_menu_format(confirm_menu, height-5, cols);

	box(confirm_win, 0, 0);
	print_in_middle(confirm_win, 0, 0, width, print_title, COLOR_PAIR(5));
	print_in_middle(confirm_win, 2, 0, width, text, COLOR_PAIR(5));

	set_menu_back(confirm_menu, COLOR_PAIR(5));
	set_menu_fore(confirm_menu, COLOR_PAIR(3));

	/*
		make sure the colums are filled top->down, not left->right
	*/
	menu_opts_off(confirm_menu, O_ROWMAJOR);

	post_menu(confirm_menu);

	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(confirm_win);
		switch(ch)
		{
			case KEY_DOWN:
				menu_driver(confirm_menu, REQ_DOWN_ITEM);
				break;

			case KEY_UP:
				menu_driver(confirm_menu, REQ_UP_ITEM);
				break;

			case KEY_LEFT:
				if(cols > 1)
					menu_driver(confirm_menu, REQ_LEFT_ITEM);
				else
					quit = 1;

				break;

			/*
				not so nice code: if we only have one column we fall
				trough...
			*/
			case KEY_RIGHT:
				if(cols > 1)
				{
					menu_driver(confirm_menu, REQ_RIGHT_ITEM);
					break;
				}

//			case KEY_RIGHT:
			case 32:
			case 10: // enter
			{
				ITEM *cur;

				cur = current_item(confirm_menu);

				size = strlen((char *)item_name(cur))+1;
				if(size == 0)
				{
					(void)vrprint.error(-1, "Internal Error", "could not determine the size of the selection (in: %s).", __FUNCTION__);
					return(NULL);
				}

				if(!(select_ptr = malloc(size)))
					return(NULL);

				strlcpy(select_ptr, item_name(cur), size);

				quit=1;
				break;
			}

//			case KEY_LEFT:
			case KEY_F(10):
			case 'q':
			case 27:
				quit=1;
				break;
		}
	}

	/*
		cleanup
	*/
	unpost_menu(confirm_menu);
	free_menu(confirm_menu);

	for(i = 0; i < n_choices; ++i)
		free_item(menu_items[i]);

	free(menu_items);

	del_panel(my_panels[0]);

	destroy_win(confirm_win);

	free(print_title);

	update_panels();
	doupdate();

	return(select_ptr);
}

int
status_print(WINDOW *local_win, char *fmt, ...)
{
	va_list ap;
	char	long_str[100];

	if(!local_win || !fmt)
		return(-1);

	werase(local_win);
	box(local_win, 0 , 0);

	va_start(ap, fmt);
	vsprintf(long_str, fmt, ap);
	va_end(ap);

	wattron(local_win, COLOR_PAIR(3));
	mvwprintw(local_win, 0, 2, " Status ");
	wattroff(local_win, COLOR_PAIR(3));

	wattron(local_win, COLOR_PAIR(4) | A_BOLD);
	mvwprintw(local_win, 1, 2, "%s", long_str);
	wattroff(local_win, COLOR_PAIR(4)| A_BOLD);
	wrefresh(local_win);

	return(0);
}
