/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

int
main(int argc, char *argv[])
{
	Interfaces	interfaces;
	Zones		zones;
	Services	services;
	Rules		rules;

	PANEL		*main_panels[4];
	int		retval=0,
			optch,
			max_height,
			max_width;

	static char	optstring[] = "c:d:";

	struct rgx_	reg;

	int		debuglvl = 0;

	/* some defaults */
	vuurmuur_semid = -1;
	vuurmuur_shmid = -1;
	vuurmuurlog_semid = -1;
	vuurmuurlog_shmid = -1;

	libvuurmuur_verbose = 0;
//	debug = 0;
	conf.loglevel_cmdline = 0;

	vrprint.logger = "vuurmuur_conf";
	vrprint.error = libvuurmuur_stdoutprint_error;
	vrprint.warning = libvuurmuur_stdoutprint_warning;
	vrprint.info = libvuurmuur_stdoutprint_info;
	vrprint.debug = libvuurmuur_stdoutprint_debug;

	/* configfiles */
	memset(conf.configfile_location, 0, sizeof(conf.configfile_location));
	strlcpy(conf.configfile_location, VUURMUUR_CONFIGFILE, sizeof(conf.configfile_location));

	memset(vccnf.configfile_location, 0, sizeof(vccnf.configfile_location));
	strlcpy(vccnf.configfile_location, VUURMUURCONF_CONFIGFILE, sizeof(vccnf.configfile_location));

	vccnf.win_fore = COLOR_BLUE;
	vccnf.win_back = COLOR_WHITE;

	while((optch = getopt(argc, argv, optstring)) != -1 )
	{
		switch(optch)
		{
			/* configfile */
			case 'c' :

				if(strlcpy(conf.configfile_location, optarg, sizeof(conf.configfile_location)) >= sizeof(conf.configfile_location))
				{
					(void)vrprint.error(-1, "Error", "commandline arg too long (-c).");
					exit(EXIT_FAILURE);
				}
				break;

			case 'd' :
				/* debugging */
				fprintf(stdout, "vuurmuur-conf: debugging enabled.\n");

				/* convert the debug string and check the result */
				debuglvl = atoi(optarg);
				if(debuglvl < 0 || debuglvl > HIGH)
				{
					fprintf(stdout, "vuurmuur-conf: illegal debug level: %d\n", debuglvl);
					debuglvl = 0;
				}

				fprintf(stdout, "vuurmuur-conf: debug level: %d\n", debuglvl);
				break;

			default:
				(void)vrprint.error(-1, "Error", "unknown option.");
				exit(EXIT_FAILURE);
		}
	}

	/* Initialize curses */
	(void)initscr();
	(void)start_color();
	(void)cbreak();
	(void)noecho();
	(void)keypad(stdscr, (bool)TRUE);

	/* Initialize all the colors */
	(void)init_pair(CP_RED_BLUE,     COLOR_RED,    COLOR_BLUE);
	(void)init_pair(CP_GREEN_BLUE,   COLOR_GREEN,  COLOR_BLUE);
	(void)init_pair(CP_WHITE_BLUE,   COLOR_WHITE,  COLOR_BLUE);
	(void)init_pair(CP_YELLOW_BLUE,  COLOR_YELLOW, COLOR_BLUE);
	(void)init_pair(CP_BLUE_WHITE,   COLOR_BLUE,   COLOR_WHITE);
	(void)init_pair(CP_YELLOW_RED,   COLOR_YELLOW, COLOR_RED);
	(void)init_pair(CP_WHITE_RED,    COLOR_WHITE,  COLOR_RED);
	(void)init_pair(CP_WHITE_GREEN,  COLOR_WHITE,  COLOR_GREEN);
	(void)init_pair(CP_RED_WHITE,    COLOR_RED,    COLOR_WHITE);
	(void)init_pair(CP_YELLOW_WHITE, COLOR_YELLOW, COLOR_WHITE);
	(void)init_pair(CP_CYAN_BLUE,    COLOR_CYAN,   COLOR_BLUE);
	(void)init_pair(CP_CYAN_WHITE,   COLOR_CYAN,   COLOR_WHITE);
	(void)init_pair(CP_GREEN_WHITE,  COLOR_GREEN,  COLOR_WHITE);
	(void)init_pair(CP_CYAN_GREEN,   COLOR_CYAN,   COLOR_GREEN);
	(void)init_pair(CP_BLACK_WHITE,  COLOR_BLACK,  COLOR_WHITE);
	(void)init_pair(CP_WHITE_CYAN,   COLOR_WHITE,  COLOR_CYAN);
	(void)init_pair(CP_WHITE_BLACK,  COLOR_WHITE,  COLOR_BLACK);
	(void)init_pair(CP_WHITE_YELLOW, COLOR_WHITE,  COLOR_YELLOW);

	vccnf.win_fore = COLOR_BLUE;
	vccnf.win_back = COLOR_WHITE;

	init_pair(CP_WIN,       vccnf.win_fore, vccnf.win_back);
	init_pair(CP_WIN_REV,   vccnf.win_back, vccnf.win_fore);
	init_pair(CP_WIN_MARK,  COLOR_RED,      vccnf.win_back);
	init_pair(CP_WIN_FIELD, COLOR_WHITE,    COLOR_BLUE);

	vccnf.color_win       = COLOR_PAIR(CP_WIN);
	vccnf.color_win_rev   = COLOR_PAIR(CP_WIN_REV);
	vccnf.color_win_mark  = COLOR_PAIR(CP_WIN_MARK);
	vccnf.color_win_field = COLOR_PAIR(CP_WIN_FIELD);

	vccnf.color_bgd     = COLOR_PAIR(CP_WHITE_BLUE);
	vccnf.color_bgd_rev = COLOR_PAIR(CP_BLUE_WHITE);



	/* create the three main windows */
	if(!(status_win = create_newwin(3, COLS, LINES-3, 0, NULL, (chtype)COLOR_PAIR(CP_WHITE_BLUE))))
		exit(EXIT_FAILURE);
	if(!(top_win = create_newwin(3, COLS, 0, 0, NULL, (chtype)COLOR_PAIR(CP_WHITE_BLUE))))
		exit(EXIT_FAILURE);
	if(!(main_win = create_newwin(LINES-6, COLS, 3, 0, NULL, (chtype)COLOR_PAIR(CP_WHITE_BLUE))))
		exit(EXIT_FAILURE);

	getmaxyx(stdscr, max_height, max_width);
	if(!(mainlog_win = newwin(max_height-8, max_width-2, 4, 1)))
		exit(EXIT_FAILURE);

	(void)wbkgd(mainlog_win, (chtype)COLOR_PAIR(CP_WHITE_BLUE));

	/* Attach a panel to each window */
	main_panels[0] = new_panel(top_win);
	main_panels[1] = new_panel(main_win);
	main_panels[2] = new_panel(status_win);
	main_panels[3] = new_panel(mainlog_win);
 
	(void)update_panels();
	(void)doupdate();

	/* init the vrprint functions for the Gui */
	vrprint.error = vuumuurconf_print_error;
	vrprint.warning = vuumuurconf_print_warning;

	if(status_print(status_win, "This is Vuurmuur-conf %d.%d.%d, Copyright (c) 2003-2005 by Victor Julien", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB) < 0)
		exit(EXIT_FAILURE);

	/* setup regexes */
	if(setup_rgx(1, &reg) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up regular expressions failed.");
		exit(EXIT_FAILURE);
	}

	// startup
	if(startup_screen(debuglvl, &rules, &zones, &services, &interfaces, &reg) < 0)
	{
		// delete panels and windows
		(void)del_panel(main_panels[0]);
		(void)del_panel(main_panels[1]);
		(void)del_panel(main_panels[2]);
		(void)del_panel(main_panels[3]);
	
		(void)destroy_win(top_win);
		(void)destroy_win(main_win);
		(void)destroy_win(status_win);
		// clear screen
		(void)refresh();
		// end ncurses mode
		(void)endwin();

		exit(EXIT_FAILURE);
	}

	status_print(status_win, "Ready.");

	/* setup statuslist */
	(void)setup_statuslist(debuglvl);
	/* mainmenu loop */
	(void)draw_top_menu_main(debuglvl, top_win);
	
	mm_status_checkall(debuglvl, NULL, &rules, &zones, &interfaces, &services);
	/* main menu loop */
	while(main_menu(debuglvl, &rules, &zones, &interfaces, &services, &reg) == 0);
	/* clean up the status list */
	d_list_cleanup(debuglvl, &VuurmuurStatus.StatusList);

	/* detach from shared memory, if we were attached */
	if(vuurmuur_shmp != NULL && vuurmuur_shmp != (char *)(-1) && vuurmuur_shmtable != 0)
	{
		if(SILENT_LOCK(vuurmuur_semid))
		{
			vuurmuur_shmtable->configtool.connected = 3;
			SILENT_UNLOCK(vuurmuur_semid);
		}
		(void)shmdt(vuurmuur_shmp);
	}
	if(vuurmuurlog_shmp != NULL && vuurmuurlog_shmp != (char *)(-1) && vuurmuurlog_shmtable != 0)
	{
		if(SILENT_LOCK(vuurmuurlog_semid))
		{
			vuurmuurlog_shmtable->configtool.connected = 3;
			SILENT_UNLOCK(vuurmuurlog_semid);
		}
		(void)shmdt(vuurmuurlog_shmp);
	}

	/* delete panels and windows */
	(void)del_panel(main_panels[0]);
	(void)del_panel(main_panels[1]);
	(void)del_panel(main_panels[2]);
	(void)del_panel(main_panels[3]);

	(void)destroy_win(mainlog_win);
	(void)destroy_win(top_win);
	(void)destroy_win(main_win);
	(void)destroy_win(status_win);
	/* clear screen */
	(void)refresh();

	/* end ncurses mode */
	(void)endwin();

	/* set error functions to the stdout versions */
	vrprint.error = libvuurmuur_stdoutprint_error;
	vrprint.warning = libvuurmuur_stdoutprint_warning;
	vrprint.info = libvuurmuur_stdoutprint_info;
	vrprint.debug = libvuurmuur_stdoutprint_debug;

	/* unload the backends */
	if(unload_backends(debuglvl, &PluginList) < 0)
	{
		(void)vrprint.error(-1, "Error", "unloading the backends failed.");
		retval=-1;
	}

	/* cleanup regexes */
	(void)setup_rgx(0, &reg);

	/* cleanup the lists */
	(void)destroy_serviceslist(debuglvl, &services);
	(void)destroy_zonedatalist(debuglvl, &zones);
	(void)rules_cleanup_list(debuglvl, &rules);
	(void)destroy_interfaceslist(debuglvl, &interfaces);
	return(retval);
}


void print_in_middle(WINDOW *win, int starty, int startx, int width, char *string, chtype color)
{
	int	length,
		x,
		y;
	float	temp;

	if(!win)
		win = stdscr;

	getyx(win, y, x);

	if(startx != 0)
		x = startx;

	if(starty != 0)
		y = starty;

	if(width == 0)
		width = 80;

	length = (int)strlen(string);
	temp = (width - length)/ 2;
	x = startx + (int)temp;
	wattron(win, color);
	mvwprintw(win, y, x, "%s", string);
	wattroff(win, color);
	refresh();
}

/*	create_newwin

	creates a window, sets it background and sets a title.

	Returns a pointer to the window or NULL in case of failure.
*/
WINDOW *
create_newwin(int height, int width, int starty, int startx, char *title, chtype ch)
{
	WINDOW *local_win;
	int size;
	char *title_ptr;

	// create the window
	if(!(local_win = newwin(height, width, starty, startx)))
		return(NULL);

	// box and background
	(void)box(local_win, 0 , 0);
	(void)wbkgd(local_win, ch);

	// draw title if we have one
	if(title != NULL)
	{
		size = (size_t)strlen(title);
		if(size+4 <= width)
		{
			title_ptr = malloc(size+3);

			if(title_ptr == NULL)
			{

			}
			else
			{
				snprintf(title_ptr, size+3, " %s ", title);

				mvwprintw(local_win, 0, (width-strlen(title_ptr))/2, title_ptr);

				free(title_ptr);
			}
		}
		else
		{
			(void)vrprint.warning("Warning", "title '%s' too long, window will be drawn without a title.", title);
		}
	}

	return(local_win);
}

// destroys a window, but first removes a border
void
destroy_win(WINDOW *local_win)
{
	// first remove the border
	wborder(local_win, ' ', ' ', ' ',' ',' ',' ',' ',' ');
	wrefresh(local_win);
	delwin(local_win);
}

/*	get_vuurmuur_pid

	Gets the pid and shm_id from vuurmuur.

	Returncodes:
		-1: error
		    otherwise the pid of vuurmuur
*/
pid_t
get_vuurmuur_pid(char *vuurmuur_pidfile_location, int *shmid)
{
	FILE	*fp;
	pid_t	pid=-1;
	char	line[32],
		pid_c[16],
		shm_c[16];

	if(!(fp = fopen(vuurmuur_pidfile_location, "r")))
		return(-1);

	if(fgets(line, sizeof(line), fp) != NULL)
	{
		sscanf(line, "%15s %15s", pid_c, shm_c);
		pid = atol(pid_c);
		*shmid = atoi(shm_c);
	}
	else
	{
		// no need to return, because pid isn't touched, so still -1
		(void)vrprint.error(-1, "Error", "Empty or corrupted pid file: '%s' (in: get_vuurmuur_pid).", vuurmuur_pidfile_location);
	}
  
	if(fclose(fp) < 0)
		return(-1);
  
	return(pid);
}


/*	startup_screen

	This is the splash-screen which calls the startup functions,
	like loading the plugins, zones, services etc.

	Returncodes:
		 0: ok
		-1: error
*/
int
startup_screen(const int debuglvl, Rules *rules, Zones *zones, Services *services, Interfaces *interfaces, struct rgx_ *reg)
{
	WINDOW	*startup_win,
		*startup_print_win;
	PANEL	*startup_panel[2];
	int	retval = 0,
		maxy = 0,
		maxx = 0,
		y = 0,
		x = 0,
		width  = 50,
		heigth = 15,
		result = 0,
		config_done = 0,
		cnfresult = 0;

	// get screensize and set windowlocation
	getmaxyx(stdscr, maxy, maxx);
	y = (maxy - heigth)/2;
	x = (maxx - width)/2;

	// create the windows and panels
	startup_win = create_newwin(heigth, width, y, x, "Vuurmuur-conf", COLOR_PAIR(CP_YELLOW_RED)|A_BOLD);
	startup_print_win = newwin(1, width-14, y+heigth-3, x+6);
	wbkgd(startup_print_win, COLOR_PAIR(CP_YELLOW_RED)|A_BOLD);
	startup_panel[0] = new_panel(startup_win);
	startup_panel[1] = new_panel(startup_print_win);
	update_panels();
	doupdate();

	// print the logo: it looks a bit weird here because of escape sequences
	// also print version
	mvwprintw(startup_win, 3, 4, "  \\\\   //           |\\\\ //|            ");
	mvwprintw(startup_win, 4, 4, "   \\\\ // | | | | |] ||\\//|| | | | | |] ");
	mvwprintw(startup_win, 5, 4, "    \\//  \\/| \\/| |\\ ||   || \\/| \\/| |\\ ");
	mvwprintw(startup_win, 6, 4, "                                Config ");
	mvwprintw(startup_win, 7, 4, "  ------------------------------------ ");
	mvwprintw(startup_win, 9, 4, "  Copyright (c) 2003-2005 by Victor Julien ");
	mvwprintw(startup_win, 10, 4, "  Version: %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
	mvwprintw(startup_win, 12, 5, "[");
	mvwprintw(startup_win, 12, 4+width-12, "]");

	/* initialize the config */
	werase(startup_print_win); wprintw(startup_print_win, "Initializing config..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	while(!config_done)
	{
		result = init_config(debuglvl, conf.configfile_location, &conf);
		if(result == VR_CNF_E_UNKNOWN_ERR || result == VR_CNF_E_PARAMETER)
			return(-1);
		else if(result == VR_CNF_E_FILE_PERMISSION)
		{
			return(-1);
		}
		else if(	result == VR_CNF_E_FILE_MISSING ||
				result == VR_CNF_E_MISSING_VAR  ||
				result == VR_CNF_E_ILLEGAL_VAR  ||
				result == VR_CNF_W_MISSING_VAR  ||
				result == VR_CNF_W_ILLEGAL_VAR)
		{
			if(confirm("Problem with config", "Do you want to edit the config now?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 1))
			{
				/* this prompt the user with the config menu */
				while((cnfresult = config_menu(debuglvl)) == 0);
				if(cnfresult < 0)
					return(-1);
			}
			else
			{
				/* if the user doesn't want to solve the problem we exit if we had an error
				   in case of a warning, we continue
				*/
				if(result == VR_CNF_E_MISSING_VAR || result == VR_CNF_E_FILE_MISSING)
					return(-1);
				else
				{
//TODO: print warning to warn the user that the config is not yet ok?
					config_done = 1;
				}
			}
		}
		else if(result == VR_CNF_OK)
		{
			werase(startup_print_win); wprintw(startup_print_win, "Initializing config... OK"); update_panels(); doupdate();
			config_done = 1;
		}
		else
		{
			(void)vrprint.error(-1, "Error", "Unknown return code from init_config. This can't be good.");
			return(-1);
		}

		if(config_done == 0)
		{
			werase(startup_print_win); wprintw(startup_print_win, "Reinitializing config..."); update_panels(); doupdate();
		}
	}

	/* config done, so now we can use logprinting */
	if(debuglvl >= LOW)
		vrprint.info = vuumuurconf_print_info;
	else
		vrprint.info = libvuurmuur_logprint_info;

	vrprint.debug = libvuurmuur_logprint_debug;


	/* initialize the vuurmuur conf config */
	config_done = 0;
	werase(startup_print_win); wprintw(startup_print_win, "Initializing settings..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	while(!config_done)
	{
		result = init_vcconfig(debuglvl, vccnf.configfile_location, &vccnf);
		if(result == VR_CNF_E_UNKNOWN_ERR || result == VR_CNF_E_PARAMETER)
			return(-1);
		else if(result == VR_CNF_E_FILE_PERMISSION)
		{
			return(-1);
		}
		else if(	result == VR_CNF_E_FILE_MISSING ||
				result == VR_CNF_E_MISSING_VAR  ||
				result == VR_CNF_E_ILLEGAL_VAR  ||
				result == VR_CNF_W_MISSING_VAR  ||
				result == VR_CNF_W_ILLEGAL_VAR)
		{
			if(confirm("Problem with settings", "Do you want to edit the settings now?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 1))
			{
				/* this prompt the user with the config menu */
				cnfresult = edit_vcconfig(debuglvl);
				if(cnfresult < 0)
					return(-1);
			}
			else
			{
				/* if the user doesn't want to solve the problem we exit if we had an error
				   in case of a warning, we continue
				*/
				if(result == VR_CNF_E_MISSING_VAR || result == VR_CNF_E_FILE_MISSING)
					return(-1);
				else
				{
//TODO: print warning to warn the user that the config is not yet ok?
					config_done = 1;
				}
			}
		}
		else if(result == VR_CNF_OK)
		{
			werase(startup_print_win); wprintw(startup_print_win, "Initializing settings... OK"); update_panels(); doupdate();
			config_done = 1;
		}
		else
		{
			(void)vrprint.error(-1, "Error", "Unknown return code from init_vcconfig. This can't be good.");
			return(-1);
		}

		if(config_done == 0)
		{
			werase(startup_print_win); wprintw(startup_print_win, "Reinitializing settings..."); update_panels(); doupdate();
		}
	}


	/* now load the backends */
	werase(startup_print_win); wprintw(startup_print_win, "Loading plugins..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	if(load_backends(debuglvl, &PluginList) < 0)
	{
		(void)vrprint.error(-1, "Error", "Unable to load backends.");
		return(-1);
	}
	werase(startup_print_win); wprintw(startup_print_win, "Loading plugins... OK"); update_panels(); doupdate();


	/* init services */
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Services..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	result = init_services(debuglvl, services, reg);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "Intializing the services failed.");
		return(-1);
	}
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Services... OK"); update_panels(); doupdate();

	/* init interfaces */
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Interfaces..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	result = init_interfaces(debuglvl, interfaces);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "Intializing the interfaces failed.");
		return(-1);
	}
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Interfaces... OK"); update_panels(); doupdate();

	/* init zones */
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Zones..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	result = init_zonedata(debuglvl, zones, interfaces, reg);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "Intializing the zones failed.");
		return(-1);
	}
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Zones... OK"); update_panels(); doupdate();

	/* init rules */
	werase(startup_print_win); wprintw(startup_print_win, "Initializing Rules..."); update_panels(); doupdate();
	if(debuglvl > LOW) sleep(1);
	result = rules_init_list(debuglvl, rules, reg);
	if(result < 0)
	{
		werase(startup_print_win); wprintw(startup_print_win, "Initializing Rules... FAILED"); update_panels(); doupdate();
		sleep(1);
	}
	else
	{
		werase(startup_print_win); wprintw(startup_print_win, "Initializing Rules... OK"); update_panels(); doupdate();
	}
	/* okay the rules are now loaded */
	vccnf.rules_loaded = 1;

	/* load the blockfile if any */
	if(conf.use_blocklist)
	{
		/* init blocklist */
		werase(startup_print_win); wprintw(startup_print_win, "Initializing BlockList..."); update_panels(); doupdate();
		if(debuglvl > LOW) sleep(1);
		result = blocklist_read_file(debuglvl, zones, &BlockList, 0);
		if(result < 0)
		{
			werase(startup_print_win); wprintw(startup_print_win, "Initializing BlockList... FAILED"); update_panels(); doupdate();
			sleep(1);
		}
		else
		{
			werase(startup_print_win); wprintw(startup_print_win, "Initializing BlockList... OK"); update_panels(); doupdate();
		}
	}


	/*
		try to connect to vuurmuur trough shm
	*/
	vuurmuur_shmtable = NULL;
	werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur..."); update_panels(); doupdate();
	vuurmuur_pid = get_vuurmuur_pid("/var/run/vuurmuur.pid", &vuurmuur_shmid);
	if(vuurmuur_shmid > 0)
	{
		/* attach to shared memory */
		vuurmuur_shmp = shmat(vuurmuur_shmid, 0, 0);
		if(vuurmuur_shmp == (char *)(-1))
		{
			(void)vrprint.error(-1, "Error", "Attaching to shared memory failed: %s.", strerror(errno));
		}
		else
		{
			vuurmuur_shmtable = (struct SHM_TABLE *)vuurmuur_shmp;
			vuurmuur_semid = vuurmuur_shmtable->sem_id;

			/* now try to connect to the shared memory */
			if(LOCK(vuurmuur_semid))
			{
				vuurmuur_shmtable->configtool.connected = 1;
				snprintf(vuurmuur_shmtable->configtool.name, sizeof(vuurmuur_shmtable->configtool.name), "Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
				UNLOCK(vuurmuur_semid);

				werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur... OK"); update_panels(); doupdate();
			}
			else
			{
				werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur... FAILED"); update_panels(); doupdate();
				vuurmuur_shmp = NULL;
				sleep(1);
			}
		}
	}
	else
	{
		werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur... FAILED"); update_panels(); doupdate();
		vuurmuur_shmp = NULL;
		sleep(1);
	}

	/*
		try to connect to vuurmuur trough shm
	*/
	vuurmuurlog_shmtable = NULL;
	werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur_log..."); update_panels(); doupdate();
	vuurmuurlog_pid = get_vuurmuur_pid("/var/run/vuurmuur_log.pid", &vuurmuurlog_shmid);
	if(vuurmuurlog_shmid > 0)
	{
		/* attach to shared memory */
		vuurmuurlog_shmp = shmat(vuurmuurlog_shmid, 0, 0);
		if(vuurmuurlog_shmp == (char *)(-1))
		{
			(void)vrprint.error(-1, "Error", "attaching to shared memory failed: %s.", strerror(errno));
		}
		else
		{
			vuurmuurlog_shmtable = (struct SHM_TABLE *)vuurmuurlog_shmp;
			vuurmuurlog_semid = vuurmuurlog_shmtable->sem_id;

			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "vuurmuur_log: sem_id: %d.", vuurmuurlog_semid);

			/* now try to connect to the shared memory */
			if(LOCK(vuurmuurlog_semid))
			{
				vuurmuurlog_shmtable->configtool.connected = 1;
				snprintf(vuurmuurlog_shmtable->configtool.name, sizeof(vuurmuurlog_shmtable->configtool.name), "Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
				UNLOCK(vuurmuurlog_semid);

				werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur_log... OK"); update_panels(); doupdate();
			}
			else
			{
				werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur_log... FAILED"); update_panels(); doupdate();
				vuurmuurlog_shmp = NULL;
				sleep(1);
			}
		}
	}
	else
	{
		werase(startup_print_win); wprintw(startup_print_win, "Connecting to Vuurmuur_log... FAILED"); update_panels(); doupdate();
		vuurmuurlog_shmp = NULL;
		sleep(1);
	}


	/* cleanup */
	del_panel(startup_panel[0]);
	del_panel(startup_panel[1]);

	destroy_win(startup_print_win);
	destroy_win(startup_win);

	update_panels();
	doupdate();

	return(retval);
}

/*	draw_field_active_mark

	Draws marks before and after the active field. If prev is non-NULL we first remove the
	marks from the previous field.
*/
void
draw_field_active_mark(const FIELD *cur, const FIELD *prev, WINDOW *formwin, FORM *form, const chtype ch)
{
	int	pos_x,
		pos_y,
		x,
		y,
		off_row,
		wrk_buff;

	/* safety */
	if(!cur || !formwin || !form)
		return;

	/* if supplied we remove the previous marking */
	if(prev)
	{
		if(field_info(prev, &y, &x, &pos_y, &pos_x, &off_row, &wrk_buff) < 0)
			return;

		mvwprintw(formwin, pos_y+1, pos_x, " ");
		mvwprintw(formwin, pos_y+1, pos_x + x + 3, " ");
	}

	/* draw our marking */
	if(field_info(cur, &y, &x, &pos_y, &pos_x, &off_row, &wrk_buff) < 0)
		return;

	wattron(formwin, ch);
	mvwprintw(formwin, pos_y+1, pos_x, ">");
	mvwprintw(formwin, pos_y+1, pos_x + x + 3, "<");
	wattroff(formwin, ch);
	wrefresh(formwin);

	/* restore cursor position */
	pos_form_cursor(form);
	return;
}


/*	copy_field2buf

	copies a buffer to another...

	Will copy for bufsize - 1 to leave space for '\0'.
	
	Returncodes:
		1: ok
		0: error
*/
int
copy_field2buf(char *buf, char *fieldbuf, size_t bufsize)
{
	size_t i = 0;

	/* safety */
	if(!buf || !fieldbuf)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(0);
	}

	/* copy while:
		1. we are inside the target buffers size
		2. we are inside the sources buffer size
	*/
	for(	i = 0;	fieldbuf[i] != ' '   &&
			i < bufsize - 1      &&
			i < strlen(fieldbuf);
		i++)
	{
		buf[i] = fieldbuf[i];
	}
	buf[i] = '\0';

	return(1);
}
